import uuid
from typing import Any, Iterable, Optional, Union

import click
from typing_extensions import Literal

import globus_sdk

from gcs_cli.helpers import requires_login
from gcs_cli.parsing import (
    EXPLICIT_NULL,
    collection_create_and_update_params,
    collection_id_arg,
    globus_command,
)
from gcs_cli.prettyprint import FORMAT_TEXT_RECORD, formatted_print
from gcs_cli.sdk_client.data import (
    CollectionDocument,
    GoogleCloudStorageCollectionPolicies,
    POSIXCollectionPolicies,
    POSIXStagingCollectionPolicies,
)
from gcs_cli.services.gcs import get_client


@globus_command("update", short_help="Update a Collection definition")
@collection_id_arg
@collection_create_and_update_params(create=False)
@click.option(
    "--delete-protected/--no-delete-protected",
    is_flag=True,
    help="Set the delete_protected flag for a mapped collection",
    default=None,
)
@requires_login
def collection_update(
    collection_id: uuid.UUID,
    delete_protected: Optional[bool],
    posix_sharing_group_allow: Optional[Iterable[str]] = None,
    posix_sharing_group_deny: Optional[Iterable[str]] = None,
    posix_staging_sharing_group_allow: Optional[Iterable[str]] = None,
    posix_staging_sharing_group_deny: Optional[Iterable[str]] = None,
    verify: Optional[
        Union[Literal["force"], Literal["disable"], Literal["default"]]
    ] = None,
    **kwargs: Any,
) -> None:
    """
    Update a Collection on the Endpoint into which the GCS CLI is logged in.
    """
    client = get_client()

    # convert keyword args as follows:
    # - filter out Nones
    # - pass through EXPLICIT_NULL as None
    converted_kwargs = {
        k: (str(v) if isinstance(v, uuid.UUID) else v if v != EXPLICIT_NULL else None)
        for k, v in kwargs.items()
        if v is not None
    }

    if delete_protected:
        converted_kwargs["delete_protected"] = True
    elif delete_protected is False:
        converted_kwargs["delete_protected"] = False

    if posix_sharing_group_allow is not None or posix_sharing_group_deny is not None:
        # Added in 5.4.8 for POSIX connector
        policies = {}
        if posix_sharing_group_allow is not None:
            policies["sharing_groups_allow"] = posix_sharing_group_allow

        if posix_sharing_group_deny is not None:
            policies["sharing_groups_deny"] = posix_sharing_group_deny

        converted_kwargs["policies"] = POSIXCollectionPolicies(**policies)

    if (
        posix_staging_sharing_group_allow is not None
        or posix_staging_sharing_group_deny is not None
    ):
        # Added in 5.4.10 for POSIX Staging connector
        policies = {}
        if posix_staging_sharing_group_allow is not None:
            policies["sharing_groups_allow"] = posix_staging_sharing_group_allow

        if posix_staging_sharing_group_deny is not None:
            policies["sharing_groups_deny"] = posix_staging_sharing_group_deny

        converted_kwargs["policies"] = POSIXStagingCollectionPolicies(**policies)

    # These are added in GCS 5.4.4, so if neither --enable-https or
    # --disable-https are set, then we'll not try to set the value for
    # backward compatibility
    if converted_kwargs.get("enable_https") and converted_kwargs.get("disable_https"):
        raise click.exceptions.UsageError(
            "Only one of --enable-https and --disable-https may be used"
        )
    if converted_kwargs.get("enable_https") is False:
        converted_kwargs.pop("enable_https")

    if converted_kwargs.get("disable_https"):
        converted_kwargs["enable_https"] = False
    converted_kwargs.pop("disable_https", None)

    # policy constructor ignores None
    if "google_project_id" in converted_kwargs:
        converted_kwargs["policies"] = GoogleCloudStorageCollectionPolicies(
            project=converted_kwargs.get("google_project_id")
        )
        del converted_kwargs["google_project_id"]
        if kwargs["google_project_id"] == EXPLICIT_NULL:
            converted_kwargs["policies"]["project"] = None

    if verify is not None:
        if verify.lower() == "force":
            converted_kwargs["force_verify"] = True
            converted_kwargs["disable_verify"] = False
        elif verify.lower() == "disable":
            converted_kwargs["force_verify"] = False
            converted_kwargs["disable_verify"] = True
        else:
            converted_kwargs["force_verify"] = False
            converted_kwargs["disable_verify"] = False

    doc = CollectionDocument(**converted_kwargs)

    try:
        res = client.update_collection(collection_id, doc)
    except globus_sdk.GlobusAPIError as e:
        # Check if the caller used a connector-specific option for the wrong storage
        # gateway type.
        if e.http_status == 422 and "policies" in converted_kwargs:
            raise click.ClickException(
                "This request used an option not supported by the collection type. "
                "Review the options and try again.\n"
                f'The service responded with "{e.message}"'
            )
        raise e

    formatted_print(res, fields=[("code", "code")], text_format=FORMAT_TEXT_RECORD)
