import pwd
from typing import (
    ClassVar,
    List,
    TYPE_CHECKING,
    Type,
)

from .base import (
    Check,
    CheckInvoker,
    MissingPropertyCheck,
    ObjectIsNotNoneCheck,
    PropertyConstraintCheck,
)

from ..models.credential import BlackPearlCredential, PosixCredential, S3Credential

if TYPE_CHECKING:
    from ..models.credential import Credential


class CredentialCheck(CheckInvoker):
    checkers: ClassVar[List[Type[Check["Credential"]]]] = []
    object_name = "credential"


@CredentialCheck.checker
class UC01(ObjectIsNotNoneCheck["Credential"]):
    ...


@CredentialCheck.property_checker(property_name="username")
class UC02(MissingPropertyCheck["Credential"]):
    ...


@CredentialCheck.property_checker(property_name="identity_id")
class UC03(MissingPropertyCheck["Credential"]):
    ...


@CredentialCheck.property_checker(property_name="identity_username")
class UC04(MissingPropertyCheck["Credential"]):
    ...


@CredentialCheck.property_checker(property_name="username")
class UC05(PropertyConstraintCheck["Credential"]):
    constraint_name = "guest collection local account exists"

    def value_is_ok(self, value: "Credential") -> bool:
        ok = True
        if isinstance(value, (PosixCredential)):
            if value.username:
                try:
                    pwd.getpwnam(value.username)
                except (KeyError, TypeError):
                    ok = False
        return ok


@CredentialCheck.property_checker(property_name="s3_key_id")
class UC06(MissingPropertyCheck["Credential"]):
    def value_is_ok(self, value: "Credential") -> bool:
        ok = True
        if isinstance(value, S3Credential):
            ok = super().value_is_ok(value)
        return ok


@CredentialCheck.property_checker(property_name="s3_secret_key")
class UC07(MissingPropertyCheck["Credential"]):
    def value_is_ok(self, value: "Credential") -> bool:
        ok = True
        if isinstance(value, S3Credential):
            ok = super().value_is_ok(value)
        return ok


@CredentialCheck.property_checker(property_name="access_id")
class UC08(MissingPropertyCheck["Credential"]):
    def value_is_ok(self, value: "Credential") -> bool:
        ok = True
        if isinstance(value, BlackPearlCredential):
            ok = super().value_is_ok(value)
        return ok


@CredentialCheck.property_checker(property_name="secret_key")
class UC09(MissingPropertyCheck["Credential"]):
    def value_is_ok(self, value: "Credential") -> bool:
        ok = True
        if isinstance(value, BlackPearlCredential):
            ok = super().value_is_ok(value)
        return ok
