import os

from typing import (
    ClassVar,
    List,
    Type,
    TYPE_CHECKING,
)

from .base import (
    Check,
    CheckInvoker,
    MissingPropertyCheck,
    ObjectIsNotNoneCheck,
    PropertyConstraintCheck,
)

from ..models.storage_gateway import (
    BlackPearlStorageGateway,
    CephStorageGateway,
    HPSSStorageGateway,
    S3StorageGateway,
)

if TYPE_CHECKING:
    from ..models.storage_gateway import StorageGateway  # noqa: F401


class StorageGatewayCheck(CheckInvoker["StorageGateway"]):
    checkers: ClassVar[List[Type[Check["StorageGateway"]]]] = []
    object_name = "storage_gateway"


@StorageGatewayCheck.checker
class SG01(ObjectIsNotNoneCheck["StorageGateway"]):
    ...


@StorageGatewayCheck.property_checker(property_name="display_name")
class SG02(MissingPropertyCheck["StorageGateway"]):
    ...


@StorageGatewayCheck.property_checker(property_name="display_name")
class SG03(PropertyConstraintCheck["StorageGateway"]):
    constraint_name = "value contains all whitespace"

    def value_is_ok(self, value: "StorageGateway") -> bool:
        ok = True
        if value.display_name is not None:
            display_name = value.display_name.strip()
            if not display_name:
                ok = False
        return ok


@StorageGatewayCheck.property_checker(property_name="allowed_domains")
class SG04(MissingPropertyCheck["StorageGateway"]):
    ...


@StorageGatewayCheck.property_checker(property_name="allowed_domains")
class SG05(PropertyConstraintCheck["StorageGateway"]):
    constraint_name = "must be set"

    def value_is_ok(self, value: "StorageGateway") -> bool:
        ok = True
        if value.allowed_domains is not None:
            allowed_domains = value.allowed_domains
            if len(allowed_domains) == 0:
                ok = False
        return ok


@StorageGatewayCheck.property_checker(property_name="allowed_domains")
class SG06(PropertyConstraintCheck["StorageGateway"]):
    constraint_name = ">1 allowed_domains without identity mappings"

    def value_is_ok(self, value: "StorageGateway") -> bool:
        ok = True
        if value.allowed_domains is not None:
            allowed_domains = value.allowed_domains
            if len(allowed_domains) > 1 and not value.identity_mappings:
                ok = False
        return ok


@StorageGatewayCheck.property_checker(property_name="load_dsi_module")
class SG07(PropertyConstraintCheck["StorageGateway"]):
    constraint_name = "load_dsi_module"
    allowed_dsis = ["file", "blackpearl", "s3", "ceph", "hpss_local"]

    def value_is_ok(self, value: "StorageGateway") -> bool:
        ok = True
        if (
            value.load_dsi_module is not None
            and value.load_dsi_module not in self.allowed_dsis
        ):
            ok = False
        return ok


@StorageGatewayCheck.property_checker(property_name="s3_endpoint")
class SG08(MissingPropertyCheck["StorageGateway"]):
    def value_is_ok(self, value: "StorageGateway") -> bool:
        ok = True
        if isinstance(value, (BlackPearlStorageGateway, S3StorageGateway)):
            ok = super().value_is_ok(value)
        return ok


@StorageGatewayCheck.property_checker(property_name="s3_user_credential_required")
class SG09(MissingPropertyCheck["StorageGateway"]):
    def value_is_ok(self, value: "StorageGateway") -> bool:
        ok = True
        if isinstance(value, S3StorageGateway):
            ok = super().value_is_ok(value)
        return ok


@StorageGatewayCheck.property_checker(property_name="bp_access_id_file")
class SG10(MissingPropertyCheck["StorageGateway"]):
    def value_is_ok(self, value: "StorageGateway") -> bool:
        ok = True
        if isinstance(value, BlackPearlStorageGateway):
            ok = super().value_is_ok(value)
        return ok


@StorageGatewayCheck.property_checker(property_name="bp_access_id_file")
class SG11(PropertyConstraintCheck["StorageGateway"]):
    # For blackpearl, the bp_access_id_file must exist on both the v4 and v5
    # nodes
    constraint_name = "bp_access_id_file does not exist"

    def value_is_ok(self, value: "StorageGateway") -> bool:
        ok = True
        if (
            isinstance(value, BlackPearlStorageGateway)
            and value.bp_access_id_file
            and not os.path.exists(value.bp_access_id_file)
        ):
            ok = False
        return ok


@StorageGatewayCheck.property_checker(property_name="ceph_admin_bucket")
class SG12(PropertyConstraintCheck["StorageGateway"]):
    # For ceph, in v4 we allowed specifying the admin bucket, but that is
    # currently not configurable in v5
    constraint_name = "unsupported custom admin bucket name"

    def value_is_ok(self, value: "StorageGateway") -> bool:
        ok = True
        if (
            isinstance(value, CephStorageGateway)
            and value.ceph_admin_bucket
            and value.ceph_admin_bucket != "admin"
        ):
            ok = False
        return ok


@StorageGatewayCheck.property_checker(property_name="ceph_admin_key_id")
class SG13(MissingPropertyCheck["StorageGateway"]):
    def value_is_ok(self, value: "StorageGateway") -> bool:
        ok = True
        if isinstance(value, CephStorageGateway):
            ok = super().value_is_ok(value)
        return ok


@StorageGatewayCheck.property_checker(property_name="ceph_admin_key_id")
class SG14(PropertyConstraintCheck["StorageGateway"]):
    constraint_name = "key id is empty"

    def value_is_ok(self, value: "StorageGateway") -> bool:
        ok = True
        if isinstance(value, CephStorageGateway) and not value.ceph_admin_key_id:
            ok = False
        return ok


@StorageGatewayCheck.property_checker(property_name="ceph_admin_secret_key")
class SG15(MissingPropertyCheck["StorageGateway"]):
    def value_is_ok(self, value: "StorageGateway") -> bool:
        ok = True
        if isinstance(value, CephStorageGateway):
            ok = super().value_is_ok(value)
        return ok


@StorageGatewayCheck.property_checker(property_name="ceph_admin_secret_key")
class SG16(PropertyConstraintCheck["StorageGateway"]):
    constraint_name = "secret key is empty"

    def value_is_ok(self, value: "StorageGateway") -> bool:
        ok = True
        if isinstance(value, CephStorageGateway) and not value.ceph_admin_secret_key:
            ok = False
        return ok


@StorageGatewayCheck.property_checker(property_name="login_name")
class SG17(MissingPropertyCheck["StorageGateway"]):
    def value_is_ok(self, value: "StorageGateway") -> bool:
        ok = True
        if isinstance(value, HPSSStorageGateway):
            ok = super().value_is_ok(value)
        return ok


# Removed: SG18 HPSS login_name must be 'hpssftp'


@StorageGatewayCheck.property_checker(property_name="authentication_mech")
class SG19(MissingPropertyCheck["StorageGateway"]):
    def value_is_ok(self, value: "StorageGateway") -> bool:
        ok = True
        if isinstance(value, HPSSStorageGateway):
            ok = super().value_is_ok(value)
        return ok


@StorageGatewayCheck.property_checker(property_name="authentication_mech")
class SG20(PropertyConstraintCheck["StorageGateway"]):
    constraint_name = "only 'unix' and 'krb5' are supported"

    def value_is_ok(self, value: "StorageGateway") -> bool:
        ok = True
        if isinstance(value, HPSSStorageGateway):
            if value.authentication_mech not in ["unix", "krb5"]:
                ok = False
        return ok


@StorageGatewayCheck.property_checker(property_name="authenticator")
class SG21(MissingPropertyCheck["StorageGateway"]):
    def value_is_ok(self, value: "StorageGateway") -> bool:
        ok = True
        if isinstance(value, HPSSStorageGateway):
            ok = super().value_is_ok(value)
        return ok


@StorageGatewayCheck.property_checker(property_name="authenticator")
class SG22(PropertyConstraintCheck["StorageGateway"]):
    constraint_name = (
        "use the format 'auth_keyfile|auth_keytab:<absolute_path_to_file>'"
    )

    def value_is_ok(self, value: "StorageGateway") -> bool:
        ok = True
        if isinstance(value, HPSSStorageGateway):
            tokens: List[str] = []
            if value.authenticator is not None:
                tokens = value.authenticator.split(":")
            if len(tokens) != 2:
                ok = False
            elif tokens[0] not in ["auth_keyfile", "auth_keytab"]:
                ok = False
            elif not tokens[1].startswith("/"):
                ok = False
        return ok


@StorageGatewayCheck.property_checker(property_name="authenticator")
class SG23(PropertyConstraintCheck["StorageGateway"]):
    constraint_name = "the HPSS credentials file does not exist"

    def value_is_ok(self, value: "StorageGateway") -> bool:
        ok = True
        if isinstance(value, HPSSStorageGateway):
            tokens: List[str] = []
            if value.authenticator is not None:
                tokens = value.authenticator.split(":")
            if len(tokens) == 2 and not os.path.exists(tokens[1]):
                ok = False
        return ok


@StorageGatewayCheck.property_checker(property_name="uda_checksum")
class SG24(MissingPropertyCheck["StorageGateway"]):
    def value_is_ok(self, value: "StorageGateway") -> bool:
        ok = True
        if isinstance(value, HPSSStorageGateway):
            ok = super().value_is_ok(value)
        return ok
