import pwd
from typing import cast, Dict, Iterable, List, Optional

import click
from globus_sdk import GlobusError, IdentityMap


from .decorators import requires_logged_in, requires_migration_plan
from ..models import MigrationPlan, S3Credential, S3CredentialYamlDict, S3StorageGateway
from ..parsers import S3MapEntry
from ..util import get_api


@click.command(
    "s3-credential-create", help="Create credential resources from S3 map file"
)
@click.option(
    "--domain",
    required=False,
    help="Domain to use for identities. Optional if there is only one domain "
    "associated with the migration plan.",
)
@click.argument(
    "USER",
    nargs=-1,
)
@requires_logged_in
@requires_migration_plan
def s3_credential_create(
    domain: Optional[str],
    user: Iterable[str],
    migration_plan: MigrationPlan,
) -> None:
    assert migration_plan.storage_gateway is not None
    if migration_plan.storage_gateway.load_dsi_module != "s3":
        raise click.ClickException(
            "This command can only be used with endpoints using the s3 connector"
        )
    storage_gateway = cast(S3StorageGateway, migration_plan.storage_gateway)
    mapfile_parser = storage_gateway.s3_mapfile_parser
    if mapfile_parser is None:
        raise click.ClickException("The storage gateway is missing an s3 map file.")

    credential_entries: Dict[str, S3MapEntry] = {}
    identity_map = IdentityMap(get_api("auth.globus.org"))
    domains = storage_gateway.allowed_domains

    if not domains:
        raise click.ClickException(
            "The storage gateway has no domains configured. Please use "
            "one of the identity-mapping commands to add an allowed domain "
            "to your endpoint."
        )

    if not domain:
        if len(domains) != 1:
            raise click.ClickException(
                "The storage gateway allows multiple domains, you'll need to "
                "explicitly indicate which credentials need to be associated with "
                "a particular domain."
            )
        else:
            domain = domains[0]
    elif domain not in domains:
        raise click.ClickException(
            f"The storage gateway is not configured to allow domain {domain}. "
            "Please use one of the identity-mapping commands to add an allowed "
            "domain to your endpoint."
        )

    # Decide which credentials to create, either all users or specific users
    if not user:
        for pwent in pwd.getpwall():
            if pwent.pw_name in mapfile_parser:
                globus_user = f"{pwent.pw_name}@{domain}"
                identity_map.add(globus_user)
                credential_entries[globus_user] = mapfile_parser[pwent.pw_name]
    else:
        for u in user:
            if u not in mapfile_parser:
                raise click.ClickException(f"No usable S3 mapfile entry for {u}")
            globus_user = f"{u}@{domain}"
            identity_map.add(globus_user)
            credential_entries[globus_user] = mapfile_parser[u]

    bad_user_identities = set()
    for identity_username, entry in credential_entries.items():
        try:
            identity = identity_map[identity_username]
        except (GlobusError, KeyError):
            bad_user_identities.add(identity_username)
            continue

        # First look for a credential that matches this identity_username
        # If there is not one, it was created by either by a
        # previous run of this program or by being attached to a guest
        # collection.
        credential_already_exists = False
        for cred in cast(List[S3Credential], migration_plan.credentials):
            if (
                cred.identity_username == identity_username
                and cred.username == identity_username
            ):
                # Already exists. If it has matching S3 key info, ignore
                # this one. If it has different S3 key info, ask for guidance
                if (
                    cred.s3_key_id != entry.access_key_id
                    or cred.s3_secret_key != entry.secret_key
                ):
                    if click.confirm(
                        "There already exists a credential for globus account "
                        f"{identity_username} and local user {entry.username} "
                        "with a different key. Replace the credential?"
                    ):
                        cred.s3_key_id = entry.access_key_id
                        cred.s3_secret_key = entry.secret_key
                credential_already_exists = True
                break

        if not credential_already_exists:
            # For this credential which is being created without
            # information an X.509 subject, we'll set the subject
            # to the identity_username. This must have a value in order to
            # generate a uuid5 for the filename for the credential that won't
            # change when other identity mapping commands are done.
            credential_data: S3CredentialYamlDict = {
                "username": identity_username,
                "subject": identity_username,
                "gsi_authz_callout_index": None,
                "identity_id": identity["id"],
                "identity_username": identity_username,
                "s3_key_id": entry.access_key_id,
                "s3_secret_key": entry.secret_key,
            }
            credential = S3Credential(credential_data)
            migration_plan.add_credential(credential)
            click.echo(f"Creating credential for {identity_username}")
    if bad_user_identities:
        click.echo(
            "\n".join(
                [
                    "Unable to resolve identities, credentials "
                    "were not created for them:"
                ]
                + [f"    {b}" for b in bad_user_identities]
            )
        )
    migration_plan.dump_credentials()
    if bad_user_identities:
        click.get_current_context().exit(1)
