import os
import pathlib
from typing import Any, Dict, Optional

from globus_sdk.tokenstorage import SQLiteAdapter

import globus_sdk

from . import util


AUTH_CLIENT_DATA_CONFIG_KEY = "auth_client_data"


def _get_template_client_id() -> str:
    GLOBUS_ENV = os.environ.get("GLOBUS_SDK_ENVIRONMENT", "default")
    default_id = "974dc87d-3261-4c7f-8e93-5f298fe7f43e"

    return {
        "sandbox": "48ddc584-3bcb-431e-94ca-2967643c90ef",
        "test": "0a692758-e147-4d19-8f7d-af5623874d17",
        "preview": "c9585fa7-31bc-4b7d-8155-7f8868d1d9e9",
        "integration": "30e01f99-6943-496c-9b4a-42c17d813edb",
        "staging": "cbb9e120-ac12-41ee-b966-dcd454cc0596",
    }.get(GLOBUS_ENV, default_id)


def _get_tokenstorage_db() -> str:
    storagedb_dir = pathlib.Path("~/.globus/gcs_migration").expanduser()
    os.makedirs(storagedb_dir, exist_ok=True, mode=0o0700)
    return str(storagedb_dir / "tokenstore.db")


def _get_tokenstorage_namespace() -> str:
    GLOBUS_PROFILE = os.environ.get("GLOBUS_PROFILE", "default")
    GLOBUS_ENV = os.environ.get("GLOBUS_SDK_ENVIRONMENT")

    if "/" in GLOBUS_PROFILE:
        raise ValueError("Not allowed to have '/' in GLOBUS_PROFILE!")
    if GLOBUS_ENV:
        return GLOBUS_ENV + "/" + GLOBUS_PROFILE
    else:
        return GLOBUS_PROFILE


def token_storage_adapter() -> SQLiteAdapter:
    if not hasattr(token_storage_adapter, "_instance"):
        old_umask = os.umask(0o77)
        try:
            # namespace is equal to the current environment
            token_storage_adapter._instance = SQLiteAdapter(  # type: ignore
                _get_tokenstorage_db(), namespace=_get_tokenstorage_namespace()
            )
        finally:
            os.umask(old_umask)
    return token_storage_adapter._instance  # type: ignore


def load_client_data() -> Dict[str, Any]:
    adapter = token_storage_adapter()
    conf = adapter.read_config(AUTH_CLIENT_DATA_CONFIG_KEY)
    return conf or {}


def internal_native_client() -> globus_sdk.NativeAppAuthClient:
    template_id = load_client_data().get("template_id", _get_template_client_id())
    return globus_sdk.NativeAppAuthClient(
        template_id,
        transport_params=util.get_transport_params(),
    )


def internal_auth_client(
    force_new_client: bool = False,
) -> globus_sdk.ConfidentialAppAuthClient:
    """
    Looks up the values for this CLI's Instance Client in config

    If none exists or force_new_client is True,
    registers a new Instance Client with GLobus Auth

    Returns a ConfidentialAppAuthClient
    """
    client_data = load_client_data()
    client_id: Optional[str] = client_data.get("client_id")
    client_secret: Optional[str] = client_data.get("client_secret")
    template_id: str = client_data.get("template_id", _get_template_client_id())
    template_client = internal_native_client()
    existing = client_id and client_secret

    # if we are forcing a new client, delete any existing client
    if force_new_client and existing:
        assert client_id and client_secret
        existing_client = globus_sdk.ConfidentialAppAuthClient(
            client_id, client_secret, transport_params=util.get_transport_params()
        )
        try:
            existing_client.delete("/v2/api/clients/{}".format(client_id))

        # if the client secret has been invalidated or the client has
        # already been removed, we continue on
        except globus_sdk.AuthAPIError:
            pass

    # if we require a new client to be made
    if force_new_client or not existing:
        # register a new instance client with auth
        body = {"client": {"template_id": template_id, "name": "GCS v4 Migration"}}
        res = template_client.post("/v2/api/clients", data=body)

        # get values and write to config
        credential_data = res["included"]["client_credential"]
        client_id = client_data["client_id"] = credential_data["client"]
        client_secret = client_data["client_secret"] = credential_data["secret"]

        token_storage_adapter().store_config(AUTH_CLIENT_DATA_CONFIG_KEY, client_data)

    assert client_id and client_secret
    return globus_sdk.ConfidentialAppAuthClient(
        client_id,
        client_secret,
        app_name="GCS v4 Migration",
        transport_params=util.get_transport_params(),
    )
