import os
import pathlib
from typing import cast, Generic, Iterable, List, Optional, Tuple, TypeVar, Union
from typing_extensions import TypedDict

from .do_not_modify import DO_NOT_MODIFY
from ..parsers import (
    BlackPearlAccessIdFileParser,
    BlackPearlConfig,
    CephConfig,
    GridFTPOptionParser,
    HPSSConfig,
    parse_restrict_paths,
    S3Config,
    S3MapfileParser,
)
from ..typing.gcs_documents import (
    BLACKPEARL_CONNECTOR_ID,
    BlackPearlStorageGatewayDict,
    BlackPearlStorageGatewayPoliciesDict,
    CEPH_CONNECTOR_ID,
    CephStorageGatewayDict,
    CephStorageGatewayPoliciesDict,
    HPSS_CONNECTOR_ID,
    HPSSStorageGatewayDict,
    HPSSStorageGatewayPoliciesDict,
    IdentityMappingDict,
    PathRestrictionsDict,
    POSIX_CONNECTOR_ID,
    PosixStorageGatewayDict,
    PosixStorageGatewayPoliciesDict,
    S3_CONNECTOR_ID,
    S3StorageGatewayDict,
    S3StorageGatewayPoliciesDict,
    StorageGatewayDict,
)
from ..util import yaml, yaml_object
from ..exceptions import UnsupportedUpdate


class StorageGatewayYamlDict(TypedDict):
    display_name: Optional[str]
    users_allow: Optional[List[str]]
    users_deny: Optional[List[str]]
    load_dsi_module: Optional[str]
    allowed_domains: Optional[List[str]]
    identity_mappings: Optional[List[IdentityMappingDict]]
    restrict_paths: Optional[PathRestrictionsDict]


class BlackPearlStorageGatewayYamlDict(StorageGatewayYamlDict, total=False):
    s3_endpoint: Optional[str]
    bp_access_id_file: Optional[str]


class HPSSStorageGatewayYamlDict(StorageGatewayYamlDict, total=False):
    login_name: Optional[str]
    authentication_mech: Optional[str]
    authenticator: Optional[str]
    uda_checksum: Optional[bool]


class PosixStorageGatewayYamlDict(StorageGatewayYamlDict, total=False):
    groups_allow: Optional[List[str]]
    groups_deny: Optional[List[str]]


class S3StorageGatewayYamlDict(StorageGatewayYamlDict, total=False):
    s3_endpoint: Optional[str]
    s3_buckets: Optional[List[str]]
    s3_user_credential_required: Optional[bool]
    s3_map_filename: Optional[str]


class CephStorageGatewayYamlDict(StorageGatewayYamlDict, total=False):
    s3_endpoint: Optional[str]
    ceph_admin_bucket: Optional[str]
    ceph_admin_key_id: Optional[str]
    ceph_admin_secret_key: Optional[str]


T = TypeVar("T", bound=StorageGatewayYamlDict)


class StorageGateway(Generic[T]):
    def __init__(self, data: Optional[T] = None) -> None:
        self.display_name: Optional[str] = None
        self.users_allow: Optional[List[str]] = None
        self.users_deny: Optional[List[str]] = None
        self.load_dsi_module: Optional[str] = None
        self.allowed_domains: Optional[List[str]] = None
        self.identity_mappings: Optional[List[IdentityMappingDict]] = None
        self.restrict_paths: Optional[PathRestrictionsDict] = None
        self.gridftp_config: Optional[GridFTPOptionParser] = None

        if data:
            self.__setstate__(data)

    @property
    def gridftp_config(self) -> Optional[GridFTPOptionParser]:
        return self._gridftp_config

    @gridftp_config.setter
    def gridftp_config(self, value: Optional[GridFTPOptionParser]) -> None:
        self._gridftp_config = value
        if value:
            if self.restrict_paths is None:
                rp = value.get_string_option("restrict_paths")
                if rp:
                    self.restrict_paths = parse_restrict_paths(rp)
            if self.load_dsi_module is None:
                self.load_dsi_module = value.get_string_option("load_dsi_module")

    @property
    def restrict_paths(self) -> Optional[PathRestrictionsDict]:
        return self._restrict_paths

    @restrict_paths.setter
    def restrict_paths(self, value: Optional[PathRestrictionsDict]) -> None:
        self._restrict_paths = value

    @property
    def allowed_domains(self) -> Optional[List[str]]:
        return self._allowed_domains

    @allowed_domains.setter
    def allowed_domains(self, value: Optional[List[str]]) -> None:
        self._allowed_domains = value

    def dump(self, root: Union[str, os.PathLike], omit_secrets: bool = False) -> None:
        path = pathlib.Path(root) / "storage_gateway.yaml"
        with open(path, "w") as f:
            print(DO_NOT_MODIFY, file=f)
            yaml.dump(self, f)

    @staticmethod
    def load(
        root: Union[str, os.PathLike], load_version: str
    ) -> Optional["StorageGateway"]:
        path = pathlib.Path(root) / "storage_gateway.yaml"

        try:
            with open(path, "r") as f:
                obj = yaml.load(f)
                if isinstance(obj, PreConnectorStorageGateway):
                    obj = PosixStorageGateway(obj.__getstate__())
                assert isinstance(obj, StorageGateway)
        except FileNotFoundError:
            return None

        return obj

    def identity_mapping_for_domains(
        self, domains: Iterable[str]
    ) -> IdentityMappingDict:
        domains_list = "|".join([d.replace(".", "\\.") for d in domains])
        return {
            "DATA_TYPE": "expression_identity_mapping#1.0.0",
            "mappings": [
                {
                    "source": "{username}",
                    "match": f"(.*)@({domains_list})",
                    "output": "{0}",
                }
            ],
        }

    def add_allowed_domains(
        self,
        domains: Iterable[str],
        identity_mappings: Optional[List[IdentityMappingDict]] = None,
    ) -> None:
        if self.allowed_domains is None:
            self.allowed_domains = []
        if self.identity_mappings is None:
            self.identity_mappings = []

        for domain in domains:
            if domain not in self.allowed_domains:
                self.allowed_domains.append(domain)

        if identity_mappings is None:
            identity_mappings = [self.identity_mapping_for_domains(domains)]

        for identity_mapping in identity_mappings:
            if identity_mapping not in self.identity_mappings:
                self.identity_mappings.append(identity_mapping)

    def add_identity_mappings(self, username_to_local: List[Tuple[str, str]]) -> None:
        if self.identity_mappings is None:
            self.identity_mappings = []
        self.identity_mappings.insert(
            0,
            {
                "DATA_TYPE": "expression_identity_mapping#1.0.0",
                "mappings": [
                    {
                        "source": "{username}",
                        "match": identity_username,
                        "literal": True,
                        "output": username,
                    }
                    for identity_username, username in username_to_local
                ],
            },
        )

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, StorageGateway):
            return NotImplemented
        return bool(self.__getstate__() == other.__getstate__())

    def __getstate__(self) -> T:
        return cast(
            T,
            {
                "display_name": self.display_name,
                "users_allow": self.users_allow,
                "users_deny": self.users_deny,
                "load_dsi_module": self.load_dsi_module,
                "allowed_domains": self.allowed_domains,
                "identity_mappings": self.identity_mappings,
                "restrict_paths": self.restrict_paths,
            },
        )

    def __setstate__(self, state: T) -> None:
        base_state = cast(StorageGatewayYamlDict, state)
        self.gridftp_config = None
        self.display_name = base_state["display_name"]
        self.users_allow = base_state["users_allow"]
        self.users_deny = base_state["users_deny"]
        self.load_dsi_module = base_state["load_dsi_module"]
        self.allowed_domains = base_state["allowed_domains"]
        self.identity_mappings = base_state["identity_mappings"]
        self.restrict_paths = base_state["restrict_paths"]

    def to_json(self) -> StorageGatewayDict:
        storage_gateway: StorageGatewayDict = {
            "DATA_TYPE": "storage_gateway#1.1.0",
            "high_assurance": False,
        }

        if self.display_name:
            storage_gateway["display_name"] = self.display_name

        if self.allowed_domains:
            storage_gateway["allowed_domains"] = self.allowed_domains

        if self.identity_mappings:
            storage_gateway["identity_mappings"] = self.identity_mappings

        if self.users_allow:
            storage_gateway["users_allow"] = self.users_allow

        if self.users_deny:
            storage_gateway["users_deny"] = self.users_deny

        if self.restrict_paths:
            storage_gateway["restrict_paths"] = self.restrict_paths

        return storage_gateway

    def update(self, other: "StorageGateway[T]") -> None:
        for attr in [
            "display_name",
            "users_allow",
            "users_deny",
            "restrict_paths",
        ]:
            old_value = getattr(self, attr)
            new_value = getattr(other, attr)
            if old_value != new_value:
                setattr(self, attr, new_value)

        unsupported_updates = []
        for attr in ["load_dsi_module", "identity_mappings", "allowed_domains"]:
            old_value = getattr(self, attr)
            new_value = getattr(other, attr)
            if new_value and old_value != new_value:
                unsupported_updates.append(
                    f"Unable to update {attr} from {old_value} to {new_value}"
                )
        if unsupported_updates:
            raise UnsupportedUpdate("\n".join(unsupported_updates))


@yaml_object(yaml)
class BlackPearlStorageGateway(StorageGateway[BlackPearlStorageGatewayYamlDict]):
    def __init__(self, data: Optional[BlackPearlStorageGatewayYamlDict] = None) -> None:
        super().__init__()
        self.s3_endpoint: Optional[str] = None
        self.bp_access_id_file: Optional[str] = None
        self._blackpearl_config: Optional[BlackPearlConfig] = None
        self._blackpearl_access_id_file_parser: Optional[
            BlackPearlAccessIdFileParser
        ] = None

        if data:
            self.__setstate__(data)

    @property
    def blackpearl_config(self) -> Optional[BlackPearlConfig]:
        return self._blackpearl_config

    @blackpearl_config.setter
    def blackpearl_config(self, value: Optional[BlackPearlConfig]) -> None:
        self._blackpearl_config = value
        if value:
            self.s3_endpoint = value.endpoint
            self.bp_access_id_file = value.access_id_file

    @property
    def blackpearl_access_id_file_parser(
        self,
    ) -> Optional[BlackPearlAccessIdFileParser]:
        if not self._blackpearl_access_id_file_parser and self.bp_access_id_file:
            self._blackpearl_access_id_file_parser = BlackPearlAccessIdFileParser(
                self.bp_access_id_file
            )
        return self._blackpearl_access_id_file_parser

    def __getstate__(self) -> BlackPearlStorageGatewayYamlDict:
        res = cast(BlackPearlStorageGatewayYamlDict, super().__getstate__())
        res["s3_endpoint"] = self.s3_endpoint
        res["bp_access_id_file"] = self.bp_access_id_file

        return res

    def __setstate__(self, state: BlackPearlStorageGatewayYamlDict) -> None:
        super().__setstate__(state)
        self.s3_endpoint = state["s3_endpoint"]
        self.bp_access_id_file = state["bp_access_id_file"]

    def to_json(self) -> BlackPearlStorageGatewayDict:
        doc = cast(BlackPearlStorageGatewayDict, super().to_json())
        doc["connector_id"] = BLACKPEARL_CONNECTOR_ID

        policies: BlackPearlStorageGatewayPoliciesDict = {
            "DATA_TYPE": "blackpearl_storage_policies#1.0.0"
        }

        if self.s3_endpoint:
            policies["s3_endpoint"] = self.s3_endpoint

        if self.bp_access_id_file:
            policies["bp_access_id_file"] = self.bp_access_id_file

        doc["policies"] = policies
        return doc

    def update(self, other: "StorageGateway[BlackPearlStorageGatewayYamlDict]") -> None:
        super().update(other)
        for attr in [
            "s3_endpoint",
            "bp_access_id_file",
        ]:
            old_value = getattr(self, attr)
            new_value = getattr(other, attr)
            if old_value != new_value:
                setattr(self, attr, new_value)

    def __str__(self) -> str:
        if self.display_name:
            return f"<BlackPearlStorageGateway {self.display_name}>"
        else:
            return "<BlackPearlStorageGateway>"


@yaml_object(yaml)
class HPSSStorageGateway(StorageGateway[HPSSStorageGatewayYamlDict]):
    def __init__(self, data: Optional[HPSSStorageGatewayYamlDict] = None) -> None:
        super().__init__()
        self.login_name: Optional[str] = None
        self.authentication_mech: Optional[str] = None
        self.authenticator: Optional[str] = None
        self.uda_checksum: Optional[bool] = None
        self._hpss_config: Optional[HPSSConfig] = None

        if data:
            self.__setstate__(data)

    @property
    def hpss_config(self) -> Optional[HPSSConfig]:
        return self._hpss_config

    @hpss_config.setter
    def hpss_config(self, value: Optional[HPSSConfig]) -> None:
        self._hpss_config = value
        if value:
            self.login_name = value.login_name
            self.authentication_mech = value.authentication_mech
            self.authenticator = value.authenticator
            self.uda_checksum = value.uda_checksum_support == "on"

    def __getstate__(self) -> HPSSStorageGatewayYamlDict:
        res = cast(HPSSStorageGatewayYamlDict, super().__getstate__())
        res["login_name"] = self.login_name
        res["authentication_mech"] = self.authentication_mech
        res["authenticator"] = self.authenticator
        res["uda_checksum"] = self.uda_checksum

        return res

    def __setstate__(self, state: HPSSStorageGatewayYamlDict) -> None:
        super().__setstate__(state)
        self.login_name = state["login_name"]
        self.authentication_mech = state["authentication_mech"]
        self.authenticator = state["authenticator"]
        self.uda_checksum = state["uda_checksum"]

    def to_json(self) -> HPSSStorageGatewayDict:
        doc = cast(HPSSStorageGatewayDict, super().to_json())
        doc["connector_id"] = HPSS_CONNECTOR_ID

        policies: HPSSStorageGatewayPoliciesDict = {
            "DATA_TYPE": "hpss_storage_policies#1.1.0",
            "authentication_mech": self.authentication_mech,
            "authenticator": self.authenticator,
            "login_name": self.login_name,
            "uda_checksum": self.uda_checksum,
        }

        doc["policies"] = policies
        return doc

    def update(self, other: "StorageGateway[HPSSStorageGatewayYamlDict]") -> None:
        super().update(other)
        for attr in [
            "login_name",
            "authentication_mech",
            "authenticator",
            "uda_checksum",
        ]:
            old_value = getattr(self, attr)
            new_value = getattr(other, attr)
            if old_value != new_value:
                setattr(self, attr, new_value)

    def __str__(self) -> str:
        if self.display_name:
            return f"<HPSSStorageGateway {self.display_name}>"
        else:
            return "<HPSSStorageGateway>"


@yaml_object(yaml)
class PosixStorageGateway(StorageGateway[PosixStorageGatewayYamlDict]):
    def __init__(self, data: Optional[PosixStorageGatewayYamlDict] = None) -> None:
        super().__init__()

        self.groups_allow: Optional[List[str]] = None
        self.groups_deny: Optional[List[str]] = None

        if data:
            self.__setstate__(data)

    def __getstate__(self) -> PosixStorageGatewayYamlDict:
        res = cast(PosixStorageGatewayYamlDict, super().__getstate__())
        res["groups_allow"] = self.groups_allow
        res["groups_deny"] = self.groups_deny

        return res

    def __setstate__(self, state: PosixStorageGatewayYamlDict) -> None:
        super().__setstate__(state)
        self.groups_allow = state["groups_allow"]
        self.groups_deny = state["groups_deny"]

    def to_json(self) -> PosixStorageGatewayDict:
        doc = cast(PosixStorageGatewayDict, super().to_json())
        doc["connector_id"] = POSIX_CONNECTOR_ID

        policies: PosixStorageGatewayPoliciesDict = {
            "DATA_TYPE": "posix_storage_policies#1.0.0"
        }

        if self.groups_allow:
            policies["groups_allow"] = self.groups_allow

        if self.groups_deny:
            policies["groups_deny"] = self.groups_deny

        doc["policies"] = policies
        return doc

    def update(self, other: "StorageGateway[PosixStorageGatewayYamlDict]") -> None:
        super().update(other)
        for attr in ["groups_allow", "groups_deny"]:
            old_value = getattr(self, attr)
            new_value = getattr(other, attr)
            if old_value != new_value:
                setattr(self, attr, new_value)

    def __str__(self) -> str:
        if self.display_name:
            return f"<PosixStorageGateway {self.display_name}>"
        else:
            return "<PosixStorageGateway>"


@yaml_object(yaml)
class S3StorageGateway(StorageGateway[S3StorageGatewayYamlDict]):
    s3_user_credential_required: Optional[bool] = True

    def __init__(self, data: Optional[S3StorageGatewayYamlDict] = None) -> None:
        super().__init__()
        self.s3_endpoint: Optional[str] = None
        self.s3_buckets: Optional[List[str]] = None
        self.s3_user_credential_required: Optional[bool] = True
        self.s3_map_filename: Optional[str] = None
        self._s3_config: Optional[S3Config] = None
        self._s3_mapfile_parser: Optional[S3MapfileParser] = None

        if data:
            self.__setstate__(data)

    def identity_mapping_for_domains(
        self, domains: Iterable[str]
    ) -> IdentityMappingDict:
        domains_list = "|".join([d.replace(".", "\\.") for d in domains])
        return {
            "DATA_TYPE": "expression_identity_mapping#1.0.0",
            "mappings": [
                {
                    "source": "{username}",
                    "match": f"(.*@({domains_list}))",
                    "output": "{0}",
                }
            ],
        }

    @property
    def s3_config(self) -> Optional[S3Config]:
        return self._s3_config

    @s3_config.setter
    def s3_config(self, value: Optional[S3Config]) -> None:
        self._s3_config = value
        if value:
            if value.use_https:
                self.s3_endpoint = f"https://{value.host_name}"
            else:
                self.s3_endpoint = f"http://{value.host_name}"
            self.s3_map_filename = value.s3_map_filename

    @property
    def s3_mapfile_parser(self) -> Optional[S3MapfileParser]:
        if not self._s3_mapfile_parser and self.s3_map_filename:
            process_user: Optional[str] = None
            if self.gridftp_config:
                process_user = self.gridftp_config.get_string_option("process_user")

            self._s3_mapfile_parser = S3MapfileParser(
                self.s3_map_filename, process_user=process_user
            )
        return self._s3_mapfile_parser

    def __getstate__(self) -> S3StorageGatewayYamlDict:
        res = cast(S3StorageGatewayYamlDict, super().__getstate__())
        res["s3_endpoint"] = self.s3_endpoint
        res["s3_buckets"] = self.s3_buckets
        res["s3_user_credential_required"] = self.s3_user_credential_required
        res["s3_map_filename"] = self.s3_map_filename

        return res

    def __setstate__(self, state: S3StorageGatewayYamlDict) -> None:
        super().__setstate__(state)
        self._s3_config = None
        self._s3_mapfile_parser = None
        self.s3_endpoint = state["s3_endpoint"]
        self.s3_buckets = state["s3_buckets"]
        self.s3_user_credential_required = state["s3_user_credential_required"]
        self.s3_map_filename = state["s3_map_filename"]

    def to_json(self) -> S3StorageGatewayDict:
        doc = cast(S3StorageGatewayDict, super().to_json())
        doc["connector_id"] = S3_CONNECTOR_ID

        policies: S3StorageGatewayPoliciesDict = {
            "DATA_TYPE": "s3_storage_policies#1.0.0"
        }

        if self.s3_endpoint:
            policies["s3_endpoint"] = self.s3_endpoint

        if self.s3_buckets:
            policies["s3_buckets"] = self.s3_buckets

        if self.s3_user_credential_required:
            policies["s3_user_credential_required"] = self.s3_user_credential_required

        doc["policies"] = policies
        return doc

    def update(self, other: "StorageGateway[S3StorageGatewayYamlDict]") -> None:
        super().update(other)
        for attr in [
            "s3_endpoint",
            "s3_buckets",
            "s3_user_credential_required",
            "s3_map_filename",
        ]:
            old_value = getattr(self, attr)
            new_value = getattr(other, attr)
            if old_value != new_value:
                setattr(self, attr, new_value)

    def __str__(self) -> str:
        if self.display_name:
            return f"<S3StorageGateway {self.display_name}>"
        else:
            return "<S3StorageGateway>"


@yaml_object(yaml)
class CephStorageGateway(StorageGateway[CephStorageGatewayYamlDict]):
    def __init__(self, data: Optional[CephStorageGatewayYamlDict] = None) -> None:
        super().__init__()
        self.s3_endpoint: Optional[str] = None
        self.ceph_admin_bucket: Optional[str] = None
        self.ceph_admin_key_id: Optional[str] = None
        self.ceph_admin_secret_key: Optional[str] = None
        self._ceph_config: Optional[CephConfig] = None

        if data:
            self.__setstate__(data)

    def dump(self, root: Union[str, os.PathLike], omit_secrets: bool = False) -> None:
        if omit_secrets:
            without_secrets = CephStorageGateway(data=self.__getstate__())
            without_secrets.ceph_admin_secret_key = None
            without_secrets.dump(root, omit_secrets=False)
        else:
            super().dump(root, False)

    @property
    def ceph_config(self) -> Optional[CephConfig]:
        return self._ceph_config

    @ceph_config.setter
    def ceph_config(self, value: Optional[CephConfig]) -> None:
        self._ceph_config = value
        if value:
            if value.use_https:
                self.s3_endpoint = f"https://{value.host_name}"
            else:
                self.s3_endpoint = f"http://{value.host_name}"
            if value.ceph_rgw_admin_bucket:
                self.ceph_admin_bucket = value.ceph_rgw_admin_bucket
            if value.ceph_rgw_admin_access_key_id:
                self.ceph_admin_key_id = value.ceph_rgw_admin_access_key_id
            if value.ceph_rgw_admin_secret_access_key:
                self.ceph_admin_secret_key = value.ceph_rgw_admin_secret_access_key

    def __getstate__(self) -> CephStorageGatewayYamlDict:
        res = cast(CephStorageGatewayYamlDict, super().__getstate__())
        res["s3_endpoint"] = self.s3_endpoint
        res["ceph_admin_bucket"] = self.ceph_admin_bucket
        res["ceph_admin_key_id"] = self.ceph_admin_key_id
        res["ceph_admin_secret_key"] = self.ceph_admin_secret_key

        return res

    def __setstate__(self, state: CephStorageGatewayYamlDict) -> None:
        super().__setstate__(state)
        self.s3_endpoint = state["s3_endpoint"]
        self.ceph_admin_bucket = state["ceph_admin_bucket"]
        self.ceph_admin_key_id = state["ceph_admin_key_id"]
        self.ceph_admin_secret_key = state["ceph_admin_secret_key"]

    def to_json(self) -> CephStorageGatewayDict:
        doc = cast(CephStorageGatewayDict, super().to_json())
        doc["connector_id"] = CEPH_CONNECTOR_ID

        policies: CephStorageGatewayPoliciesDict = {
            "DATA_TYPE": "ceph_storage_policies#1.0.0"
        }

        if self.s3_endpoint:
            policies["s3_endpoint"] = self.s3_endpoint

        if self.ceph_admin_key_id:
            policies["ceph_admin_key_id"] = self.ceph_admin_key_id

        if self.ceph_admin_secret_key:
            policies["ceph_admin_secret_key"] = self.ceph_admin_secret_key

        doc["policies"] = policies
        return doc

    def update(self, other: "StorageGateway[CephStorageGatewayYamlDict]") -> None:
        super().update(other)
        for attr in [
            "s3_endpoint",
            "ceph_admin_bucket",
            "ceph_admin_key_id",
            "ceph_admin_secret_key",
        ]:
            old_value = getattr(self, attr)
            new_value = getattr(other, attr)
            if old_value != new_value:
                setattr(self, attr, new_value)

    def __str__(self) -> str:
        if self.display_name:
            return f"<CephStorageGateway {self.display_name}>"
        else:
            return "<CephStorageGateway>"


# Backward compatible storage gateway yaml object for deserializing
# storage_gateway.yaml from migration plans that predate multiple connector
# support
@yaml_object(yaml)
class PreConnectorStorageGateway(PosixStorageGateway):
    yaml_tag = "!StorageGateway"
