from typing import Dict, List, Optional
from urllib.parse import quote

from lxml import etree
import requests
from requests.exceptions import RequestException


class InCommonMetadata:
    _idplist_xml: Optional[etree._Element] = None
    _entity_id_cache: Dict[str, str] = {}
    _domain_cache: Dict[str, List[str]] = {}

    def __init__(
        self,
        organization: Optional[str] = None,
        entity_id: Optional[str] = None,
        domains: Optional[List[str]] = None,
    ):
        self._organization: Optional[str]
        self._entity_id: Optional[str]
        self._domains: Optional[List[str]]

        self.domains = domains
        self.entity_id = entity_id
        self.organization = organization

    @classmethod
    def entity_id_for_organization(cls, organization: str) -> Optional[str]:
        entity_id = cls._entity_id_cache.get(organization)
        if not entity_id:
            idplist = cls.idplist_xml()
            if idplist is not None:
                for idp in idplist.findall("idp"):
                    org = idp.find("Organization_Name")
                    name = idp.find("Display_Name")
                    if (org is not None and org.text == organization) or (
                        name is not None and name.text == organization
                    ):
                        rands = idp.find("RandS")
                        if rands is not None and rands.text == "1":
                            valid = True
                        if valid:
                            entity_id_val = idp.attrib.get("entityID")
                            if isinstance(entity_id_val, bytes):
                                entity_id_val = entity_id_val.decode()
                            if entity_id_val:
                                cls._entity_id_cache[organization] = entity_id_val
                                return entity_id_val
                        break
        return entity_id

    @classmethod
    def domains_for_entity_id(cls, entity_id: str) -> Optional[List[str]]:
        domains = cls._domain_cache.get(entity_id)
        if not domains:
            try:
                r = requests.get(
                    f"https://mdq.incommon.org/entities/{quote(entity_id, safe=':')}"
                )
                data = r.content
            except RequestException:
                return None

            incommon_xml = etree.fromstring(data)
            domains_xpath = (
                "/md:EntityDescriptor[@entityID=$entity_id]//shibmd:Scope/text()"
            )
            XMLNS = {
                "md": "urn:oasis:names:tc:SAML:2.0:metadata",
                "shibmd": "urn:mace:shibboleth:metadata:1.0",
            }
            domains_val = etree.XPath(domains_xpath, namespaces=XMLNS)(
                incommon_xml, entity_id=entity_id
            )
            if domains_val is not None:
                assert not isinstance(domains_val, (bool, float))
                cls._domain_cache[entity_id] = list(
                    d for d in set(str(s) for s in domains_val)
                )
                return cls._domain_cache[entity_id]
        return domains

    @classmethod
    def idplist_xml(cls) -> Optional[etree._Element]:
        if cls._idplist_xml is None:
            IDPLIST_XML_URL = "https://cilogon.org/include/idplist.xml"
            try:
                r = requests.get(IDPLIST_XML_URL)
                data = r.content
                if data:
                    cls._idplist_xml = etree.fromstring(data)
            except RequestException:
                pass
        return cls._idplist_xml

    @property
    def entity_id(self) -> Optional[str]:
        return self._entity_id

    @entity_id.setter
    def entity_id(self, value: Optional[str]) -> None:
        self._entity_id = value
        if value and self.domains is None:
            self.domains = self.domains_for_entity_id(value)

    @property
    def domains(self) -> Optional[List[str]]:
        return self._domains

    @domains.setter
    def domains(self, value: Optional[List[str]]) -> None:
        self._domains = value

    @property
    def organization(self) -> Optional[str]:
        return self._organization

    @organization.setter
    def organization(self, value: Optional[str]) -> None:
        self._organization = value

        if value and self.entity_id is None:
            self.entity_id = self.entity_id_for_organization(value)
