from dataclasses import dataclass
import pwd
import typing as t


@dataclass
class S3MapEntry:
    username: str
    access_key_id: str
    secret_key: str
    token: t.Optional[str]

    def __eq__(self, other: t.Any) -> bool:
        if not isinstance(other, S3MapEntry):
            return NotImplemented
        return self is other or (
            self.username == other.username
            and self.access_key_id == other.access_key_id
            and self.secret_key == other.secret_key
            and self.token == other.token
        )


class S3MapfileParser:
    def __init__(
        self, path: t.Optional[str] = None, process_user: t.Optional[str] = None
    ) -> None:
        self.mapfile_entries: t.Dict[str, t.Optional[S3MapEntry]] = {}

        if path is None:
            path = "$HOME/.globus/s3"

        self.path = path
        self.per_user = "$HOME" in path or "$LOGNAME" in path

        if not self.per_user:
            self._parse_s3_mapfile_entries(self.path)
        elif process_user:
            self.per_user = False
            self._parse_s3_mapfile_entries(self._s3_user_mapfile(process_user))

    def _s3_user_mapfile(self, username: str) -> str:
        pwent = pwd.getpwnam(username)
        home = pwent.pw_dir
        logname = pwent.pw_name
        return self.path.replace("$HOME", home).replace("$LOGNAME", logname)

    def _parse_s3_mapfile_entries(
        self, path: str, username: t.Optional[str] = None
    ) -> None:
        with open(path, "r") as f:
            for line in f:
                entry = self._parse_s3_entry_line(line.strip())
                if entry:
                    if not username or entry.username == username:
                        self.mapfile_entries[entry.username] = entry

    @staticmethod
    def _parse_s3_entry_line(line: str) -> t.Optional[S3MapEntry]:
        if line:
            parts = line.split(";")
            token: t.Optional[str] = None
            if len(parts) == 4:
                token = parts.pop()
            if len(parts) != 3:
                raise IOError("Invalid S3 mapfile line")
            return S3MapEntry(parts[0], parts[1], parts[2], token)
        return None

    def __contains__(self, username: str) -> bool:
        if username not in self.mapfile_entries and self.per_user:
            try:
                self._parse_s3_mapfile_entries(
                    self._s3_user_mapfile(username), username=username
                )
            except (IOError, KeyError):
                self.mapfile_entries[username] = None
        return self.mapfile_entries.get(username) is not None

    def __getitem__(self, username: str) -> S3MapEntry:
        if username not in self.mapfile_entries and self.per_user:
            try:
                self._parse_s3_mapfile_entries(self._s3_user_mapfile(username))
            except IOError:
                self.mapfile_entries[username] = None
        res = self.mapfile_entries[username]
        if res is None:
            raise KeyError(username)
        return res
