import time
from typing import List, Mapping, Optional

from flask import current_app

from globus.manager.api.errors import AuthenticationTimeout

from .filter import Filter
from .identity_set import IdentitySet


class AuthenticationTimeoutFilter(Filter):
    """
    This class provides a callable Filter interface to implement
    high assurance- or last authentication-based identity filtering.
    """

    def __init__(
        self,
        require_session: bool = False,
        authentication_timeout: Optional[int] = None,
        require_mfa: bool = False,
    ):
        """
        :param require_session:
            Flag indicating whether the identity must have been
            authenticated within the Globus Auth session associated
            with this API call.
        :param authentication_timeout:
            Integer (in minutes) indicating the maximum time which has
            elapsed since authentication with a given identity to determine
            whether it can still be considered valid. If `require_session`
            is True, then this must be within the current session, if not,
            then any Globus Auth authentication by this identity is accepted.
            If this is None, then no time-based filtering is done by this
            filter.
        :param require_mfa:
            Flag indicating whether the identity must have been
            authenticated within the Globus Auth session using multi-factor
            authentication.
        """
        self.require_session = require_session
        self.authentication_timeout = authentication_timeout
        self.require_mfa = require_mfa

    def __call__(self, identity_set: IdentitySet, auth_state) -> IdentitySet:
        """
        Filter the identity_set_data, removing all items which do not
        meet this filter's authentication timeout requirements.

        :param identity_set:
            Identity set information, filtered in place
        :param auth_state:
            Authentication state for this request

        :returns:
            Returns the subset of identity_set_data which meets the
            require_session and authentication_timeout requirements of this
            filter.
        """
        invalid_identities: List[Mapping] = []

        current_app.logger.debug(
            "%s: require_session=%s authentication_timeout=%s require_mfa=%s input=%s",
            self,
            self.require_session,
            self.authentication_timeout,
            self.require_mfa,
            identity_set,
        )

        authentications: Mapping[str, int] = {}

        only_auth_clients = len(identity_set) == len(
            [
                i
                for i in identity_set
                if i["username"].endswith("@clients.auth.globus.org")
            ]
        )

        if not only_auth_clients:
            if self.require_session:
                authentications = auth_state.session_authentications
            else:
                authentications = auth_state.non_session_authentications

        if self.authentication_timeout:
            oldest_allowed = time.time() - self.authentication_timeout * 60

            invalid_identities = [
                i
                for i in identity_set
                if authentications.get(i["id"], 0) < oldest_allowed
                and not i["username"].endswith("@clients.auth.globus.org")
            ]
        else:
            invalid_identities = [
                i
                for i in identity_set
                if authentications.get(i["id"], None) is None
                and not i["username"].endswith("@clients.auth.globus.org")
            ]

        for i in invalid_identities:
            identity_set.remove(i)

        if self.require_mfa:
            invalid_identities.extend(
                i
                for i in identity_set
                if not auth_state.multi_factor_authentications.get(i["id"], False)
                and not i["username"].endswith("@clients.auth.globus.org")
            )

        for i in invalid_identities:
            if i in identity_set:
                identity_set.remove(i)

        if invalid_identities:
            detail = AuthenticationTimeout(
                identities=invalid_identities,
                high_assurance=self.require_session,
                require_mfa=self.require_mfa,
            )
            identity_set.error_details.insert(0, detail)

        current_app.logger.debug("%s: result=%s", self, identity_set)
        return identity_set

    def __repr__(self):
        return (
            "<AuthenticationTimeoutFilter "
            "session={} "
            "authentication_timeout={} "
            "mfa={}>".format(
                self.require_session, self.authentication_timeout, self.require_mfa
            )
        )
