from __future__ import annotations

import uuid
from typing import Dict, List, Optional, Sequence, TypeVar, Union, cast

from cachelib import SimpleCache
from flask import current_app, g

from globus_sdk import AccessTokenAuthorizer, GroupsClient
from globus_sdk.scopes import GroupsScopes

import globus.manager.api.errors as errors
from globus.manager.api.authentication import AuthenticationState
from globus.manager.api.models import Collection, Role
from globus.manager.constants import PrincipalType, RoleType

from .identity_in_identity_set_filter import IdentityInIdentitySetFilter
from .identity_set import IdentitySet

# Dependent token => Groups API response
groups_cache = SimpleCache(default_timeout=30)


class RoleSpec:
    """
    This class parses the role specification from the RoleAuthorizer
    specification in the API documentation
    """

    def __init__(
        self: RoleSpec,
        role_string: str,
        collection: Optional[Collection] = None,
    ):
        assert role_string.startswith("collection:") or role_string.startswith(
            "endpoint:"
        )

        resource: Optional[str] = None
        self.mapped_collection_role = False
        self.original_collection = collection

        if role_string.startswith("collection:"):
            role, resource = role_string.rsplit(":", 1)
            assert resource in ["*", "{collection_id}", "{mapped_collection_id}"]
            if resource == "*":
                resource = None
            else:
                assert collection is not None
                if resource == "{mapped_collection_id}":
                    assert collection.mapped_collection is not None
                    collection = cast(Collection, collection.mapped_collection)
                    self.mapped_collection_role = True
                resource = collection.id
        else:
            role = role_string

        if resource is not None:
            assert uuid.UUID(resource) is not None

        self.role = RoleType(role)
        self.role_name = self.role.name.split("_", 1)[1]
        self.resource = resource
        self.collection = collection

    def __repr__(self):
        if (
            self.role == RoleType.endpoint_owner
            or self.role == RoleType.endpoint_administrator
            or self.role == RoleType.endpoint_activity_manager
            or self.role == RoleType.endpoint_activity_monitor
        ):
            return "<RoleSpec {}>".format(self.role.value)
        else:
            if self.resource and self.mapped_collection_role:
                return f"<RoleSpec {self.role.value}:{{mapped_collection_id}}>"
            elif self.resource:
                return f"<RoleSpec {self.role.value}:{{collection_id}}>"
            else:
                return "<RoleSpec {}:*>".format(self.role.value)

    def __str__(self):
        if (
            self.role == RoleType.endpoint_owner
            or self.role == RoleType.endpoint_administrator
            or self.role == RoleType.endpoint_activity_manager
            or self.role == RoleType.endpoint_activity_monitor
        ):
            return self.role.value
        else:
            if self.resource:
                return "{}:{}".format(self.role.value, self.collection)
            else:
                return "{}:*".format(self.role.value)


RoleSpecs = TypeVar("RoleSpecs", str, List[str], RoleSpec, List[RoleSpec])


class RoleAuthorizer:
    """
    This class provides methods to process role-based access control
    policies for endpoints or collections. Note that if there is no
    subscription on the endpoint, only the `endpoint:owner` role is
    considered.

    """

    def __init__(
        self: RoleAuthorizer,
        role_specs: Union[str, RoleSpec, Sequence[Union[str, RoleSpec]]],
        caching: bool = False,  # Controlled rollout
    ):
        """
        Initialize a RoleAuthorizer object.

        :param role_specs:
            The roles to authorize. This input may be in any of these forms:
            -   String version of one or more role names such as
                `endpoint:administrator` separated by the `,` character
            -   List of strings of role names such as `endpoint:administrator`
            -   A RoleSpec object
            -   A List of RoleSpec objects
        """
        self.caching = caching

        rs: List[Union[str, RoleSpec]]
        if isinstance(role_specs, str):
            rs = cast(List[Union[str, RoleSpec]], role_specs.split(","))
        elif isinstance(role_specs, RoleSpec):
            rs = [role_specs]
        else:
            rs = cast(List[Union[str, RoleSpec]], role_specs)

        self.role_specs: List[RoleSpec] = []
        for r in rs:
            if isinstance(r, RoleSpec):
                role_spec = r
            else:
                role_spec = RoleSpec(r)
            self.role_specs.append(role_spec)

    def authorize_identity(
        self, identity_id: str, auth_state: AuthenticationState
    ) -> Dict:
        """
        Attempt to authorize the caller as identity_id (a string UUID)
        using the authorization information in the current request's
        authentication state.

        :param identity_id:
            A string representation of the uuid to authorize.
        :param auth_state:
            The caller's authentication state. This is passed in rather
            than based on `g` to handle the gridftp case where the
            authentication state is created based on assertions from transfer.

        :returns:
            This function returns an entry from the auth_state's
            identity_set_data property that contains the information
            about the identity which was authorized, with two additional
            entries:

            -   _role_name: A string containing the name of the role
                the caller has, such as `endpoint:administrator`
            -   _role_id: A string representation of the UUID of
                this role.

        :raises errors.IdNotInIdentitySetError:
            The `auth_state` does not have any identity which matches
            `identity_id`.
        :raises errors.MissingRequiredRoleError:
            The `auth_state` has the given identity, but there is no
            role assignment for that identity.
        """
        if self.caching:
            if not hasattr(g, "role_cache"):
                g.role_cache = {}
            role_cache = g.role_cache

        idset = IdentitySet(identity_set=auth_state.identity_set_data)

        current_app.logger.debug(
            "Filtering identity_set %s with %s", idset, identity_id
        )

        # Filter out identities which don't match the requested id first
        f = IdentityInIdentitySetFilter(identity_id)
        f(idset, auth_state)

        if not idset:
            raise errors.IdNotInIdentitySetError(detail=idset.error_details[0])

        if current_app.subscription:
            enforced_roles = [
                spec
                for spec in self.role_specs
                if not (spec.collection and spec.collection.is_deleted)
                and not (
                    spec.original_collection and spec.original_collection.is_deleted
                )
            ]

        else:
            enforced_roles = [
                r for r in self.role_specs if r.role == RoleType.endpoint_owner
            ]

        identity = idset[0]

        # See if we previously cached roles for this identity
        roles: list[dict[str, str | None]] | None = None
        if self.caching:
            roles = role_cache.get(identity["id"])

        if roles is None:
            # Not cached so query it and cache it for next time
            roles = [
                {"id": r.id, "role": r.role, "resource": r.resource}
                for r in Role.query.filter(
                    ~Role.is_deleted,  # type: ignore
                    Role.principal_type == PrincipalType.identity.value,
                    Role.identity_id == identity["id"],
                )
            ]
            if self.caching:
                role_cache[identity["id"]] = roles

        for role in roles or []:
            for spec in enforced_roles:
                if role["role"] == spec.role and (
                    spec.resource is None or role["resource"] == spec.resource
                ):
                    res_identity = identity.copy()
                    res_identity["_role_name"] = spec.role.value
                    res_identity["_role_id"] = role["id"]
                    return res_identity

        # No luck with identity roles. Let's search roles owned by groups.

        group_role_count: int | None = None
        if self.caching:
            group_role_count = getattr(g, "group_role_count", None)

        if group_role_count is None:
            group_role_count = Role.query.filter(
                ~Role.is_deleted,  # type: ignore
                Role.principal_type == PrincipalType.group,
            ).count()
            if self.caching:
                g.group_role_count = group_role_count

        if group_role_count > 0:
            my_group_ids: list[str] = []
            try:
                # Look up the caller's groups response. Cached on first attempt
                groups_response = None
                groups_token = auth_state.dependent_token(
                    required_scope=GroupsScopes.view_my_groups_and_memberships,
                )
                if groups_token:
                    groups_response = groups_cache.get(groups_token)

                if not groups_response:
                    groups_client = GroupsClient(
                        authorizer=AccessTokenAuthorizer(groups_token),
                        environment=current_app.environment,
                    )
                    groups_response = groups_client.get_my_groups().data
                    groups_cache.set(groups_token, groups_response)

                # Construct the list of group IDs for this caller
                for group in groups_response:
                    for membership in group["my_memberships"]:
                        if identity["id"] == membership["identity_id"]:
                            my_group_ids.append(group["id"])
            except Exception as e:
                current_app.logger.warning(
                    "Group resolution failure: %s %s",
                    groups_response,
                    e,
                )

            for group_id in my_group_ids:
                # See if this group's roles were cached a the previous run
                roles = None
                if self.caching:
                    roles = role_cache.get(group_id)

                if roles is None:
                    # Not cached so query it and cache it for next time
                    roles = [
                        {"id": r.id, "role": r.role, "resource": r.resource}
                        for r in Role.query.filter(
                            ~Role.is_deleted,  # type: ignore
                            Role.principal_type == PrincipalType.group,
                            Role.identity_id == group_id,
                        )
                    ]

                    if self.caching:
                        role_cache[group_id] = roles

                for role in roles or []:
                    for spec in enforced_roles:
                        if role["role"] == spec.role and (
                            spec.resource is None or role["resource"] == spec.resource
                        ):
                            res_identity = identity.copy()
                            res_identity["_role_name"] = spec.role.value
                            res_identity["_role_id"] = role["id"]
                            return res_identity

        raise errors.MissingRequiredRoleError(roles=enforced_roles)

    def authorize_identities(self, auth_state: AuthenticationState) -> IdentitySet:
        """
        Attempt to authorize the caller based on the roles associated
        with this authorizer.

        :param auth_state:
            The caller's authentication state. This is passed in rather
            than based on `g` to handle the gridftp case where the
            authentication state is created based on assertions from transfer.

        :returns:
            This function returns an IdentitySet containing the list of
            the caller's identities which are authorized by this
            role authorizer. This IdentitySet may be empty if no role
            is assigned for this identity_set. In that case
            IdentitySet.error_details may include information about why
            it is empty.
        """
        identities = IdentitySet()
        details = []
        for i in auth_state.identity_set_data:
            try:
                identity = self.authorize_identity(i["id"], auth_state)
                identities.append(identity)
            except errors.APIBaseError as e:
                current_app.logger.debug("Exception (%s) %s", type(e), e)
                if e.detail:
                    details.append(e.detail)
        if not identities:
            identities.error_details.extend(details)

        return identities

    def __repr__(self):
        return "<RoleAuthorizer {}>".format(self.role_specs)

    def __str__(self):
        return "role_authorizer allowed_roles={}".format(
            ",".join(str(s) for s in self.role_specs)
        )


EndpointRoleAuthorizer = RoleAuthorizer(
    [
        "endpoint:owner",
        "endpoint:administrator",
        "endpoint:activity_manager",
        "endpoint:activity_monitor",
    ]
)
