from __future__ import annotations

import datetime
import enum
import re
import uuid
from dataclasses import dataclass
from typing import (
    Any,
    Callable,
    Dict,
    Generator,
    Hashable,
    Iterable,
    List,
    Mapping,
    Optional,
    Set,
    Type,
    Union,
)

import marshmallow as ma
from flask import (
    Response,
    current_app,
    g,
    jsonify,
    request,
    url_for,
)

import globus.manager.api.errors as errors
import globus.manager.api.models as m
import globus.manager.constants as c
from globus.manager.api.authentication.authstate import AuthenticationState
from globus.manager.api.authorization import (
    AuthenticationTimeoutFilter,
    EndpointRoleAuthorizer,
    IdentitySet,
    PublicAuthorizer,
    RoleAuthorizer,
    RoleSpec,
    StorageGatewayAuthorizer,
)
from globus.manager.api.authorization.decorators import (
    requires_data_access_scope,
    requires_storage_gateway_policy,
    requires_subscription,
)
from globus.manager.api.blueprints.decorators import (
    Base62Marker,
    ConverterWithMetadata,
    Paginator,
    authenticated,
    document_authorization,
    enum_list_converter,
    expects_json,
    paginated,
    query_parameter,
    requires_one_role,
    result_data,
    returns_error,
)
from globus.manager.api.blueprints.utils import (
    check_subscription,
    create_object,
    delete_object,
    patch_object,
    put_object,
    update_object,
    validate_unique_custom_domain,
)
from globus.manager.api.models.result import Result
from globus.manager.api.schemas.api import data_type_schema as data_types
from globus.manager.constants import RoleType
from globus.manager.database import db

from .openapi import (
    OpenAPIBlueprint,
    PublicAuthorizer as OpenAPIPublicAuthorizer,
    RoleAuthorizer as OpenAPIRoleAuthorizer,
    StorageGatewayAuthorizer as OpenAPIStorageGatewayAuthorizer,
)

blueprint = OpenAPIBlueprint(
    "collections",
    __name__,
    tag="Collections",
    description="""
    The collection management API allows globus users to manage guest
    collections and administrators to manage mapped collections.

    A mapped collection is a collection that maps the Globus Auth
    identity of the user accessing the collection to a local
    account in the Storage Gateway's user space. Mapped collections
    can only be created by those with an `administrator` role on
    the Endpoint.

    A guest collection is a collection that uses the collection
    creator's credentials to access the Storage Gateway data
    interface. ACLs may be added to a Guest Collection by using the
    Globus Transfer API.

    Operations on Mapped Collections require an `administrator` role.

    """,
    menu_weight=7,
    url_prefix="/api/",
)


class CollectionListFilterName(enum.Enum):
    mapped_collections = "mapped_collections"
    guest_collections = "guest_collections"
    managed_by_me = "managed_by_me"
    created_by_me = "created_by_me"
    created_at = "created_at"
    last_access = "last_access"


@dataclass
class CollectionListFilter:
    filter_name: CollectionListFilterName
    operator: Optional[str] = None
    operand: Optional[datetime.date] = None

    def __eq__(self, other: Any) -> bool:
        if isinstance(other, CollectionListFilter):
            return (
                self.filter_name == other.filter_name
                and self.operator == other.operator
                and self.operand == other.operand
            )
        elif isinstance(other, CollectionListFilterName):
            return (
                self.filter_name == other
                and self.operator is None
                and self.operand is None
            )
        elif isinstance(other, str):
            filter_name = [n.name for n in CollectionListFilterName if n.name == other]
            if filter_name:
                return (
                    self.filter_name == other
                    and self.operator is None
                    and self.operand is None
                )
        return False


def collection_list_filter_converter() -> Type[ConverterWithMetadata]:
    class _CollectionListFilterConverter(ConverterWithMetadata):
        pattern = (
            r"^\s*("
            + "|".join(
                [
                    "mapped_collections",
                    "guest_collections",
                    "managed_by_me",
                    "created_by_me",
                    r"created_at\s*([<>]=?|=)\s*(\d{4}-\d{2}-\d{2})",
                    r"last_access\s*([<>]=?|=)\s*(\d{4}-\d{2}-\d{2})",
                ]
            )
            + r")\s*$"
        )

        metadata = {
            "style": "form",
            "explode": False,
            "schema": {
                "type": "array",
                "items": {
                    "type": "string",
                    "pattern": pattern,
                },
            },
        }

        def to_python(self, value: str) -> List[CollectionListFilter]:
            if value:
                res: List[CollectionListFilter] = []
                bad_items: List[str] = []
                for item in value.split(","):
                    m = re.match(self.pattern, item)
                    if not m:
                        bad_items.append(item)
                    elif m.group(1).startswith("created_at"):
                        operator = m.group(2)
                        operand = datetime.date(
                            *[int(x) for x in m.group(3).split("-")]
                        )
                        res.append(
                            CollectionListFilter(
                                CollectionListFilterName("created_at"),
                                operator,
                                operand,
                            )
                        )
                    elif m.group(1).startswith("last_access"):
                        operator = m.group(4)
                        operand = datetime.date(
                            *[int(x) for x in m.group(5).split("-")]
                        )
                        res.append(
                            CollectionListFilter(
                                CollectionListFilterName("last_access"),
                                operator,
                                operand,
                            )
                        )
                    else:
                        res.append(
                            CollectionListFilter(CollectionListFilterName(m.group(1)))
                        )

            if bad_items:
                raise errors.BadRequestError(
                    f"Unsupported filter(s): {','.join(bad_items)}"
                )
            return res

        def to_url(self, value: Any) -> str:
            return str(value)

    return _CollectionListFilterConverter


class CollectionGetInclude(enum.Enum):
    private_policies = "private_policies"


class _AccessChecker:
    """
    Verify caller's access rights to this endpoint and its objects using cached
    responses to minimize overhead and cleanup the call sites.
    """

    def __init__(
        self: _AccessChecker,
        auth_state: AuthenticationState,
    ) -> None:
        self._auth_state = auth_state

    #
    # Endpoint operations
    #

    def is_endpoint_public(self) -> bool:
        key = "__is_endpoint_public"
        if not hasattr(self, key):
            setattr(self, key, m.Endpoint.query.one().public)
        return getattr(self, key)

    def has_endpoint_role(self) -> bool:
        key = "__has_endpoint_role"
        if not hasattr(self, key):
            authorizer = RoleAuthorizer(
                [
                    RoleSpec("endpoint:owner"),
                    RoleSpec("endpoint:administrator"),
                    RoleSpec("endpoint:activity_manager"),
                    RoleSpec("endpoint:activity_monitor"),
                ],
                caching=True,
            )
            idset = authorizer.authorize_identities(self._auth_state)
            setattr(self, key, idset)
        return getattr(self, key)

    def endpoint_admin_identities(self) -> IdentitySet:
        """
        Return IdentitySet containing caller's identities that have the endpoint
        admin role or are endpoint owner.
        """
        key = "__endpoint_admin_identities"
        if not hasattr(self, key):
            authorizer = RoleAuthorizer(
                [
                    RoleSpec("endpoint:owner"),
                    RoleSpec("endpoint:administrator"),
                ],
                caching=True,
            )
            idset = authorizer.authorize_identities(self._auth_state)
            setattr(self, key, idset)
        return getattr(self, key)

    def is_endpoint_admin(self) -> bool:
        """
        Caller is endpoint admin or endpoint owner.
        """
        # No reason to cache result
        return bool(self.endpoint_admin_identities())

    #
    # Gateway operations
    #
    def in_gateway_domains(self, gateway: Optional[m.StorageGateway] = None) -> bool:
        """
        Check if the caller is authorized (in domain, ignore timeout) of the gateway. If
        gateway is not provided, check if the caller is authorized by any gateway on the
        endpoint.
        """

        def _in_gateway_domains(
            gateway: m.StorageGateway, auth_state: AuthenticationState
        ) -> bool:
            authorizer = StorageGatewayAuthorizer(
                gateway, enforce_authentication_timeout=False
            )
            id_set = authorizer.authorize_identities(auth_state)
            return bool(id_set)

        if gateway:
            return _in_gateway_domains(gateway, self._auth_state)

        for gw in m.StorageGateway.query.filter(
            ~m.StorageGateway.deleted  # type: ignore
        ):
            if _in_gateway_domains(gw, self._auth_state):
                return True

        return False

    def gateway_authenticated_identities(
        self,
        gateway: m.StorageGateway,
        idset: Optional[IdentitySet] = None,
    ) -> IdentitySet:
        if idset is None:
            idset = IdentitySet(self._auth_state.identity_set_data)

        key = "__gateway_authenticated_identities"
        if not hasattr(self, key):
            setattr(self, key, {})
        cache = getattr(self, key)

        def _idset_to_key(idset: IdentitySet, gateway_id: str) -> Hashable:
            sorted_ids = sorted([id["id"] for id in idset])
            unique_ids = set(sorted_ids)
            return (*unique_ids, gateway_id)

        cache_key = _idset_to_key(idset, gateway.id)

        if cache_key not in cache:
            sg_filter = AuthenticationTimeoutFilter(
                require_session=gateway.high_assurance,
                authentication_timeout=gateway.authentication_timeout_mins,
                require_mfa=gateway.require_mfa,
            )
            cache[cache_key] = sg_filter(idset, self._auth_state)

        return cache[cache_key]

    #
    # Collection operations
    #

    def created_by_caller(self, col: m.Collection) -> bool:
        """
        Collection.id is in auth_state.identity_set
        """
        # No reason to cache
        return col.identity_id in self._auth_state.identity_set

    def has_collection_role(self, col: Optional[m.Collection] = None) -> bool:
        if col:
            cache_key = str(col.id)
            collection_roles = [
                "collection:administrator:{collection_id}",
                "collection:activity_manager:{collection_id}",
                "collection:activity_monitor:{collection_id}",
                "collection:access_manager:{collection_id}",
            ]
        else:
            cache_key = "ANY"
            collection_roles = [
                "collection:administrator:*",
                "collection:activity_manager:*",
                "collection:activity_monitor:*",
                "collection:access_manager:*",
            ]

        key = "__has_collection_role"
        if not hasattr(self, key):
            setattr(self, key, {})
        cache = getattr(self, key)

        if cache_key not in cache:
            role_specs = [RoleSpec(r, col) for r in collection_roles]
            authorizer = RoleAuthorizer(role_specs, caching=True)
            idset = authorizer.authorize_identities(self._auth_state)
            cache[cache_key] = bool(idset)

        return cache[cache_key]

    def collection_admin_identities(self, col: m.Collection) -> IdentitySet:
        """
        Return IdentitySet containing caller's identities that have the collection
        admin role.
        """
        key = "__collection_admin_identities"
        if not hasattr(self, key):
            setattr(self, key, {})
        cache = getattr(self, key)

        if col.id not in cache:
            collection_roles = [
                "collection:administrator:{collection_id}",
            ]
            role_specs = [RoleSpec(r, col) for r in collection_roles]
            collection_role_authorizer = RoleAuthorizer(role_specs, caching=True)
            idset = collection_role_authorizer.authorize_identities(self._auth_state)
            cache[col.id] = idset

        return cache[col.id]

    def is_collection_admin(
        self,
        collection: Union[m.Collection, Iterable[m.Collection]],
        inherited: bool = False,
    ) -> bool:
        """
        Caller has collection admin role on all collections in 'col'.

        inherited: bool
        Return True if collection is a guest collection and the caller has the
        admin role on the mapped collection.
        """
        collections = collection if isinstance(collection, Iterable) else [collection]

        col: m.Collection
        for col in collections:
            if not bool(self.collection_admin_identities(col)):
                if not inherited:
                    return False
                if not col.is_guest_collection:
                    return False
                if not bool(self.collection_admin_identities(col.mapped_collection)):
                    return False
        return True

    def collection_is_visible(self, col: m.Collection) -> bool:
        """
        Collection is visible to the caller, although the caller does not necessarily
        meet the authentication timeout for the collection's storage gateway.
        """
        # Everyone can see public collections
        if col.public:
            return True

        # Endpoint roles can see all collections
        if self.has_endpoint_role():
            return True

        # Collection roles can view the collection
        if self.has_collection_role(col):
            return True

        # Mapped collection roles can see private guest collections
        if col.is_guest_collection and self.has_collection_role(col.mapped_collection):
            return True

        # Gateway-authorized callers can see private collections on the gateway
        if self.in_gateway_domains(col.storage_gateway):
            return True

        # TODO: Caller is ACL authorized. Until then, private guests are
        # just hidden.

        return False

    def any_public_collections(self) -> bool:
        key = "__any_public_collections"
        if not hasattr(self, key):
            setattr(self, key, False)
            col: m.Collection
            for col in m.Collection.query.filter(~m.Collection.is_deleted):  # type: ignore
                if col.public:
                    setattr(self, key, True)
                    break

        return getattr(self, key)


class _VisibleCollectionFilter:
    def __init__(
        self,
        query: Iterable[m.Collection],
        access_checker: _AccessChecker,
        filter_created_by_me: bool,
        filter_managed_by_me: bool,
    ) -> None:
        self._query = query
        self._access_checker = access_checker
        self._filter_created_by_me = filter_created_by_me
        self._filter_managed_by_me = filter_managed_by_me

    def __iter__(self) -> Generator[m.Collection, None, None]:
        for col in self._query:
            # Skip collections not created by me, if requested
            if (
                self._filter_created_by_me
                and not self._access_checker.created_by_caller(col)
            ):
                continue

            # Skip collections not managed by me, if requested
            if (
                self._filter_managed_by_me
                and not self._access_checker.has_collection_role(col)
            ):
                continue

            # Skip non-visible collections. Most expensive check last.
            if not self._access_checker.collection_is_visible(col):
                continue

            yield col

    def _new_filter_query(self, func, *args, **kwargs) -> "_VisibleCollectionFilter":
        return _VisibleCollectionFilter(
            getattr(self._query, func)(*args, **kwargs),
            self._access_checker,
            self._filter_created_by_me,
            self._filter_managed_by_me,
        )

    def filter(self, *args, **kwargs) -> "_VisibleCollectionFilter":
        return self._new_filter_query("filter", *args, **kwargs)

    def order_by(self, *args, **kwargs) -> "_VisibleCollectionFilter":
        return self._new_filter_query("order_by", *args, **kwargs)


@blueprint.route(  # type: ignore
    "/collections",
    methods=["GET"],
    summary="List the collections on this endpoint",
    operationId="listCollections",
)
@result_data(
    schemas=[data_types.Collection],
    description="List collections response",
)
@returns_error(
    error_type=errors.ForbiddenError,
)
@returns_error(
    error_type=errors.NotFoundError,
    response_id="NotFound",
)
@returns_error(
    error_type=errors.AuthenticationTimeoutError,
)
@query_parameter(
    name_spec="<uuid:mapped_collection_id>",
    description="""
    Filter collections which were created using this mapped_collection_id
    """,
)
@query_parameter(
    name_spec="<uuid:storage_gateway_id>",
    description="""
    Filter collections which were created using this storage_gateway_id
    """,
)
@query_parameter(
    name_spec="filter",
    description="""
    Filter to apply to the return set
    """,
    converter=collection_list_filter_converter(),
    default=[],
)
@query_parameter(
    name_spec="include",
    description="Document values to include",
    converter=enum_list_converter(CollectionGetInclude),
    default=[],
)
@document_authorization(
    authorization=OpenAPIStorageGatewayAuthorizer(scopes=["{storage_gateway_id}"])
)
@document_authorization(
    authorization=OpenAPIRoleAuthorizer(
        roles=[
            RoleType.endpoint_owner.value,
            RoleType.endpoint_administrator.value,
            RoleType.endpoint_activity_manager.value,
            RoleType.endpoint_activity_monitor.value,
            RoleType.collection_administrator.value + ":{collection_id}",
            RoleType.collection_activity_manager.value + ":{collection_id}",
            RoleType.collection_activity_monitor.value + ":{collection_id}",
            RoleType.collection_access_manager.value + ":{collection_id}",
            RoleType.collection_administrator.value + ":{mapped_collection_id}",
            RoleType.collection_activity_manager.value + ":{mapped_collection_id}",
        ]
    )
)
@document_authorization(authorization=OpenAPIPublicAuthorizer())
@paginated
@authenticated
def collections_get(
    *,
    page_size: int,
    marker: Optional[Paginator],
    mapped_collection_id: Optional[uuid.UUID],
    storage_gateway_id: Optional[uuid.UUID],
    filter: List[CollectionListFilter],
    include: List[CollectionGetInclude],
) -> Result:
    """
    This operation requires either the endpoint to have the `public` property
    set to true or the caller to have a role that allows viewing this
    Collection.

    The result of this can be limited by using the `filter` query parameter to
    choose which of the visible collections to return.  This is a
    comma-separated list of filters to apply to the result set:

    <dl>
    <dt>mapped_collections</dt>
        <dd>Only collections with collection_type equal to <em>mapped</em>.</dd>
    <dt>guest_collections</dt>
        <dd>Only collections with collection_type equal to <em>guest</em>.</dd>
    <dt>managed_by_me</dt>
        <dd>Only collections where one of caller's identities (either directly or
        via a group role assignment) is granted a role on the collection.</dd>
    <dt>created_by_me</dt>
        <dd>Only collections where one of the caller's identities matches the
        `identity_id` property of the collection.</dd>
    <dt>last_access < YYYY-MM-DD</dt>
    <dt>last_access <= YYYY-MM-DD</dt>
    <dt>last_access <= YYYY-MM-DD</dt>
    <dt>last_access = YYYY-MM-DD</dt>
    <dt>last_access >= YYYY-MM-DD</dt>
    <dt>last_access < YYYY-MM-DD</dt>
        <dd>Only collections accessed before or after the given date</dd>
    <dt>created_at < YYYY-MM-DD</dt>
    <dt>created_at <= YYYY-MM-DD</dt>
    <dt>created_at <= YYYY-MM-DD</dt>
    <dt>created_at = YYYY-MM-DD</dt>
    <dt>created_at >= YYYY-MM-DD</dt>
    <dt>created_at < YYYY-MM-DD</dt>
        <dd>Only collections created before or after the given date</dd>
    </dl>

    The result can also be limited by including the `mapped_collection_id`
    query parameter. This limits the response to guest collections which have
    been created using the specified mapped collection.

    Normally, only public collection configuration policy data is included in
    the response. If the query parameter `include=private_policies` is passed
    to this API, and the caller has an administrator role on this collection,
    the response will include all private policies for the collection as well.
    """
    include_private_policies = CollectionGetInclude.private_policies in include
    filter_managed_by_me = CollectionListFilterName.managed_by_me in filter
    filter_created_by_me = CollectionListFilterName.created_by_me in filter
    access_checker = _AccessChecker(g.auth_state)

    #
    # Begin the query and apply optional filters
    #

    query = _VisibleCollectionFilter(
        m.Collection.query.filter(~m.Collection.is_deleted),  # type: ignore
        access_checker,
        filter_created_by_me,
        filter_managed_by_me,
    )

    # Caller wants only guest collections attached to this mapped collection
    if mapped_collection_id:
        query = query.filter(
            m.Collection.collection_type == c.CollectionType.guest.name,
            m.Collection.mapped_collection_id == str(mapped_collection_id),
        )

    # Caller wants only collections attached to this storage gateway
    if storage_gateway_id:
        query = query.filter(
            m.Collection.storage_gateway_id == str(storage_gateway_id),
        )

    opnames = {
        "<": "__lt__",
        "<=": "__le__",
        ">": "__gt__",
        ">=": "__ge__",
        "=": "__eq__",
    }
    # The query can support the non-role based filters, but roles are
    # more complicated because of group permissions
    for f in filter:
        if f == CollectionListFilterName.mapped_collections:
            query = query.filter(m.Collection.is_mapped_collection)
        elif f == CollectionListFilterName.guest_collections:
            query = query.filter(m.Collection.is_guest_collection)
        elif f == CollectionListFilterName.created_by_me:
            filter_created_by_me = True
        elif f.filter_name == CollectionListFilterName.created_at:
            assert f.operator is not None
            assert f.operand is not None

            query = query.filter(
                getattr(m.Collection.created_at_date, opnames[f.operator])(f.operand)
            )
        elif f.filter_name == CollectionListFilterName.last_access:
            assert f.operator is not None
            assert f.operand is not None

            if f.operator == "<" or f.operator == "<=":
                query = query.filter(
                    (getattr(m.Collection.last_access, opnames[f.operator])(f.operand))
                    | (m.Collection.last_access.is_(None))
                )
            else:
                query = query.filter(
                    getattr(m.Collection.last_access, opnames[f.operator])(f.operand)
                )

    #
    # Apply authorization checks
    #

    # Some callers shouldn't even be using this endpoint
    if (
        not access_checker.is_endpoint_public()
        and not access_checker.has_endpoint_role()  # Any role on endpoint
        and not access_checker.in_gateway_domains()  # Any gateway
        and not access_checker.has_collection_role()  # Role on any collection
        and not access_checker.any_public_collections()
    ):
        raise errors.ForbiddenError()

    if mapped_collection_id:
        # Verify that the mapped collection exists
        mapped = m.Collection.query.filter(
            ~m.Collection.is_deleted,  # type: ignore
            m.Collection.id == str(mapped_collection_id),
        ).one_or_none()

        # If mapped_collection_id does not exist or the caller does not have
        # access to it...
        if mapped is None or not access_checker.collection_is_visible(mapped):
            # We know from above that caller has a right to be here so return
            # 404 not 403
            raise errors.NotFoundError(
                f"mapped collection {mapped_collection_id} does not exist"
            )

    if include_private_policies:
        # Check that caller has admin privileged on all visible collections
        if (
            not filter_created_by_me  # Fast path
            and not filter_managed_by_me  # Fast path
            and not access_checker.is_endpoint_admin()
            and not access_checker.is_collection_admin(query, inherited=True)
        ):
            raise errors.ForbiddenError(
                "You must have administrator role for all returned collections when "
                "requesting private policies."
            )

        # Check that admin role identity has authenticated recently enough to view
        # private policies on all returned collections.
        ep_admin_idset = access_checker.endpoint_admin_identities()
        for col in query:
            g_admin_idset = access_checker.collection_admin_identities(col)
            m_admin_idset = IdentitySet()
            if col.is_guest_collection:
                m_admin_idset = access_checker.collection_admin_identities(
                    col.mapped_collection
                )

            admin_idset = IdentitySet(
                ep_admin_idset
                + [a for a in g_admin_idset if a not in ep_admin_idset]
                + [
                    a
                    for a in m_admin_idset
                    if a not in ep_admin_idset and a not in g_admin_idset
                ]
            )

            idset = access_checker.gateway_authenticated_identities(
                col.storage_gateway,
                IdentitySet(admin_idset),
            )

            if not idset:
                raise errors.AuthenticationTimeoutError(
                    high_assurance=col.storage_gateway.high_assurance,
                    require_mfa=col.storage_gateway.require_mfa,
                    detail=idset.error_detail,
                )

    # Check that the caller meets the authentication requirements of all HA gateways.
    for col in query:
        if not col.high_assurance:
            continue

        idset = access_checker.gateway_authenticated_identities(col.storage_gateway)
        if not idset:
            raise errors.AuthenticationTimeoutError(
                high_assurance=col.high_assurance,
                require_mfa=col.require_mfa,
                detail=idset.error_detail,
            )

    #
    # Now collect up to page_size records for the caller.
    #

    # Apply the marker, if any
    if marker:
        query = query.filter(m.ConfigurationObject.id >= marker.marker)

    # Order the collections returned by ID so that markers make sense
    query = query.order_by(m.ConfigurationObject.id)

    next_page_token = None
    results: List[Dict[str, Any]] = []
    for col in query:
        if len(results) < page_size:
            serialized_collection = None
            if include_private_policies:
                serialized_collection = data_types.Collection.private_schema.dump(col)
                if (
                    not access_checker.is_endpoint_admin()
                    and col.is_guest_collection
                    and not access_checker.is_collection_admin([col.mapped_collection])
                ):
                    serialized_collection.pop("root_path", None)
            else:
                serialized_collection = data_types.Collection.public_schema.dump(col)

            results.append(serialized_collection)
        else:
            next_page_token = Base62Marker(marker_data=col.id)
            break

    res = data_types.Result.dump(Result(data=results, marker=next_page_token))

    response = jsonify(res)
    response.status_code = res["http_response_code"]

    return response


@blueprint.route(
    "/collections/<uuid:collection_id>",
    methods=["GET"],
    summary="Get information about a collection",
    path_parameters={"collection_id": "Id of the collection"},
    operationId="getCollection",
)
@result_data(
    schemas=[data_types.Collection],
    max_items=1,
    description="List collections response",
)
@returns_error(
    error_type=errors.NotFoundError,
    response_id="NotFound",
)
@query_parameter(
    name_spec="include",
    description="Document values to include",
    converter=enum_list_converter(CollectionGetInclude),
    default=[],
)
@document_authorization(
    authorization=OpenAPIStorageGatewayAuthorizer(scopes=["{storage_gateway_id}"])
)
@document_authorization(
    authorization=OpenAPIRoleAuthorizer(
        roles=[
            RoleType.endpoint_owner.value,
            RoleType.endpoint_administrator.value,
            RoleType.endpoint_activity_manager.value,
            RoleType.endpoint_activity_monitor.value,
            RoleType.collection_administrator.value + ":{collection_id}",
            RoleType.collection_activity_manager.value + ":{collection_id}",
            RoleType.collection_activity_monitor.value + ":{collection_id}",
            RoleType.collection_access_manager.value + ":{collection_id}",
            RoleType.collection_administrator.value + ":{mapped_collection_id}",
            RoleType.collection_activity_manager.value + ":{mapped_collection_id}",
        ]
    )
)
@document_authorization(authorization=OpenAPIPublicAuthorizer())
@authenticated
def collections_get_by_id(
    *, include: List[CollectionGetInclude], collection_id: uuid.UUID
):
    """
    This operation requires either the endpoint to have the `public` property
    set to `true` or the caller to have a role that allows viewing this
    Endpoint.  Some property visibility is limited for users who do not have an
    `administrator` role.

    Normally, only public collection configuration policy data is included in
    the response. If the query parameter `include=private_policies` is passed
    to this API, and the caller has an administrator role on this collection,
    the response will include all private policies for the collection as well.
    """
    include_private_policies = CollectionGetInclude.private_policies in include

    query = m.Collection.query.filter(
        m.Collection.id == str(collection_id), ~m.Collection.is_deleted  # type: ignore
    )
    col = query.one_or_none()

    if not col:
        raise errors.NotFoundError()

    endpoint_is_public = PublicAuthorizer().authorize_identities(g.auth_state)

    caller_has_endpoint_role = bool(
        EndpointRoleAuthorizer.authorize_identities(g.auth_state)
    )

    if not _check_collection_visibility(
        col,
        endpoint_is_public=endpoint_is_public,
        caller_has_endpoint_role=caller_has_endpoint_role,
    ):
        if endpoint_is_public:
            raise errors.NotFoundError()
        else:
            raise errors.ForbiddenError()

    res = None
    if include_private_policies:
        authorized_admins = _caller_is_collection_admin(col)
        if authorized_admins:
            datum = data_types.Collection.private_schema.dump(col)
            if col.is_guest_collection and not _caller_is_collection_admin(
                col.mapped_collection
            ):
                datum.pop("root_path", None)
            res = data_types.Result.dump(Result(data=[datum]))
        else:
            detail = None
            current_app.logger.debug("details=%s", authorized_admins.error_details)
            if authorized_admins.error_details:
                detail = authorized_admins.error_details[0]
            raise errors.AuthorizationError(detail=detail)
    else:
        res = data_types.Result.dump(
            Result(data=[data_types.Collection.public_schema.dump(col)])
        )

    response = jsonify(res)
    response.status_code = res["http_response_code"]

    return response


def _check_collection_visibility(
    collection,
    endpoint_is_public=False,
    caller_has_endpoint_role=False,
    filter_managed_by_me=False,
):
    """
    Perform collection visibility access checks based on the caller's
    identity set, collection, and endpoint policies.

    :param collection:
        The collection model to check for visibility.
    :param endpoint_is_public:
        True if the Endpoint.public property is True.
    :param caller_has_endpoint_role:
        True if the caller has any endpoint role.
    :param filter_managed_by_me:
        If true, the collection is visible only if the caller has an
        endpoint or collection role.

    :returns:
        Returns True if the collection is visible, False otherwise.
    """
    visible = caller_has_endpoint_role
    if visible:
        current_app.logger.debug("%s visible: caller has endpoint role", collection)

    if not visible and collection.public and not filter_managed_by_me:
        current_app.logger.debug("%s visible: collection is public", collection)
        visible = True

    if not visible:
        current_app.logger.debug("Checking for role granting access to %s", collection)

        collection_roles = [
            "collection:administrator:{collection_id}",
            "collection:activity_manager:{collection_id}",
            "collection:activity_monitor:{collection_id}",
            "collection:access_manager:{collection_id}",
        ]
        role_specs = [RoleSpec(r, collection) for r in collection_roles]
        collection_role_authorizer = RoleAuthorizer(role_specs)
        caller_has_collection_role = collection_role_authorizer.authorize_identities(
            g.auth_state
        )
        if bool(caller_has_collection_role):
            current_app.logger.debug(
                "%s visible: %s has a role on it",
                collection,
                caller_has_collection_role,
            )
            visible = True

    if not visible and not filter_managed_by_me:
        # TODO: ACL authorizer
        pass

    if not visible and not filter_managed_by_me:
        sg_authorizer = StorageGatewayAuthorizer(collection.storage_gateway)
        sg_authorized_identities = sg_authorizer.authorize_identities(g.auth_state)

        if bool(sg_authorized_identities):
            current_app.logger.debug(
                "%s visible: %s policies", collection, collection.storage_gateway
            )
            visible = True

    return visible


def _caller_is_collection_admin(collection) -> IdentitySet:
    """
    Check that the caller's identity set includes an identity with a
    collection:administrator role for this collection authenticated within the
    authentication_timeout_mins for the storage gateway this collection
    was created on.

    :param collection:
        The collection model to check for access to.
    :returns:
        This function returns an authorized IdentitySet object, which
        may be interpreted as a bool or checked for error details
    """
    current_app.logger.debug(
        "Checking whether caller is endpoint owner or collection_admin for %s",
        collection.id,
    )
    if not current_app.subscription:
        authorizer = RoleAuthorizer("endpoint:owner")
    else:
        authorizer = RoleAuthorizer(
            [
                RoleSpec("collection:administrator:{collection_id}", collection),
                RoleSpec("endpoint:administrator"),
            ]
        )

    idset = authorizer.authorize_identities(g.auth_state)

    current_app.logger.debug("identities to consider: %s", idset)

    # Enforce authorization timeouts
    idset_filter = AuthenticationTimeoutFilter(
        require_session=collection.storage_gateway.high_assurance,
        authentication_timeout=(collection.storage_gateway.authentication_timeout_mins),
        require_mfa=collection.require_mfa,
    )

    idset = idset_filter(idset, g.auth_state)

    current_app.logger.debug("result is %s", idset)
    return idset


@blueprint.route(
    "/collections",
    methods=["POST"],
    summary="Create a collection",
    operationId="postCollection",
)
@result_data(
    schemas=[data_types.Collection],
    http_status=201,
    max_items=1,
    description="Create collections response",
)
@returns_error(
    error_type=errors.ForbiddenError,
    error_detail=[
        errors.AuthenticationTimeout,
        errors.CredentialNotFound,
        errors.IdNotInIdentitySet,
        errors.InvalidCredential,
        errors.InvalidUser,
        errors.MissingRequiredRole,
        errors.NotFromAllowedDomain,
        errors.SubscriptionRequired,
    ],
)
@returns_error(
    error_type=errors.ConflictError,
    error_detail=[errors.ResourceConflict],
)
@document_authorization(
    authorization=OpenAPIStorageGatewayAuthorizer(scopes=["{storage_gateway_id}"])
)
@document_authorization(
    authorization=OpenAPIRoleAuthorizer(
        roles=[RoleType.endpoint_administrator.value, RoleType.endpoint_owner.value]
    )
)
@authenticated
@expects_json(
    schema=data_types.Collection,
    description="Collection definition",
    required=True,
)
def collections_post():
    """
    This is used to create either a mapped or a guest collection. When created,
    a "collection:administrator" role for that collection will be created using
    the caller's identity.

    The collection is assigned a unique DNS name. For guest collections, this
    DNS name begins with "g-". By default, for mapped collections this name
    begins with "m-", but a user with an "endpoint:administrator" role may
    assign a custom domain name for a mapped collection.

    In order to create a guest collection, the caller must have an identity
    that matches the Storage Gateway policies.

    In order to create a mapped collection, the caller must have an
    "endpoint:administrator" or "endpoint:owner" role.
    """
    collection_doc = g.validated_data
    with current_app.appsync_client.pause_notifications():
        collection_type = collection_doc["collection_type"]
        assert collection_type in c.CollectionType
        if collection_type == c.CollectionType.mapped:
            collection = collections_post_mapped()
        elif collection_type == c.CollectionType.guest:
            collection = collections_post_guest()

    if collection.owner_role:
        create_object(collection.owner_role)
        g.created_objects.append(str(collection.owner_role.id))

    try:
        collection.set_owner_string(collection.identity_id)
    except errors.APIBaseError as e:
        current_app.logger.warn(
            f"Unable to set owner_string on new collection {collection.id}: {e}"
        )

    res = data_types.Result.dump(
        Result(
            http_response_code=201,
            message="Created collection {}".format(collection.id),
            data=[data_types.Collection.dump(collection)],
        )
    )

    response = jsonify(res)
    response.status_code = res["http_response_code"]
    response.headers["Location"] = url_for(
        ".collections_get_by_id",
        collection_id=collection.id,
        _external=True,
        _scheme="https",
    )

    return response


def allocate_domain(collection_type, mapped_collection=None) -> str:
    """
    Allocate a new domain name for a collection and update endpoint state in
    appsync.

    :param collection_type:
        Enum value indicating whether this is a mapped or guest collection.
    :param mapped_collection:
        Mapped collection object which this collection is being
        created on, if this is a guest collection; otherwise None.

    :returns:
        Returns a string containing the fully-qualified name to use for
        this collection
    """

    def update_domain_counter(endpoint, collection):
        nonlocal mapped_collection

        collection["domain_name"] = endpoint.allocate_domain_name(
            collection["collection_type"], mapped_collection
        )

    collection = {"collection_type": collection_type.name}
    endpoint = m.Endpoint.query.one()
    update_object(endpoint, update_domain_counter, collection)
    return collection["domain_name"]


def validate_custom_domain_label(
    domain_name: str, collection_id: str | None = None
) -> str:
    """
    Ensure that the custom domain name for a mapped collection
    is valid, according to the following rules:

    -   The domain must not start with "g-" or "m-"
    -   If it contains a ".", the entire domain must be a child
        of the endpoint's wildcard domain
    -   The domain must not be used by any other collection.

    :param domain_name:
        Client-requested domain name.
    :param collection_id:
        ID of the collection we are updating. If None, then assume this
        is a create call. If we are updating, we don't error
        if the name is in use by itself.

    :returns:
        Returns the fully qualified domain name.

    :raises errors.ValidationError:
        The requested domain is not allowed because it does not match
        the required format.
    :raises errors.ConflictError:
        The requested domain is not allowed because it conflicts with
        an existing collection.
    """
    # Limited support for custom domains: must not start with g- or m-
    # and must be subdomains of the wildcard domain
    domain: Optional[str] = None
    if "." in domain_name:
        label, domain = domain_name.lower().split(".", 1)
    else:
        label = domain_name
        domain = None

    if label == "" or not re.match(r"^(?![0-9]+$)(?!-)[a-zA-Z0-9-]{,63}(?<!-)$", label):
        raise errors.ValidationError("Invalid domain_name value")

    if re.match(r"^(m-|g-)", label):
        raise errors.ValidationError("Invalid domain_name value")

    endpoint_domain = m.Endpoint.query.one().collection_subdomain
    if domain:
        if domain != endpoint_domain:
            raise errors.ValidationError("Invalid domain_name value")

    new_domain_name = "{}.{}".format(label, endpoint_domain)

    in_use = m.Collection.query.filter(
        m.Collection.domain_name == new_domain_name
    ).one_or_none()

    if in_use and ((not collection_id) or in_use.id != collection_id):
        raise errors.ConflictError(
            "domain_name already in use", resource_ids=[in_use.id]
        )

    return new_domain_name


def identity_data_for_requested_identity_id(
    storage_gateway: m.StorageGateway,
    identity_id: str | None,
    authorized_identities: IdentitySet,
    collection_type: c.CollectionType,
) -> Mapping[str, Any]:
    """
    Determine which identity data in authorized_identities to use when
    creating a collection. If the caller has provided a requested `identity_id`,
    then return the identity from `authorized_identities` has that id;
    otherwise, choose the first identity in the `authorized_identities` list.

    :param storage_gateway:
        Storage gateway to use to select an identity which maps if identity_id
        is None.
    :param identity_id:
        Optional string id of the requested identity. If this is not None, it
        must match the 'id' property of one of the elements of
        `authorized_identities`.
    :param authorized_identities:
        an IdentitySet
    :param collection_type:
        If this value is CollectionType.guest, then require a valid mapping for
        the user; otherwise select the identity in the identity set.

    :return:
        The member of `authorized_identities` whose `id` property
        matches the `identity_id`, or the first member of `authorized_identities`
        if `identity_id` is None.

    :raises errors.IdNotInIdentitySetError:
        No member of `authorized_identities` matches the requested
        `identity_id`.
    """
    if identity_id:
        authorized = [i for i in authorized_identities if i["id"] == identity_id]
        if not authorized:
            raise errors.IdNotInIdentitySetError(id=identity_id)
        return authorized[0]
    elif collection_type == c.CollectionType.mapped:
        return g.authorized_identities[0]
    else:
        accounts = storage_gateway.map_identities(authorized_identities)
        for account in accounts:
            user_credential = m.UserCredential.query.filter(
                ~m.UserCredential.deleted,  # type: ignore
                m.UserCredential.storage_gateway_id == storage_gateway.id,
                m.UserCredential.identity_id == account.identity_id,
                m.UserCredential.username == account.username,
            ).one_or_none()
            if user_credential:
                return [
                    i for i in authorized_identities if i["id"] == account.identity_id
                ][0]
    raise errors.BadRequestError("No valid mapping for identity set")


def get_storage_gateway_from_collection_doc():
    storage_gateway_id = g.validated_data.get("storage_gateway_id")
    user_credential_id = g.validated_data.get("user_credential_id")
    mapped_collection_id = g.validated_data.get("mapped_collection_id")
    storage_gateway = None
    user_credential = None
    mapped_collection = None

    if storage_gateway_id:
        storage_gateway = m.StorageGateway.query.filter(
            ~m.StorageGateway.deleted, m.StorageGateway.id == str(storage_gateway_id)
        ).one_or_none()

    if user_credential_id:
        user_credential = m.UserCredential.query.filter(
            ~m.UserCredential.is_deleted, m.UserCredential.id == str(user_credential_id)
        ).one_or_none()
        if user_credential:
            storage_gateway = user_credential.storage_gateway

    if mapped_collection_id:
        mapped_collection = m.Collection.query.filter(
            ~m.Collection.is_deleted,  # type: ignore
            m.Collection.id == str(mapped_collection_id),
        ).one_or_none()
        if mapped_collection:
            storage_gateway = mapped_collection.storage_gateway

    if storage_gateway is None:
        raise errors.BadRequestError("Invalid storage_gateway_id")

    g.storage_gateway = storage_gateway
    g.user_credential = user_credential
    g.mapped_collection = mapped_collection

    return storage_gateway


@requires_storage_gateway_policy(
    storage_gateway_selector=get_storage_gateway_from_collection_doc
)
@requires_subscription
@requires_data_access_scope
def collections_post_guest() -> m.Collection:
    collection_doc = g.validated_data

    user_credential = None
    storage_gateway = None
    identity_id: uuid.UUID | str | None = None
    domain_name = None
    endpoint = m.Endpoint.query.one()

    # Collection documents may have
    # user_credential_id:
    #   User credential associated with a guest collection. This must not
    #   be present in a mapped collection
    # identity_id:
    #   Auth identity of the creator of the collection. This can be guessed
    #   from either the user_credential_id or first entry that passes the
    #   identity mapping authorization step, or may be explicitly set.
    #
    # Whatever is present (if allowed for the collection type) must be
    # consistent:
    #
    # After resolving the storage_gateway_id and user_credential_id to
    # objects, user_credential.storage_gateway_id must equal
    # storage_gateway.id
    #
    # If identity_id is present it must be in the caller's identity set,
    # and in the list of authorized identities based on policies.
    storage_gateway = g.storage_gateway

    # If the caller has requested a specific user_credential_id be used,
    # verify that it exists
    user_credential_id = collection_doc.get("user_credential_id")
    if user_credential_id:
        user_credential = g.user_credential
        if user_credential is None:
            if storage_gateway:
                raise errors.CredentialNotFoundError(
                    accounts=storage_gateway.map_identities(g.authorized_identities)
                )
            else:
                raise errors.BadRequestError("Invalid user_credential_id")
        elif not user_credential.provisioned:
            raise errors.InvalidCredentialError(
                "Your credential requires some initial setup.",
                user_credential_id=user_credential.id,
            )
        elif user_credential.invalid:
            raise errors.InvalidCredentialError(user_credential_id=user_credential.id)

    # If the caller has not requested a specific identity_id be used,
    # default to that of the user_credential if known
    identity_id = collection_doc.get("identity_id")
    if not identity_id and user_credential:
        identity_id = user_credential.identity_id

    # Check that user_credential is associated with the requested storage
    # gateway
    if user_credential and storage_gateway:
        if user_credential.storage_gateway_id != storage_gateway.id:
            raise errors.BadRequestError(
                "User Credential {} not associated with Storage Gateway {}".format(
                    user_credential.id, storage_gateway.id
                )
            )

    # Check that user_credential is associated with the requested
    # user_credential
    if identity_id and user_credential:
        if user_credential.identity_id != str(identity_id):
            raise errors.BadRequestError(
                "User Credential {} not associated with identity {}".format(
                    user_credential.id, identity_id
                )
            )

    check_subscription(
        storage_gateway.connector,
        high_assurance=storage_gateway.high_assurance,
        collection_type=collection_doc["collection_type"],
    )

    if not g.mapped_collection:
        raise errors.BadRequestError("Invalid mapped_collection_id")

    if not g.mapped_collection.allow_guest_collections:
        raise errors.ForbiddenError(
            "Collection {} does not allow guest collections".format(
                g.mapped_collection.id
            )
        )

    authorized_identity = identity_data_for_requested_identity_id(
        storage_gateway,
        str(identity_id) if identity_id else None,
        g.authorized_identities,
        c.CollectionType.guest,
    )

    domain_name = allocate_domain(
        collection_doc["collection_type"], g.mapped_collection
    )

    # Create collection
    try:
        collection = m.Collection.from_collections_post(
            storage_gateway=storage_gateway,
            mapped_collection=g.mapped_collection,
            user_credential=user_credential,
            endpoint=endpoint,
            domain_name=domain_name,
            authorized_identity=authorized_identity,
            doc=g.validated_data,
        )
    except ValueError as v:
        raise errors.ValidationError(str(v) or None)

    create_object(collection)
    g.created_objects.append(collection.id)

    return collection


@requires_one_role(
    role_names=["endpoint:owner", "endpoint:administrator"],
    storage_gateway_selector=get_storage_gateway_from_collection_doc,
)
def collections_post_mapped():
    collection_doc = g.validated_data
    storage_gateway = g.storage_gateway
    endpoint = m.Endpoint.query.one()

    if collection_doc.get("allow_guest_collections", False):
        check_collection_type = c.CollectionType.guest
    else:
        check_collection_type = collection_doc["collection_type"]

    # Check for HA and connector support in the endpoint's subscription
    check_subscription(
        storage_gateway.connector,
        high_assurance=storage_gateway.high_assurance,
        collection_type=check_collection_type,
    )

    authorized_identity = identity_data_for_requested_identity_id(
        storage_gateway,
        collection_doc.get("identity_id"),
        g.authorized_identities,
        c.CollectionType.mapped,
    )

    if "domain_name" not in collection_doc:
        domain_name = allocate_domain(collection_doc["collection_type"])
    else:
        domain_name = validate_custom_domain_label(collection_doc["domain_name"])

    # Create collection
    try:
        collection = m.Collection.from_collections_post(
            storage_gateway=storage_gateway,
            endpoint=endpoint,
            domain_name=domain_name,
            authorized_identity=authorized_identity,
            doc=g.validated_data,
        )
    except ValueError as v:
        raise errors.ValidationError(str(v) or None)

    create_object(collection)
    g.created_objects.append(collection.id)

    return collection


def _update_collection(collection, collection_doc, update_func) -> Response:
    """
    Common routine for put and patch for collections.
    """
    if collection is None:
        raise errors.NotFoundError()

    field: str
    if collection.is_guest_collection:
        for field in ["user_message", "user_message_link"]:
            if field in collection_doc:
                exc = ma.ValidationError(
                    "Invalid field for guest collection.", field_name=field
                )
                raise errors.ValidationError(exc=exc)

    current_app.logger.debug(
        "Updating {} {} with {}".format(
            collection, collection.configuration_hash, collection_doc
        )
    )

    if collection_doc.get("allow_guest_collections", False):
        check_subscription(
            collection.storage_gateway.connector,
            high_assurance=collection.storage_gateway.high_assurance,
            collection_type=c.CollectionType.guest,
        )

    with current_app.appsync_client.pause_notifications():
        update_guests_transfer_docs = False
        guest_patch_doc = {}

        if collection.is_mapped_collection:

            def _update_force_encryption():
                if not collection_doc.get("force_encryption", False):
                    guest_patch_doc["force_encryption"] = False

            def _update_force_verify():
                if collection_doc.get("force_verify", False):
                    guest_patch_doc["disable_verify"] = False
                else:
                    guest_patch_doc["force_verify"] = False

            def _update_disable_verify():
                if collection_doc.get("disable_verify", False):
                    guest_patch_doc["force_verify"] = False
                else:
                    guest_patch_doc["disable_verify"] = False

            check: bool
            default_value: Any
            update_patch_func: Union[Callable[[], None], None]
            for check, field, default_value, update_patch_func in [
                (current_app.subscription, "enable_https", True, None),
                (True, "user_message", None, None),
                (True, "user_message_link", None, None),
                (True, "guest_auth_policy_id", None, None),
                (True, "force_encryption", False, _update_force_encryption),
                (True, "force_verify", False, _update_force_verify),
                (True, "disable_verify", False, _update_disable_verify),
            ]:
                if check:
                    if (
                        field in collection_doc
                        and getattr(collection, field, None) != collection_doc[field]
                    ) or (
                        update_func is put_object
                        and field not in collection_doc
                        and getattr(collection, field, default_value) != default_value
                    ):
                        update_guests_transfer_docs = True
                        if update_patch_func:
                            update_patch_func()

            if "domain_name" in collection_doc:
                if (
                    collection.has_domain
                    and collection_doc["domain_name"] != collection.domain.domain_name
                ):
                    raise errors.ValidationError(
                        "Unable to change domain_name on a collection with a custom domain"
                    )
                domain_name = validate_custom_domain_label(
                    collection_doc["domain_name"], collection_id=collection.id
                )
                # The above returns fqdn, where input may be partial
                collection_doc["domain_name"] = domain_name

        else:
            if "domain_name" in collection_doc:
                if collection.domain_name != collection_doc["domain_name"]:
                    raise errors.ValidationError(
                        "Unable to change domain_name on a guest collection"
                    )

        dependent_objs = [collection.endpoint]
        if collection.is_guest_collection:
            dependent_objs.append(collection.mapped_collection)
        update_func(collection, collection_doc, dependent_objs=dependent_objs)

        if update_guests_transfer_docs:
            ep_serial = (
                db.session.query(m.ConfigurationObject.serial)
                .filter(m.ConfigurationObject.tag == c.ConfigurationObjectTag.ENDPOINT)
                .scalar()
            )

            for guest in m.Collection.query.filter(
                ~m.Collection.deleted,  # type: ignore
                m.Collection.is_guest_collection,
                m.Collection.mapped_collection_id == collection.id,
            ):
                current_app.logger.debug(
                    f"Creating PendingUpdate to update guest {guest}"
                )
                pending_update = m.PendingUpdate(
                    object_id=guest.id,
                    change_time=datetime.datetime.now(tz=datetime.timezone.utc),
                    endpoint_serial=ep_serial,
                    collection_serial=collection.serial,
                    update_type=m.PendingUpdate.PATCH,
                    data=guest_patch_doc if guest_patch_doc else None,
                )
                db.session.add(pending_update)

        data = []
        data.append(data_types.Collection.dump(collection))
        db.session.commit()

        res = data_types.Result.dump(
            Result(message="Updated collection {}".format(collection.id), data=data)
        )

    response = jsonify(res)
    response.status_code = res["http_response_code"]
    g.updated_objects.append(collection.id)

    return response


def get_collection_id_from_path():
    collection_id = request.view_args.get("collection_id")
    collection = None
    collection = m.Collection.query.filter(
        ~m.Collection.is_deleted, m.Collection.id == str(collection_id)
    ).one_or_none()

    g.collection = collection
    return collection


@blueprint.route(
    "/collections/<uuid:collection_id>",
    methods=["PATCH"],
    summary="Update a collection",
    path_parameters={"collection_id": "Id of the collection"},
    operationId="patchCollection",
)
@result_data(
    schemas=[],
    description="Update collections response",
)
@authenticated
@requires_one_role(
    role_names=[
        "endpoint:owner",
        "endpoint:administrator",
        "collection:administrator:{collection_id}",
        "collection:administrator:{mapped_collection_id}",
    ],
    collection_selector=get_collection_id_from_path,
)
@expects_json(
    schema=data_types.Collection,
    patch_schema=True,
    description="Changes to apply to the collection",
    required=True,
)
def collections_patch_by_id(collection_id):
    """
    Updates a collection, changing only the properties included in the input
    document. It optionally returns a document containing the document after
    the change is applied.  Items explicitly set to null in the input are
    removed from the collection document.
    """
    return _update_collection(g.collection, g.validated_data, patch_object)


@blueprint.route(
    "/collections/<uuid:collection_id>",
    methods=["PUT"],
    summary="Update a collection",
    path_parameters={"collection_id": "Id of the collection"},
    operationId="putCollection",
)
@result_data(
    schemas=[],
    description="Update collections response",
)
@authenticated
@requires_one_role(
    role_names=[
        "endpoint:owner",
        "endpoint:administrator",
        "collection:administrator:{collection_id}",
        "collection:administrator:{mapped_collection_id}",
    ],
    collection_selector=get_collection_id_from_path,
)
@expects_json(
    schema=data_types.Collection,
    description="New collection definition",
    required=True,
)
def collections_put_by_id(collection_id):
    """
    Update a collection, completely replacing its definition with the new
    document. It returns a document containing the collection after the
    update has been applied.
    """
    return _update_collection(g.collection, g.validated_data, put_object)


@blueprint.route(
    "/collections/<uuid:collection_id>",
    methods=["DELETE"],
    summary="Delete a collection",
    path_parameters={"collection_id": "Id of the collection"},
    operationId="deleteCollection",
)
@result_data(
    schemas=[],
    description="Delete collections response",
)
@authenticated
@requires_one_role(
    role_names=[
        "endpoint:owner",
        "endpoint:administrator",
        "collection:administrator:{collection_id}",
        "collection:administrator:{mapped_collection_id}",
    ],
    collection_selector=get_collection_id_from_path,
)
def collections_delete_by_id(collection_id: uuid.UUID) -> Response:
    """
    Deletes a collection owned by the caller or an endpoint administrator.
    If the collection has the delete_protection property set to true, the
    collection can not be deleted.

    When a collection is deleted, all collection-specific roles and
    sharing_policies are also deleted.

    If a mapped collection is deleted, then all guest collections and roles
    associated them are also deleted.
    """
    if g.collection is None:
        raise errors.NotFoundError()

    collection = g.collection
    ep_serial = (
        db.session.query(m.ConfigurationObject.serial)
        .filter(m.ConfigurationObject.tag == c.ConfigurationObjectTag.ENDPOINT)
        .scalar()
    )
    if collection.is_mapped_collection and collection.delete_protected:
        raise errors.ForbiddenError(
            f"Collection {collection.id} has delete protection enabled"
        )

    with current_app.appsync_client.pause_notifications(), m.db.session.no_autoflush:
        # (soft) Delete the collection
        delete_object(collection, update=True)
        collection_serial = collection.serial

        current_app.logger.debug(
            f"collection {collection.id} soft deleted, processing dependent objects"
        )
        # Mark other objects locally to be deleted later.
        for gc in m.Collection.query.filter(
            # We want to add deleted collections as well for any dangling soft-deleted
            # guests that occurred before the upgrade to the GCS version that supports async
            # operations
            m.Collection.mapped_collection_id == collection.id,
            m.Collection.is_guest_collection,
        ):
            gcid = gc.id
            for r in m.Role.query.filter(m.Role.resource == gc.id):
                current_app.logger.debug(
                    f"Creating PendingUpdate to delete guest role {r}"
                )
                pending_update: m.PendingUpdate = m.PendingUpdate(
                    object_id=r.id,
                    endpoint_serial=ep_serial,
                    change_time=datetime.datetime.now(tz=datetime.timezone.utc),
                    collection_serial=gc.serial,
                    update_type=m.PendingUpdate.DELETE,
                )
                db.session.add(pending_update)

            current_app.logger.debug(f"Creating PendingUpdate to delete guest {gc}")
            pending_update = m.PendingUpdate(
                object_id=gcid,
                change_time=datetime.datetime.now(tz=datetime.timezone.utc),
                endpoint_serial=ep_serial,
                collection_serial=collection_serial,
                update_type=m.PendingUpdate.DELETE,
            )
            db.session.add(pending_update)

        # Delete appsync representations of the role objects
        # for this collection.
        for r in m.Role.query.filter(m.Role.resource == collection.id):
            current_app.logger.debug(f"Creating PendingUpdate to delete role {r}")
            pending_update = m.PendingUpdate(
                object_id=r.id,
                change_time=datetime.datetime.now(tz=datetime.timezone.utc),
                endpoint_serial=ep_serial,
                collection_serial=collection_serial,
                update_type=m.PendingUpdate.DELETE,
            )
            db.session.add(pending_update)
        pending_update = m.PendingUpdate(
            object_id=collection.id,
            change_time=datetime.datetime.now(tz=datetime.timezone.utc),
            endpoint_serial=ep_serial,
            collection_serial=collection_serial,
            update_type=m.PendingUpdate.DELETE,
        )
        db.session.add(pending_update)
        db.session.commit()

    res = data_types.Result.dump(
        Result(message="Deleted collection {}".format(str(collection_id)))
    )
    g.deleted_objects.append(collection.id)

    response = jsonify(res)
    response.status_code = res["http_response_code"]

    return response


@blueprint.route(
    "/collections/<uuid:collection_id>/domain",
    methods=["PUT"],
    summary="Set custom domain for a collection",
    path_parameters={"collection_id": "Id of the collection"},
    operationId="putCollectionDomain",
    tags=["Domains"],
)
@result_data(
    schemas=[],
    description="Set collection domain response",
)
@authenticated
@requires_one_role(role_names=["endpoint:owner", "endpoint:administrator"])
@expects_json(
    schema=data_types.Domain,
    description="New collection domain definition",
    required=True,
)
def collection_domain_put(collection_id: uuid.UUID) -> Response:
    """
    Register a new custom domain and certificate to to be used to serve
    this collection.

    The domain is used for the collection. If this is a mapped collection
    and the `wildcard` property is set to true, then all all guest collections
    associated with this collection that do not have a custom domain will be
    updated to use subdomains of that domain. Otherwise, only this collection
    will use that domain.

    This requires an `administrator` role on the Endpoint
    """
    with current_app.appsync_client.pause_notifications():
        collection: Optional[m.Collection] = m.Collection.query.filter(
            ~m.Collection.is_deleted,  # type: ignore
            m.Collection.id == str(collection_id),
        ).one_or_none()

        if collection is None:
            raise errors.NotFoundError("Collection not found")

        old_domain_name: Optional[str] = None
        try:
            old_domain = collection.domain
            old_domain_name = old_domain.domain_name
            old_wildcard = bool(old_domain and old_domain.wildcard)
        except AttributeError:
            old_wildcard = False

        if g.validated_data["domain_name"] != collection.domain_name:
            validate_unique_custom_domain(g.validated_data["domain_name"])

        update_object(collection, m.Collection.put_domain, g.validated_data)
        g.updated_objects.append(str(collection.id))

        if (
            old_wildcard or collection.domain.wildcard
        ) and old_domain_name != collection.domain.domain_name:
            _update_guest_collections_without_custom_domains(collection)

    res = data_types.Result.dump(Result(message="Updated collection"))
    response = jsonify(res)
    response.status_code = res["http_response_code"]

    return response


@blueprint.route(
    "/collections/<uuid:collection_id>/domain",
    methods=["DELETE"],
    summary="Delete custom domain for a collection",
    path_parameters={"collection_id": "Id of the collection"},
    operationId="deleteCollectionDomain",
    tags=["Domains"],
)
@result_data(
    schemas=[],
    description="Delete collection domain response",
)
@authenticated
@requires_one_role(role_names=["endpoint:owner", "endpoint:administrator"])
def collection_domain_delete(collection_id: uuid.UUID) -> Response:
    """
    Delete the custom collection domain.

    If this is a mapped collection, this will cause the collection to revert to
    a subdomain of the endpoint's domain (if it is wildcard domain) or a
    subdomain of the endpoint's data.globus.org domain. If this mapped collection
    has a wildcard domain when this is called, then all guest collections without
    custom domains will have their domains changed as well.

    If this is a guest collection, and the mapped collection it was created from
    has a custom wildcard domain, then this collection will become a subdomain
    of that domain; otherwise it will revert to a subdomain of either the
    endpoint's domain (if it is a wildcard domain) or a subdomain of the endpoint's
    data.globus.org domain.

    This requires an `administrator` role on the Endpoint.
    """
    with current_app.appsync_client.pause_notifications():
        collection = m.Collection.query.filter(
            ~m.Collection.is_deleted,  # type: ignore
            m.Collection.id == str(collection_id),
        ).one_or_none()
        if not collection:
            raise errors.NotFoundError("Collection not found")

        try:
            old_domain = collection.domain
            old_wildcard = bool(old_domain and old_domain.wildcard)
        except AttributeError:
            raise errors.NotFoundError("No domain is configured for this collection.")

        domain_name = allocate_domain(c.CollectionType(collection.collection_type))

        def delete_domain(collection_model):
            collection_model.delete_domain(new_domain=domain_name)

        update_object(collection, delete_domain)
        g.updated_objects.append(str(collection.id))

        if old_wildcard:
            _update_guest_collections_without_custom_domains(collection)

    res = data_types.Result.dump(Result(message="Updated collection"))

    response = jsonify(res)
    response.status_code = res["http_response_code"]

    return response


@blueprint.route(
    "/collections/<uuid:collection_id>/domain",
    methods=["GET"],
    summary="Get custom domain for a collection",
    path_parameters={"collection_id": "Id of the collection"},
    operationId="getCollectionDomain",
    tags=["Domains"],
)
@result_data(
    schemas=[data_types.Domain],
    description="Get collection domain response",
    max_items=1,
)
@authenticated
@requires_one_role(role_names=["endpoint:owner", "endpoint:administrator"])
def collection_domain_get(collection_id: uuid.UUID) -> Response:
    """
    Get the custom domain document associated with this collection.

    This requires an `administrator` role on the Endpoint
    """
    with current_app.appsync_client.pause_notifications():
        collection = m.Collection.query.filter(
            ~m.Collection.is_deleted,  # type: ignore
            m.Collection.id == str(collection_id),
        ).one_or_none()
        if not collection:
            raise errors.NotFoundError("Collection not found")
        old_domain = getattr(collection, "domain", None)

        if not old_domain:
            raise errors.NotFoundError("No domain is configured for this collection.")

    res = data_types.Result.dump(Result(data=[data_types.Domain.dump(old_domain)]))

    response = jsonify(res)
    response.status_code = res["http_response_code"]

    return response


@blueprint.route(
    "/collections/<uuid:collection_id>/owner_string",
    methods=["PUT"],
    summary="Set advertised owner of collection",
    path_parameters={"collection_id": "Id of the collection"},
    operationId="putCollectionOwnerString",
)
@result_data(
    schemas=[],
    description="Set collection owner string response",
)
@authenticated
@requires_one_role(
    role_names=["endpoint:owner", "collection:administrator:{collection_id}"],
    collection_selector=get_collection_id_from_path,
)
@expects_json(
    schema=data_types.OwnerString,
    description="New collection owner string",
    required=True,
)
def collection_owner_string_put(collection_id: uuid.UUID) -> Response:
    """
    Update the advertised owner string of the collection

    Modify the collection's advertised owner to match the username of one of
    the caller's linked identities.  The identity must have an administrator
    role on the collection.
    """
    new_owner_id = str(g.validated_data["identity_id"])

    # Validate the requested owner_id is in the caller's identity set
    id_set = g.auth_state.identity_set_data
    new_owner_id_data = [i for i in id_set if i["id"] == new_owner_id]
    if not new_owner_id_data:
        raise errors.IdNotInIdentitySetError(id=new_owner_id)

    if current_app.subscription:
        # Validate the requested owner_id has a role or is in a group with a role
        allowed_roles = [
            RoleSpec(
                "collection:administrator:{collection_id}", collection=g.collection
            )
        ]
    else:
        allowed_roles = [RoleSpec("endpoint:owner")]
    role_authorizer = RoleAuthorizer(allowed_roles)
    try:
        role_id = role_authorizer.authorize_identity(new_owner_id, g.auth_state)
    except errors.MissingRequiredRoleError as e:
        raise errors.MissingRequiredRoleError(
            f"The identity {new_owner_id} does not have a valid role on the collection",
            detail=e.detail if isinstance(e.detail, errors.ErrorDetail) else None,
            roles=allowed_roles,
        )

    if not role_id:
        raise errors.MissingRequiredRoleError(roles=allowed_roles)

    new_username = new_owner_id_data[0].get("username")
    g.collection.set_owner_string(new_owner_id)

    res = data_types.Result.dump(
        Result(
            message="Updated collection owner_string to "
            f"{new_username} ({new_owner_id})"
        )
    )
    response = jsonify(res)
    response.status_code = res["http_response_code"]

    return response


@blueprint.route(
    "/collections/<uuid:collection_id>/owner_string",
    methods=["DELETE"],
    summary="Reset advertised owner of collection",
    path_parameters={"collection_id": "Id of the collection"},
    operationId="deleteCollectionOwnerString",
)
@result_data(
    schemas=[],
    description="Delete collection owner string response",
)
@authenticated
@requires_one_role(
    role_names=[
        "collection:administrator:{collection_id}",
        "endpoint:administrator",
        "endpoint:owner",
    ],
    collection_selector=get_collection_id_from_path,
)
def collection_owner_string_delete(collection_id: uuid.UUID) -> Response:
    """
    Reset the advertised owner string of the collection to the endpoint's client_id.
    """
    g.collection.reset_owner_string()

    res = data_types.Result.dump(
        Result(
            message="Reset collection owner_string to "
            f"{current_app.client_id}@clients.auth.globus.org"
        )
    )
    response = jsonify(res)
    response.status_code = res["http_response_code"]

    return response


def _update_guest_collections_without_custom_domains(collection: m.Collection) -> None:
    ep_serial: int = (
        db.session.query(m.ConfigurationObject.serial)
        .filter(m.ConfigurationObject.tag == c.ConfigurationObjectTag.ENDPOINT)
        .scalar()
    )
    guest: m.Collection

    for guest in m.Collection.query.filter(
        ~m.Collection.is_deleted,  # type: ignore
        m.Collection.is_guest_collection,
        m.Collection.mapped_collection_id == collection.id,
    ):
        if not guest.has_domain:
            pending_update = m.PendingUpdate(
                object_id=guest.id,
                endpoint_serial=ep_serial,
                change_time=datetime.datetime.now(tz=datetime.timezone.utc),
                collection_serial=collection.serial,
                update_type=m.PendingUpdate.UPDATE_DOMAIN,
            )
            m.db.session.add(pending_update)
    m.db.session.commit()


@blueprint.route(
    "/collections/batch_delete",
    methods=["POST"],
    summary="Delete multiple guest collections",
)
@result_data(
    schemas=[data_types.Batch],
    max_items=1,
    http_status=202,
    description="Delete multiple collections response",
)
@returns_error(
    error_type=errors.InvalidBatchItemsError,
    error_detail=[errors.Batch],
)
@authenticated
@requires_one_role(
    role_names=["endpoint:owner", "endpoint:administrator"],
)
@expects_json(
    schema=data_types.Batch,
    description="List of collection ids to delete",
    required=True,
)
def collections_batch_guest_delete() -> Response:
    """
    Initiate the deletion of multiple guest collections. The input document
    contains a list of the IDs of collections to delete.

    If any of the collections have collection_type of "mapped", then this
    operation returns an error indicating which ones were not valid or this
    operation.

    If any of the collections do not exist or are already deleted, then they
    are silently ignored.

    Deletion does not happen immediately; it is handled in the background by
    the GCS Manager Assistant process.

    On success, this operation returns a message body containing the list of
    collections from the input that this GCS manager node will delete.
    """
    batch: Set[uuid.UUID] = set(g.validated_data["ids"])

    delete_collections = m.Collection.query.filter(
        ~m.Collection.is_deleted,  # type: ignore
        m.Collection.id.in_([str(s) for s in batch]),
    )
    valid_collections = []
    invalid_ids = []
    result_ids = []

    # Determine if there are any mapped collections in the input, if so, record
    # them to return in an error detail.
    # Filter out any collections that are already deleted
    for d in delete_collections:
        if not d.is_guest_collection:
            invalid_ids.append(uuid.UUID(d.id))
        else:
            valid_collections.append(d)

    if invalid_ids:
        raise errors.InvalidBatchItemsError(
            error_description=(
                "Unable to delete mapped collections via this interface",
            ),
            ids=invalid_ids,
        )

    ep_serial = m.Endpoint.query.one().serial
    for d in valid_collections:
        for r in d.roles:
            current_app.logger.debug(f"Creating PendingUpdate to delete guest role {r}")
            pending_update: m.PendingUpdate = m.PendingUpdate(
                object_id=r.id,
                endpoint_serial=ep_serial,
                change_time=datetime.datetime.now(tz=datetime.timezone.utc),
                collection_serial=d.serial,
                update_type=m.PendingUpdate.DELETE,
            )
            db.session.add(pending_update)

        current_app.logger.debug(f"Creating PendingUpdate to delete guest {d}")
        pending_update = m.PendingUpdate(
            object_id=d.id,
            change_time=datetime.datetime.now(tz=datetime.timezone.utc),
            endpoint_serial=ep_serial,
            collection_serial=d.serial,
            update_type=m.PendingUpdate.DELETE,
        )
        db.session.add(pending_update)
        result_ids.append(d.id)
    db.session.commit()

    results = data_types.Batch.dump({"ids": result_ids})
    res = data_types.Result.dump(Result(data=[results], http_response_code=202))
    response = jsonify(res)
    response.status_code = res["http_response_code"]
    return response


@blueprint.route(
    "/collections/<uuid:collection_id>/owner",
    methods=["PUT"],
    summary="Set collection owner",
    path_parameters={"collection_id": "ID of the collection"},
    operationId="setCollectionOwner",
)
@result_data(
    schemas=[],
    description="Set collection owner response",
)
@authenticated
@requires_one_role(
    role_names=[
        "endpoint:owner",
        "endpoint:administrator",
    ],
    collection_selector=get_collection_id_from_path,
)
@expects_json(
    schema=data_types.CollectionOwner, desription="New collection owner", required=True
)
def collection_owner_put(
    collection_id: uuid.UUID,
) -> Response:
    """
    Assign a new identity to act as the mapped collection owner. Caller must have
    an endpoint admin or owner role.

    - This is only allowed for mapped collections
    - Owner ID can not be the endpoint client ID
    """
    new_owner_id = str(g.validated_data["identity_id"])

    new_owner_id_result = current_app.auth_client.get_identities(
        ids=[new_owner_id]
    ).data.get("identities", [])

    if len(new_owner_id_result) != 1:
        current_app.logger.error("Unable to resolve identity_id=%s", new_owner_id)
        raise errors.BadRequestError(
            "Unable to resolve identity_id to a Globus Auth identity"
        )

    new_owner_id = new_owner_id_result[0]["id"]
    new_owner_username = new_owner_id_result[0]["username"]

    if new_owner_username == f"{current_app.client_id}@clients.auth.globus.org":
        current_app.logger.error("Cannot set the endpoint client as collection owner")
        raise errors.BadRequestError(
            "The endpoint client identity can not be used as the collection owner"
        )

    with current_app.appsync_client.pause_notifications():
        # Check if the collection exists
        collection: m.Collection = m.Collection.query.filter(
            ~m.Collection.is_deleted,  # type: ignore
            m.Collection.id == str(collection_id),
        ).one_or_none()

        if collection is None:
            raise errors.NotFoundError(f"Collection {collection_id} does not exist")

        # This operation is only supported on mapped collections.
        if collection.collection_type != c.CollectionType.mapped.name:
            raise errors.BadRequestError(
                "Changing the collection owner is only supported on mapped collections."
            )

        # Nothing to do if the owner is not changing
        if collection.identity_id != new_owner_id:
            # Update the collection's identity_id
            collection.identity_id = new_owner_id

            # Clear the primary bit on the old owner role
            def clear_primary(role: m.Role) -> None:
                role.primary_role = False

            # Regardless of whether or not this endpoint is managed,
            # if there is a primary admin role for this collection,
            # remove the primary flag so that if the endpoint later
            # set to managed, the collection admin roles are sane.
            if collection.owner_role is not None:
                old_admin_role: m.Role = collection.owner_role
                update_object(old_admin_role, clear_primary)
                g.updated_objects.append(old_admin_role.id)

            # Check if the new owner has an existing admin role
            admin_role: m.Role = m.Role.query.filter(
                ~m.Role.is_deleted,  # type: ignore
                m.Role.resource == collection.id,
                m.Role.identity_id == new_owner_id,
                m.Role.role == RoleType.collection_administrator,
            ).one_or_none()

            # No such role, create a new one but only if the endpoint is managed.
            if admin_role is None and current_app.subscription is not None:
                admin_role = m.Role(
                    doc=dict(
                        DATA_TYPE="role#1.0.0",
                        collection=collection.id,
                        principal=f"urn:globus:auth:identity:{new_owner_id}",
                        role="administrator",
                    ),
                    primary_role=True,
                )
                create_object(admin_role)
                g.created_objects.append(admin_role.id)
            elif admin_role is not None and admin_role.primary_role is False:
                # Regardless of whether the endpoint is managed or not, if there
                # is an existing collection admin role for the new owner, make
                # sure it is set to primary so that when the endpoint is later
                # set to managed, the collection admin roles are sane.

                # Set the primary bit
                def set_primary(role: m.Role) -> None:
                    role.primary_role = True

                update_object(admin_role, set_primary)
                g.updated_objects.append(admin_role.id)

            # Reset the owner string
            collection.reset_owner_string()

    res = data_types.Result.dump(
        Result(
            message=(
                f"Updated collection owner to {new_owner_id} ({new_owner_username})"
            )
        )
    )

    response = jsonify(res)
    response.status_code = res["http_response_code"]

    return response
