from typing import Any, Dict

import requests
from oauth2client import client as oauth_client

import globus.manager.api.errors as errors
import globus.manager.api.models as m
import globus.manager.api.schemas.connectors.box  # noqa

from .oauth_credential_api import OAuthCredentialAPI


class BoxConnector:
    CONNECTOR_ID = "7c100eae-40fe-11e9-95a3-9cb6d0d9fd63"
    DISPLAY_NAME = "Box"
    SCOPES = "root_readwrite"
    model = m.BoxStorageGateway

    def __init__(self, app):
        self.oauth_credential_api = OAuthCredentialAPI(app)
        self.oauth_credential_api.add_connector(self.CONNECTOR_ID, self)

    def initiate_flow(
        self,
        storage_gateway: m.StorageGateway,
        authorized_identity: Dict[str, Any],
        **kwargs
    ) -> oauth_client.OAuth2WebServerFlow:
        if not storage_gateway.user_credential_required:
            raise errors.UnsupportedMediaError()

        flow = oauth_client.OAuth2WebServerFlow(
            client_id=storage_gateway.client_id,
            client_secret=storage_gateway.secret,
            scope=self.SCOPES,
            redirect_uri=self.callback_url,
            auth_uri="https://account.box.com/api/oauth2/authorize",
            token_uri="https://api.box.com/oauth2/token",
            revoke_uri="https://api.box.com/oauth2/revoke",
            **kwargs,
        )
        if not getattr(storage_gateway, "allow_any_account", False):
            mapped_to = storage_gateway.map_identity(authorized_identity)
            split_username = mapped_to.split("@")
            flow.params["box_login"] = "@".join(split_username[0:2])

        return flow

    def id_token(self, access_token):
        """
        Generate profile info usable as an id_token.

        """
        id_token = None
        resource = "https://api.box.com/2.0/users/me"
        params = {"fields": "id,login,name,max_upload_size"}
        headers = {"Authorization": "Bearer {}".format(access_token)}
        response = requests.get(resource, params=params, headers=headers)

        if response.status_code >= 200 and response.status_code < 300:
            user_info = response.json()
            id_token = {
                "sub": user_info["id"],
                "email": user_info["login"],
                "name": user_info["name"],
                "max_upload": user_info["max_upload_size"],
            }
        else:
            raise errors.InvalidCredentialError(
                "Your credentials do not have sufficient permissions to access this collection",
            )
        return id_token

    @property
    def callback_url(self) -> str:
        return self.oauth_credential_api.authcallback_url
