from __future__ import annotations

import functools
import typing as t


class DocumentChecker:
    """
    Mixin that adds support for json document validation against the
    current state of a model object. This can be used to verify that the
    changes are both self-consistent and don't violate any constraints when
    the changes would be applied.

    This is implemented by the check_doc method, and the two related decorators
    @checks_property and @checks_policy_property
    """

    checkers: dict[tuple[str, str], t.Callable] = {}
    policy_checkers: dict[tuple[str, str], t.Callable] = {}
    required: set[tuple[str, str]] = set()
    required_policies: set[tuple[str, str]] = set()

    @staticmethod
    def __lookup_checker(cls, prop: str) -> t.Optional[t.Callable]:
        class_qualname = cls.__qualname__
        key = (class_qualname, prop)
        checker = DocumentChecker.checkers.get(key)
        if checker:
            return checker

        for base in cls.__bases__:
            checker = DocumentChecker.__lookup_checker(base, prop)
            if checker:
                return checker

        return checker

    @staticmethod
    def __lookup_policy_checker(cls, prop) -> t.Optional[t.Callable]:
        class_qualname = cls.__qualname__
        key = (class_qualname, prop)
        checker = DocumentChecker.policy_checkers.get(key)
        if checker:
            return checker

        for base in cls.__bases__:
            checker = DocumentChecker.__lookup_policy_checker(base, prop)
            if checker:
                return checker

        return checker

    @staticmethod
    def matches_qualname(cls, qualname: str) -> bool:
        if cls.__qualname__ == qualname:
            return True
        else:
            for base in cls.__bases__:
                if DocumentChecker.matches_qualname(base, qualname):
                    return True
        return False

    def check_document_property(self, name, value):
        checker = DocumentChecker.__lookup_checker(self.__class__, name)
        if checker:
            checker(self, value)

    def check_policies_property(self, name, value):
        checker = DocumentChecker.__lookup_policy_checker(self.__class__, name)
        if checker:
            checker(self, value)

    def check_doc(self, **doc):
        self.check_patch_doc(**doc)

        for cls, prop in self.required:
            if self.matches_qualname(self.__class__, cls):
                if prop not in doc:
                    raise ValueError(
                        "Missing value for required property {}".format(prop)
                    )

        policies = doc.get("policies")
        if policies:
            for cls, policy in self.required_policies:
                if self.matches_qualname(self.__class__, cls):
                    if policy not in policies:
                        raise ValueError(
                            "Missing value for required policy property {}".format(
                                policy
                            )
                        )

    def check_patch_doc(self, **doc):
        for prop, val in doc.items():
            self.check_document_property(prop, val)

        policies = doc.get("policies")
        if policies:
            for prop, val in policies.items():
                if prop == "DATA_TYPE":
                    if val not in self.POLICY_DATA_TYPES:
                        raise ValueError(
                            "Unsupported DATA_TYPE. "
                            f"Expected {self.POLICY_DATA_TYPE} "
                            f"but received {val}"
                        )

                self.check_policies_property(prop, val)


def checks_property(
    fn: t.Optional[t.Callable] = None, prop: str | None = None, required: bool = False
):
    """
    Decorate a method as a value checker. This registers a function to be
    called by the DocumentChecker.check_doc() method when the given property is
    available in a put or patch document.
    """
    if fn is None:
        return functools.partial(checks_property, prop=prop, required=required)

    assert prop is not None

    class_qualname = fn.__qualname__.rsplit(".", 1)[0]
    key = (class_qualname, prop)

    @functools.wraps(fn)
    def decorated_fn(*args, **kwargs):
        try:
            return fn(*args, **kwargs)
        except ValueError as e:
            if e.args:
                raise
            else:
                raise ValueError("Invalid value for the {} property".format(prop))
        except AttributeError as e:
            if e.args:
                raise
            else:
                raise AttributeError("The {} property is immutable.".format(prop))

    DocumentChecker.checkers[key] = decorated_fn

    if required:
        DocumentChecker.required.add(key)

    return decorated_fn


def checks_policy_property(
    fn: t.Optional[t.Callable] = None, prop: str | None = None, required: bool = False
):
    """
    Decorate a method as a value checker for the policy sub-document. This
    registers a function to be called by the DocumentChecker.check_doc() method
    when the given property is available in the policies sub-document of the
    put or patch document.
    """
    if fn is None:
        return functools.partial(checks_policy_property, prop=prop, required=required)

    assert prop is not None

    class_qualname = fn.__qualname__.rsplit(".", 1)[0]
    key = (class_qualname, prop)

    @functools.wraps(fn)
    def decorated_fn(*args, **kwargs):
        try:
            return fn(*args, **kwargs)
        except ValueError as e:
            if e.args:
                raise
            else:
                raise ValueError("Invalid value for the {} property".format(prop))
        except AttributeError as e:
            if e.args:
                raise
            else:
                raise AttributeError("The {} property is immutable.".format(prop))

    DocumentChecker.policy_checkers[key] = decorated_fn

    if required:
        DocumentChecker.required_policies.add(key)

    return decorated_fn
