from __future__ import annotations

import datetime
import typing as t
from collections.abc import Mapping

from flask import current_app
from OpenSSL import crypto

import globus.manager.api.errors as errors
import globus.manager.api.models as m
from globus.manager.api.schemas.api.domain import DomainDict
from globus.manager.api.schemas.appsync import Certificate as certificate_schema
from globus.manager.constants import ConfigurationObjectTag
from globus.manager.database import db
from globus.manager.utils.run_as_user import run_as_user

from .configuration_object import ConfigurationObject


class LetsEncryptCertificate(ConfigurationObject):
    """
    Model for a managed Let's Encrypt certificate for a wildcard domain. This
    uses the Domain model to handle writing the actual certificate and key to
    the filesystem so that apache/gridftp can find it.
    """

    _DATA_TYPE = "certificate#1.1.0"
    __mapper_args__ = {"polymorphic_identity": ConfigurationObjectTag.CERTIFICATE.name}
    PROPERTIES = {
        "certificate#1.0.0": ["domain_name", "certificate", "key", "chain"],
        "certificate#1.1.0": ["domain_name", "certificate", "key", "chain", "domain"],
    }
    configuration_schema = certificate_schema
    _x509_certificate: t.Optional[crypto.X509] = None
    _domain: t.Optional[m.Domain] = None

    def __init__(self, appsync_data=None, doc=None):
        if appsync_data is not None:
            args = {k: v for k, v in appsync_data.items()}
            args["tag"] = ConfigurationObjectTag.CERTIFICATE
            args["id"] = ConfigurationObjectTag.CERTIFICATE.name
            super().__init__(**args)
        elif doc is not None and (doc.get("DATA_TYPE") == self._DATA_TYPE):
            super().__init__(
                tag=ConfigurationObjectTag.CERTIFICATE,
                id=ConfigurationObjectTag.CERTIFICATE.name,
            )
            self._put(doc)
        else:
            raise ValueError("Unsupported DATA_TYPE")

    def _put(self, doc):
        for k, v in doc.items():
            if k in self.PROPERTIES[self._DATA_TYPE]:
                if k != "domain":
                    setattr(self, k, v)
        self._domain = None
        self._x509_certificate = None

    @property
    def domain_name(self) -> str:
        return self.configuration_hash["domain_name"]

    @domain_name.setter
    def domain_name(self, value: str) -> None:
        if value and value != self.configuration_hash.get("domain_name"):
            self.configuration_hash["domain_name"] = value
            if self._domain:
                self._domain.domain_name = value

    @property
    def certificate(self) -> str:
        """
        Text representation of the X.509 certificate in PEM format.
        """
        return self.configuration_hash["certificate"]

    @certificate.setter
    def certificate(self, value: str) -> None:
        if value and value != self.configuration_hash.get("certificate"):
            self.configuration_hash["certificate"] = value
            if self._domain:
                self._domain.certificate = value
            self._x509_certificate = None

    @property
    def key(self) -> str:
        """
        Text representation of the private key for this certificate in
        PEM format.
        """
        return self.configuration_hash["key"]

    @key.setter
    def key(self, value: str) -> None:
        if value and value != self.configuration_hash.get("key"):
            self.configuration_hash["key"] = value
            if self._domain:
                self._domain.private_key = value

    @property
    def chain(self) -> str:
        """
        Text representation of the certificate chain needed to validate
        this certificate.
        """
        chain = self.configuration_hash.get("chain")
        if chain:
            return chain
        raise AttributeError("The chain property is not defined for this object.")

    @chain.setter
    def chain(self, value: str) -> None:
        if value:
            stripped = value.strip()
            if stripped and stripped != self.configuration_hash.get("chain"):
                self.configuration_hash["chain"] = stripped
                if self._domain:
                    self._domain.certificate_chain = stripped

    @property
    def domain(self) -> m.Domain:
        """
        Details about the certificate's domain, including paths to the
        certificates and keys on the data transfer node.
        """
        if self._domain is None:
            doc: DomainDict = {
                "DATA_TYPE": "domain#1.0.0",
                "wildcard": True,
                "domain_name": self.domain_name,
                "certificate": self.certificate,
                "certificate_chain": self.chain,
                "private_key": self.key,
            }
            self._domain = m.Domain(doc=doc)
        return self._domain

    @domain.setter
    def domain(self, value: t.Optional[DomainDict | m.Domain]) -> None:
        if isinstance(value, m.Domain):
            self._domain = value
        elif isinstance(value, Mapping):
            self._domain = m.Domain(appsync_data=value)
        else:
            self._domain = None
        if self._domain:
            self.domain_name = self._domain.domain_name
            self.certificate = self._domain.certificate
            self.chain = self._domain.certificate_chain
            self.key = self._domain.private_key

    @property
    def expiration_time(self) -> datetime.datetime:
        """
        Expiration time as an zone-aware datetime object
        """
        cert = self.x509_certificate
        ts_bytes = cert.get_notAfter()
        assert ts_bytes is not None
        ts = ts_bytes.decode()

        if ts.endswith("Z"):
            ts = ts[:-1] + "+0000"

        return datetime.datetime.strptime(ts, "%Y%m%d%H%M%S%z")

    @property
    def x509_certificate(self) -> crypto.X509:
        """
        OpenSSL X.509 representation of this certificate
        """
        if not self._x509_certificate:
            self._x509_certificate = crypto.load_certificate(
                crypto.FILETYPE_PEM, self.certificate.encode("ascii")
            )
        return self._x509_certificate

    @staticmethod
    def create_certificate(domain):
        acme_account = m.ConfigurationObject.query.filter(
            m.ConfigurationObject.tag == ConfigurationObjectTag.ACME_ACCOUNT
        ).one()

        return LetsEncryptCertificate(
            doc=dict(
                DATA_TYPE=LetsEncryptCertificate._DATA_TYPE,
                **acme_account.get_cert_chain_and_key(domain)._asdict(),
            )
        )

    def renew(self):
        acme_account = m.ConfigurationObject.query.filter(
            m.ConfigurationObject.tag == ConfigurationObjectTag.ACME_ACCOUNT
        ).one()

        try:
            res = acme_account.get_cert_chain_and_key(self.domain_name)
            self.domain = {
                "DATA_TYPE": "domain#1.0.0",
                "wildcard": True,
                "domain_name": self.domain_name,
                "certificate": res.certificate,
                "certificate_chain": res.chain,
                "private_key": res.key,
            }
        except errors.APIBaseError as e:
            current_app.logger.error(f"Error renewing certificate {e}")
            raise
        except ValueError as e:
            current_app.logger.error(f"Error renewing certificate {e}")
            raise
        except Exception as e:
            current_app.logger.error(f"Error renewing certificate {e}")
            raise errors.BadGatewayError("Unable to renew certificate")

    @property
    def files_exist(self):
        return self.domain.files_exist

    @property
    def certificate_path(self):
        return self.domain.certificate_path

    @property
    def certificate_chain_path(self):
        return self.domain.certificate_chain_path

    @property
    def private_key_path(self):
        return self.domain.private_key_path

    @run_as_user(user="gcsweb")
    def write_config(self, delete=False):
        return self.domain.write_config(delete=delete)

    def rollback(self, data=None):
        """
        Rolls back state to a previous version of the configuration_hash.
        """
        if data:
            self.update(data)
        else:
            db.session.refresh(self)

    def __repr__(self):
        return "<LetsEncryptCertificate {}>".format(self.domain_name)
