from __future__ import annotations

import time
import typing as t
from urllib.parse import urlsplit

from flask import current_app

import globus_sdk
from globus_sdk.exc import GlobusAPIError

import globus.manager.api.errors as errors
from globus.manager.api.schemas.appsync.oidc_registration import OIDCRegistrationDict
from globus.manager.api.utils.log import log_auth_exception


class OIDCRegistration:
    """
    This class is the model for managing an external OIDC server registration
    for an endpoint. It includes methods for interacting with Globus Auth to
    register the OIDC service and delete that registration.
    """

    _DATA_TYPES = ["oidc_registration#1.0.0"]
    _DATA_TYPE = _DATA_TYPES[0]
    OIDC_REGISTRATION_PROPERTIES = {
        "oidc_registration#1.0.0": [
            "auth_client_id",
            "environment",
            "display_name",
            "discovery_url",
            "domain_name",
            "domains",
            "support_contact",
            "support_email",
            "username_claim",
            "id_claim",
            "oidc_client_id",
            "oidc_client_secret",
            "deleted",
        ]
    }
    IMMUTABLE = [
        "auth_client_id",
        "environment",
        "discovery_url",
        "domain_name",
        "domains",
        "deleted",
    ]

    def __init__(
        self,
        appsync_data: t.Optional[OIDCRegistrationDict] = None,
        doc: t.Optional[OIDCRegistrationDict] = None,
    ):
        """
        The constructor can be called with either the appsync_data or
        the doc options set. In either case, the value should be an
        OIDCRegistrationDict containing the configuration parameters. When
        passed as the appsync_data option, this class assumes that
        all data has already been validated and treats this more like
        an object restore from the database. When passed as the doc
        option, this method validates the document contents, and creates
        resources as needed to fill in the configuration.
        """
        self.auth_client_id: t.Optional[str] = None
        self.environment: t.Optional[str] = None
        self.display_name: t.Optional[str] = None
        self.discovery_url: t.Optional[str] = None
        self.domains: t.Optional[list[str]] = None
        self.support_contact: t.Optional[str] = None
        self.support_email: t.Optional[str] = None
        self.username_claim: t.Optional[str] = None
        self.id_claim: t.Optional[str] = None
        self.oidc_client_id: t.Optional[str] = None
        self.oidc_client_secret: t.Optional[str] = None
        self.domain_name: t.Optional[str] = None
        self.deleted = False

        self._auth_client_data: t.Mapping[str, t.Any] = {}
        self._auth_identity_provider_data: t.Mapping[str, t.Any] = {}

        if appsync_data is not None:
            self._put(**appsync_data)
        elif doc is not None:
            self.put(**doc)

            if self.deleted:
                return

            if self.environment is None:
                self.environment = current_app.environment

            if self.domain_name is None and self.discovery_url is not None:
                self.domain_name = urlsplit(self.discovery_url).netloc

            if self.auth_client_id is None:
                self._create_child_client()

        else:
            raise errors.InternalServerError("Invalid constructor call")

    @property
    def auth_client_data(self) -> t.Mapping[str, t.Any]:
        """
        The auth_client_data property contains a mapping with client document from
        auth for the client_id associated with this OIDCRegistration object.
        """
        if not self.auth_client_id:
            raise AttributeError(
                "The auth_client_data property is not defined for this OIDCRegistration"
            )

        if not self._auth_client_data:
            try:
                auth_response = current_app.auth_client.get(
                    "/v2/api/clients/{}".format(self.auth_client_id)
                )
                current_app.logger.debug(
                    "Get client from auth returned {0}".format(
                        auth_response.http_status
                    )
                )
                if not auth_response.data.get("client", {}).get("id"):
                    raise errors.BadGatewayError(
                        "Error retrieving Globus Auth client. Please check "
                        "that the credentials are correct."
                    )

                self._auth_client_data = auth_response.data["client"]
                current_app.logger.debug(
                    "Retrieved client with ID %s", self.auth_client_id
                )
            except GlobusAPIError as e:
                msg = (
                    "Error retrieving Globus Auth client: {0} {1}. Please check "
                    "that the credentials are correct and Globus Auth is accessible."
                ).format(e.http_status, str(e))
                raise Exception(msg)
        return self._auth_client_data

    @property
    def auth_identity_provider_data(self) -> t.Mapping[str, t.Any]:
        """
        The auth_client_data property contains a mapping with identity_provider
        document from auth for the domain_name used by this OIDCRegistration object.
        """
        if not self._auth_identity_provider_data:
            try:
                resp = current_app.auth_client.get(
                    "/v2/api/identity_providers",
                    query_params={"domains": [self.domain_name]},
                )
                data: list[dict[str, t.Any]] = resp.data["identity_providers"]

                if self._check_auth_identity_provider_exists(data):
                    self._auth_identity_provider_data = data[0]
                else:
                    raise AttributeError(
                        "The auth_identity_provider_data property is not "
                        "set on this OIDCRegistration"
                    )
            except GlobusAPIError as e:
                log_auth_exception(e)

                raise AttributeError(
                    "The auth_identity_provider_data property is not "
                    "set on this OIDCRegistration"
                )
        return self._auth_identity_provider_data

    def _check_auth_identity_provider_exists(
        self, data: list[dict[str, t.Any]]
    ) -> bool:
        """
        The following scenarios are identically treated:
            1) Domain has no IdP
            2) Domain has the generic_email IdP
        """
        if not data:
            return False

        short_name = data[0].get("short_name")
        if short_name == "generic_email":
            return False
        return True

    def _put(self, DATA_TYPE: str | None = None, **doc) -> None:
        """
        The _put method updates the contents of this object with a new
        document. This method does not check the document for validity, and
        ignores unknown properties.

        :param DATA_TYPE:
            The DATA_TYPE parameter is the versioned document data type that
            is being processed. If this is unknown, the object will use process
            this as the newest document type it understands.
        :param **doc:
            The doc parameter contains the rest of the configuration document.
        """
        properties = (
            self.OIDC_REGISTRATION_PROPERTIES.get(DATA_TYPE) if DATA_TYPE else None
        )
        if properties is None:
            properties = self.OIDC_REGISTRATION_PROPERTIES[self._DATA_TYPE]

        for prop in properties:
            if prop in doc:
                setattr(self, prop, doc[prop])

    def put(self, DATA_TYPE: str | None = None, **doc) -> None:
        """
        The put method updates the contents of this object with a new
        document. This method does checking of the document for validity and
        will only process documents of a known data type.

        :param DATA_TYPE:
            The DATA_TYPE parameter is the versioned document data type that
            is being processed. If this is unknown, the object will use process
            this as the newest document type it understands.
        :param **doc:
            The doc parameter contains the rest of the configuration document.
        """
        if DATA_TYPE not in self._DATA_TYPES:
            raise ValueError(
                "Unsupported DATA_TYPE for OIDCRegistration: {}".format(DATA_TYPE)
            )
        for prop in self.IMMUTABLE:
            if prop in doc and getattr(self, prop) and doc[prop] != getattr(self, prop):
                raise errors.ValidationError(
                    "The {} property is immutable".format(prop)
                )
        self._put(DATA_TYPE, **doc)

    def _patch(self, DATA_TYPE: str | None = None, **doc) -> None:
        """
        The _patch method updates the contents of this object with a change
        document.  This method does not check the document for validity, and
        ignores unknown properties.

        :param DATA_TYPE:
            The DATA_TYPE parameter is the versioned document data type that
            is being processed. If this is unknown, the object will use process
            this as the newest document type it understands.
        :param **doc:
            The doc parameter contains the rest of the configuration document.
        """
        properties = (
            self.OIDC_REGISTRATION_PROPERTIES.get(DATA_TYPE) if DATA_TYPE else None
        )
        if properties is None:
            properties = self.OIDC_REGISTRATION_PROPERTIES[self._DATA_TYPE]

        for prop in properties:
            if prop in doc:
                setattr(self, prop, doc[prop])

    def patch(self, DATA_TYPE: str | None = None, **doc) -> None:
        """
        The patch method updates the contents of this object with a new
        document. This method does checking of the document for validity and
        will only process documents of a known data type. Any items missing
        in the input document will retain their previous values.

        :param DATA_TYPE:
            The DATA_TYPE parameter is the versioned document data type that
            is being processed. If this is unknown, the object will use process
            this as the newest document type it understands.
        :param **doc:
            The doc parameter contains the rest of the configuration document.
        """
        if DATA_TYPE not in self._DATA_TYPES:
            raise errors.ValidationError("Unsupported DATA_TYPE {}".format(DATA_TYPE))

        for prop in self.IMMUTABLE:
            if prop in doc and getattr(self, prop) and doc[prop] != getattr(self, prop):
                raise errors.ValidationError(
                    "The {} property is immutable".format(prop)
                )

        if "display_name" in doc:
            request_body: dict[str, t.Any] = {"client": {"name": doc["display_name"]}}
            try:
                auth_response = current_app.auth_client.put(
                    "/v2/api/clients/{}".format(self.auth_client_id), data=request_body
                )
                self._auth_client_data = auth_response["client"]
            except GlobusAPIError as e:
                log_auth_exception(e)
                raise errors.BadGatewayError(
                    "Unable to update display_name: {}".format(str(e))
                )

        self._patch(DATA_TYPE, **doc)

    def _create_child_client(self) -> None:
        """
        Create a new auth client for this service as a child of the Endpoint's
        auth client.
        """
        current_app.logger.debug("Creating Globus Auth identity....")

        request_body: dict[str, t.Any] = {
            "client": {
                "name": self.display_name,
                "public_client": False,
                "redirect_uris": [],
                "visibility": "private",
                "required_idp": None,
                "preselect_idp": None,
            }
        }
        try:
            auth_response = current_app.auth_client.post(
                "/v2/api/clients", data=request_body
            )
            current_app.logger.debug(
                "Created client with auth returned {0}".format(
                    auth_response.http_status
                )
            )
            assert auth_response.http_status == 201
            assert auth_response.data["client"] and auth_response.data["client"]["id"]
            self._auth_client_data = auth_response.data["client"]
            self.auth_client_id = self._auth_client_data["id"]
            current_app.logger.debug("Created client with ID %s", self.auth_client_id)
        except GlobusAPIError as e:
            msg = (
                "Error creating Globus Auth client: {0} {1}. Please check "
                + "that the credentials are correct and Globus Auth is accessible."
            ).format(e.http_status, str(e))
            raise errors.BadGatewayError(msg)

    def _register_fqdn(self, fqdn) -> None:
        """
        Register a domain that the OIDC registration asserts with this service's
        Globus Auth client_id.
        """
        current_app.logger.debug("Registering FQDN {} with Auth".format(fqdn))

        # FQDN isn't registered yet with Auth - register it
        RETRIES = 10
        for attempt in range(1, RETRIES):
            try:
                current_app.logger.debug(
                    "Attempting to register FQDN {} with Globus Auth client {} "
                    "using parent client ID {}...".format(
                        fqdn,
                        self.auth_client_id,
                        current_app.client_id,
                    )
                )
                auth_response = current_app.auth_client.post(
                    "/v2/api/clients/{client_id}/fqdns".format(
                        client_id=self.auth_client_id
                    ),
                    data={"fqdn": fqdn},
                )
                current_app.logger.debug(
                    "Registered dns with auth returned {0}".format(
                        auth_response.http_status
                    )
                )
                assert auth_response.http_status in [200, 201]
                break
            except GlobusAPIError as e:
                msg = (
                    "Error registering FQDN with Globus Auth: {} {}. Please check "
                    "that a DNS TXT record exists at {} with value '{}'."
                ).format(
                    e.http_status,
                    str(e),
                    fqdn,
                    current_app.auth_client.client_id,
                )
                current_app.logger.info(msg)

                # Auth should probably send a 400 in the domain already
                # registered case
                if e.http_status > 500:
                    raise errors.InternalServerError(msg)
                if attempt + 1 < RETRIES:
                    current_app.logger.debug(str(e))
                    time.sleep(20)
                else:
                    raise errors.InternalServerError(msg)

    def register_idp(self) -> str:
        """
        Register the IdP with Globus Auth. This creates a new identity provider
        for this OIDC Server's domain name. This requires the OIDC service to
        be running enough to serve the OIDC autodiscovery URL.
        """
        # register fqdns with auth_client_id
        fqdns = [str(self.domain_name)] + (self.domains or [])
        registered = self.auth_client_data.get("fqdns", [])
        for fqdn in set(fqdns) - set(registered):
            self._register_fqdn(fqdn)

        # return idp if it already exists
        try:
            auth_identity_provider_data = self.auth_identity_provider_data

            if auth_identity_provider_data:
                return auth_identity_provider_data["id"]
        except AttributeError:
            pass

        # register idp
        params = {
            "identity_provider": {
                "name": self.display_name,
                "discovery_url": self.discovery_url,
                "domains": self.domains or [],
                "client_credentials": {
                    "client_id": self.oidc_client_id,
                    "client_secret": self.oidc_client_secret,
                },
                "attribute_map": {
                    "username": self.username_claim,
                    "immutable_id": self.id_claim,
                },
                "contact_email": self.support_email,
                "contact_name": self.support_contact,
            }
        }

        assert self.auth_client_id is not None
        auth_child_client = globus_sdk.ConfidentialAppAuthClient(
            self.auth_client_id,
            current_app.client_secret,
            environment=current_app.environment,
        )
        try:
            resp = auth_child_client.post("/v2/api/identity_providers", data=params)
            if resp.http_status in [200, 201]:
                current_app.logger.debug(resp)
                self._auth_identity_provider_data = resp["identity_provider"]
            else:
                raise errors.BadGatewayError(
                    "Unable to create IdP, Globus Auth returned status code {}.".format(
                        resp.http_status
                    )
                )
        except GlobusAPIError as e:
            log_auth_exception(e)
            raise errors.BadGatewayError(
                "Unable to create IdP, Globus Auth returned "
                "status code {}: {}.".format(e.http_status, str(e))
            )
        return resp["identity_provider"]["id"]

    def update_idp(self) -> None:
        """
        Update the IdP registration with Globus Auth.
        """
        auth_identity_provider_data = self.auth_identity_provider_data

        params = {
            "identity_provider": {
                "name": self.display_name,
                "client_credentials": {
                    "client_id": self.oidc_client_id,
                    "client_secret": self.oidc_client_secret,
                },
                "contact_email": self.support_email,
                "contact_name": self.support_contact,
            }
        }

        assert self.auth_client_id is not None
        auth_child_client = globus_sdk.ConfidentialAppAuthClient(
            self.auth_client_id,
            current_app.client_secret,
            environment=current_app.environment,
        )
        try:
            resp = auth_child_client.put(
                "/v2/api/identity_providers/{}".format(
                    auth_identity_provider_data["id"]
                ),
                data=params,
            )
            if resp.http_status in [200, 201]:
                current_app.logger.debug(resp)
                self._auth_identity_provider_data = resp["identity_provider"]
            else:
                raise errors.BadGatewayError(
                    "Unable to update IdP, Globus Auth returned status code {}.".format(
                        resp.http_status
                    )
                )
        except GlobusAPIError as e:
            log_auth_exception(e)
            raise errors.BadGatewayError(
                "Unable to update IdP, Globus Auth returned "
                "status code {}: {}.".format(e.http_status, str(e))
            )
        return

    def delete(self):
        """
        Delete the Auth resources associated with this OIDC server registration.
        """
        # Attempt to delete the idp if it was created, but continue if this
        # client doesn't have permission.
        try:
            idp = self.auth_identity_provider_data
            try:
                auth_child_client = globus_sdk.ConfidentialAppAuthClient(
                    self.auth_client_id,
                    current_app.client_secret,
                    environment=current_app.environment,
                )
                auth_child_client.delete(
                    "/v2/api/identity_providers/{}".format(idp["id"])
                )
            except GlobusAPIError as e:
                current_app.logger.info(
                    "Unable to delete identity provider. Globus Auth returned {}.".format(
                        str(e)
                    )
                )
                if e.http_status != 403:
                    raise e
        except AttributeError:
            pass

        # delete the child client.  *this does not delete the idp*
        if self.auth_client_id:
            current_app.auth_client.delete(
                "/v2/api/clients/{}".format(self.auth_client_id)
            )
            self._auth_client_data = {}
            self._auth_identity_provider_data = {}
        self.deleted = True

    def __repr__(self):
        return "<OIDCRegistration {}>".format(self.auth_client_id)


__all__ = ["OIDCRegistration"]
