import typing as t
import uuid

import marshmallow as ma

from .data_type import Schema, data_type, data_type_schema


@data_type("endpoint#1.0.0")
class Endpoint_1_0_0(Schema):
    __doc__ = """
    A Globus Connect Server endpoint is a deployment of Globus Connect Server
    version 5. A single endpoint may optionally include multiple data transfer
    nodes. The endpoint provides a link between a Globus Connect Server
    deployment and the Globus Transfer service. The endpoint describes services
    for accessing data via GridFTP and HTTPS and also for configuring and
    managing the policies associated with that access.
    """

    public = ma.fields.Boolean(
        default=True,
        description="""
        Flag indicating whether this endpoint is visible to all other
        Globus users. If false, only users which have been granted a role
        on the endpoint or one of its collections, or belong to a domain
        allowed access to any of its storage gateways may view it.
        """,
    )
    endpoint_id = ma.fields.UUID(
        data_key="id", description="Unique identifier for this endpoint"
    )
    display_name = ma.fields.Str(
        required=True,
        validate=[ma.validate.Length(max=128)],
        description="""
        Friendly name for the endpoint, not unique.  Unicode string, no new
        lines (`\\r` or `\\n`).  Searchable.
        """,
    )

    gcs_manager_url = ma.fields.Url(
        dump_only=True,
        description="URL of the GCS Manager API service for this endpoint",
    )
    subscription_id = ma.fields.Str(
        allow_none=True,
        description="""
        The id of the subscription that is managing this endpoint.  This may be
        the special value `DEFAULT` when using this as input to PATCH or PUT to
        use the caller's default subscription id.
        """,
    )
    network_use = ma.fields.String(
        default="normal",
        validate=[ma.validate.OneOf(["normal", "minimal", "aggressive", "custom"])],
        description="""
        Control how Globus interacts with this endpoint over the network.

        Allowed values for **network_use** are:

        * `normal`
            - The default setting. Uses an average level of concurrency and
              parallelism. The levels depend on the number of physical
              servers in the endpoint.
        * `minimal`
            - Uses a minimal level of concurrency and parallelism.
        * `aggressive`
            - Uses a high level of concurrency and parallelism.
        * `custom`
            - Uses custom values of concurrency and parallelism set by the
              endpoint admin. When setting this level, you must also set
              the **max_concurrency**, **preferred_concurrency**,
              **max_parallelism**, and **preferred_parallelism** properties.
        """,
    )
    max_concurrency = ma.fields.Int(
        validate=[ma.validate.Range(min=1)],
        description="""
        Admin-specified value when the **network_use** property's value is
        `custom`; otherwise the preset value for the specified
        **network_use**.
        """,
    )
    preferred_concurrency = ma.fields.Int(
        validate=[ma.validate.Range(min=1)],
        description="""
        Admin-specified value when the **network_use** property's value is
        `custom`; otherwise the preset value for the specified
        **network_use**.
        """,
    )

    max_parallelism = ma.fields.Int(
        validate=[ma.validate.Range(min=1)],
        description="""
        Admin-specified value when the **network_use** property's value is
        `custom`; otherwise the preset value for the specified
        **network_use**.
        """,
    )
    preferred_parallelism = ma.fields.Int(
        validate=[ma.validate.Range(min=1)],
        description="""
        Admin-specified value when the **network_use** property's value is
        `custom`; otherwise the preset value for the specified
        **network_use**.
        """,
    )

    organization = ma.fields.String(
        validate=[ma.validate.Length(max=1024)],
        description="""
        Organization that runs the server(s) represented by the endpoint.
        Unicode string, max 1024 characters, no new lines.
        """,
    )
    department = ma.fields.String(
        validate=[ma.validate.Length(max=1024)],
        description="""
        Department within organization that runs the server(s).
        Searchable. Unicode string, max 1024 characters, no new lines.
        """,
    )
    keywords = ma.fields.List(
        ma.fields.String(),
        description="""
        List of search keywords for the endpoint. Unicode
        string, max 1024 characters total across all strings.
        """,
    )
    description = ma.fields.Str(
        validate=[ma.validate.Length(max=4096)],
        description="A description of the endpoint",
    )
    contact_email = ma.fields.Email(
        description=(
            "Email address of the support contact for this endpoint. This is visible "
            + "to end users so that they may contact your organization for support."
        ),
    )
    contact_info = ma.fields.Str(
        validate=[ma.validate.Length(max=4096)],
        description=(
            "Other non-email contact information for the endpoint, e.g.  phone and "
            + "mailing address. This is visible to end users for support."
        ),
    )
    info_link = ma.fields.Url(
        description="""
        Link to a web page with more information about the endpoint. The
        administrator is responsible for running a website at this URL and
        verifying that it is accepting public connections.
        """
    )
    allow_udt = ma.fields.Boolean(
        description="Allow data transfer on this endpoint using the UDT protocol"
    )

    @ma.validates("subscription_id")
    def validate_subscription_id(self, data, **kwargs):
        if data is None or data.lower() == "default":
            return
        elif data:
            try:
                uuid.UUID(data)
            except (ValueError, AttributeError):
                raise ma.ValidationError("must be uuid or DEFAULT")

    @ma.post_load
    def clean_uuids(self, data, **kwargs):
        res = data.copy()
        if "id" in res and isinstance(res["id"], uuid.UUID):
            res["id"] = str(res["id"])
        if "subscription_id" in res and isinstance(res["subscription_id"], uuid.UUID):
            res["subscription_id"] = str(res["subscription_id"])
        return res

    @ma.post_load
    def clean_display_name(self, data, **kwargs):
        res = data.copy()
        if "display_name" in res:
            res["display_name"] = res["display_name"].strip()
        return res

    @ma.post_load
    def clean_keywords(self, data, **kwargs):
        res = data.copy()
        keywords = res.get("keywords")
        if keywords:
            res["keywords"] = [kw.strip() for kw in data["keywords"]]
        return res

    @ma.validates("display_name")
    def validate_display_name(self, data, **kwargs):
        if data and "\n" in data or "\r" in data:
            raise ma.ValidationError("Newline not allowed in display_name")

    @ma.validates("organization")
    def validate_organization(self, data, **kwargs):
        if data and ("\n" in data or "\r" in data):
            raise ma.ValidationError("Newline not allowed in organization")

    @ma.validates("department")
    def validate_department(self, data, **kwargs):
        if data and ("\n" in data or "\r" in data):
            raise ma.ValidationError("Newline not allowed in department")

    @ma.validates("keywords")
    def validate_keywords(self, data, **kwargs):
        if data:
            kw = ",".join(data)
            if len(kw) > 1024:
                raise ma.ValidationError("Longer than maximum length 1024.")

    @ma.validates_schema
    def validate_network_use(self, obj, **kwargs):
        custom_network_fields = [
            "max_concurrency",
            "preferred_concurrency",
            "max_parallelism",
            "preferred_parallelism",
        ]
        network_use = obj.get("network_use")
        if network_use is not None and network_use != "custom":
            for i in custom_network_fields:
                if obj.get(i) is not None:
                    raise ma.ValidationError(
                        "Invalid {} when network use is {}".format(i, network_use)
                    )


@data_type("endpoint#1.1.0")
class Endpoint_1_1_0(Endpoint_1_0_0):
    __doc__ = (
        t.cast(str, Endpoint_1_0_0.__doc__)
        + """
    Version 1.1.0 of the endpoint includes support for customizing the TCP port
    that the GridFTP listens on.
    """
    )

    gridftp_control_channel_port = ma.fields.Integer(
        min_inclusive=1,
        max_inclusive=65535,
        required=False,
        allow_none=True,
        description="""
        TCP port for the Globus control channel to listen on. By default,
        the control channel is passed through 443 with an ALPN header
        containing the value "ftp".
        """,
    )


@data_type("endpoint#1.2.0")
class Endpoint_1_2_0(Endpoint_1_1_0):
    __doc__ = (
        t.cast(str, Endpoint_1_1_0.__doc__)
        + """
    Version 1.2.0 of the endpoint includes read-only earliest_last_access
    to put a limit on collections which are missing a last_access value.
    """
    )

    earliest_last_access = ma.fields.Date(
        allow_none=False,
        dump_only=True,
        description="""
        Earliest date when this endpoint began tracking last_access for
        collections
        """,
    )


Endpoint = data_type_schema.Endpoint
