from typing import List

import marshmallow as ma
from typing_extensions import Literal, TypedDict

from .data_type import Schema, data_type, data_type_schema

_PATH_RE = "^[/~\\$\\*\\?\\[].*$"


class PathRestrictionsDict(TypedDict):
    DATA_TYPE: Literal["path_restrictions#1.0.0"]
    read: List[str]
    read_write: List[str]
    none: List[str]


@data_type("path_restrictions#1.0.0")
class PathRestrictions_1_0_0(Schema):
    """
    This object represents the path restrictions for a storage gateway
    or a sharing path restrictions for a mapped collection.

    The values of each of the path lists in this object are interpreted using
    the POSIX pattern matching notation as described in
    https://pubs.opengroup.org/onlinepubs/9699919799/functions/fnmatch.html[fnmatch(3)]
    with flags set to `0` with additional support for some special
    user-specific value interpolation:

    `~`, `$HOME`::

    The user's home directory if the storage gateway supports such a concept,
    `/` otherwise

    `$USER`::

    The effective Storage Gateway-specific username that is being used for data
    access. For a Guest Collection, this is the username of the identity that
    created the Guest Collection.

    These restrictions are evaluated at every data access. When
    evaluating restrictions, the user-specific interpolation is
    applied before the file name matching is evaluated.

    Globus Connect Server evaluates its path restrictions from
    longest leading expression match to shortest. When pattern
    matching characters are present, they are considered as a lower
    priority match than a literal character, with more specific
    pattern characters given precedence. The precedence is thus
    literal character, bracket expression, `?` (single-character
    wildcard), `*` (wildcard).

    If multiple path restrictions apply, all matches are applied
    from longest to shortest, with the following rules for
    overriding values:

    ### Path Restriction Override Precedence
    ++++
    <table>
      <tr>
        <th>longer restriction</th>
        <th>shorter restriction</th>
        <th>result</th>
      </tr>
    <tr>
    <td> <pre>read_write</pre>   </td><td> <pre>read</pre>          </td><td> <pre>read_write</pre></td>
    </tr>
    <tr>
    <td> <pre>read_write</pre>   </td><td> <pre>none</pre>          </td><td> <pre>read_write</pre></td>
    </tr>
    <tr>
    <td> <pre>read</pre>         </td><td> <pre>read_write</pre>    </td><td> <pre>read_write</pre></td>
    </tr>
    <tr>
    <td> <pre>read</pre>         </td><td> <pre>none</pre>          </td><td> <pre>read</pre></td>
    </tr>
    <tr>
    <td> <pre>none</pre>         </td><td> <pre>read_write</pre>    </td><td> <pre>none</pre></td>
    </tr>
    <tr>
    <td> <pre>none</pre>         </td><td> <pre>read</pre>          </td><td> <pre>none</pre></td>
    </tr>
    </table>
    ++++

    """

    read = ma.fields.List(
        ma.fields.String(validate=[ma.validate.Regexp(_PATH_RE)]),
        description="List of paths which are allowed read-only access",
    )
    read_write = ma.fields.List(
        ma.fields.String(validate=[ma.validate.Regexp(_PATH_RE)]),
        description="List of paths which are allowed read-write access",
    )
    none = ma.fields.List(
        ma.fields.String(validate=[ma.validate.Regexp(_PATH_RE)]),
        description="List of paths which are denied any access",
    )

    def load(self, *args, **kwargs) -> PathRestrictionsDict:
        return super().load(*args, **kwargs)


PathRestrictions = data_type_schema.PathRestrictions
