from typing import TYPE_CHECKING, Any, Optional, Union

if TYPE_CHECKING:
    import uuid

import globus_sdk

AuthScopes = globus_sdk.scopes.ScopeBuilder(
    "auth.globus.org",
    known_scopes=[
        "manage_projects",
    ],
)


class GetPolicyResponse(globus_sdk.response.GlobusHTTPResponse):
    @property
    def data(self) -> Any:
        return self._parsed_json["policy"]


class AuthClient(globus_sdk.ConfidentialAppAuthClient):
    """
    An Auth client that behaves exactly like a ConfidentialAppAuthClient except for API
    calls that require a bearer token.
    """

    def __init__(
        self,
        client_id: str,
        secret: str,
        *args: Any,
        environment: Optional[str],
        **kwargs: Any,
    ) -> None:
        # This will initialize all calls to Auth that require basic authorization
        if environment is not None:
            kwargs["environment"] = environment
        super().__init__(client_id, secret, *args, **kwargs)

        # Newer Auth API calls require a bearer token. Save off the environment setting
        self._environment = environment

    @property
    def bearer_auth_client(self) -> globus_sdk.AuthClient:
        if getattr(self, "_bearer_auth_client", None) is None:
            authorizer = globus_sdk.ClientCredentialsAuthorizer(
                self, AuthScopes.manage_projects
            )
            self._bearer_auth_client = globus_sdk.AuthClient(
                authorizer=authorizer, environment=self._environment
            )
        return self._bearer_auth_client

    def get_policy(
        self, policy_id: Union[str, "uuid.UUID"]
    ) -> globus_sdk.response.GlobusHTTPResponse:
        """
        GET /v2/api/policies/<policy_id>

        Get information about an authentication policy.

        This API call requires a bearer token, even when calling as a confidential client.
        """
        return GetPolicyResponse(
            self.bearer_auth_client.get(f"/v2/api/policies/{policy_id}")
        )
