import enum
from typing import Optional


class GraphQLError(Exception):
    """
    GraphQL can return a list of errors which contain at least a message,
    as well as potentially some query data returned
    """

    def __init__(self, errors, data) -> None:
        super().__init__(self, errors[0]["message"])
        self.message = errors[0]["message"]
        self.errors = errors
        self.error_type: Optional[str] = None
        if errors and errors[0].get("errorType"):
            self.error_type = errors[0]["errorType"]
        self.data = data


class GraphQLVariable:
    """
    GraphQL Variables provide a way to pass values to a query, mutation,
    or subscription outside of the query text, so the values can be
    directly serialized as json.
    """

    def __init__(self, name, valuetype, value=None):
        self.name = name
        self.valuetype = valuetype
        self._value = value

    def __str__(self):
        return "${}: {}".format(self.name, self.valuetype)

    @property
    def value(self):
        if isinstance(self._value, enum.Enum):
            return self._value.value
        else:
            return self._value

    def variable(self):
        return {self.name: self.value}


class GraphQLArgument:
    """
    Any GraphQL field selection may include arguments that are passed to
    the resolvers. The values of the arguments may be either variable
    references or literals. This version supports string, int, float, and
    boolean literals.
    """

    def __init__(self, name, value):
        self.name = name
        self.value = self._value_str(value)

    @staticmethod
    def _value_str(value):
        if isinstance(value, enum.Enum):
            value = value.value

        if isinstance(value, str):
            return '"{0}"'.format(value.replace("\\", "\\\\").replace('"', '\\"'))
        elif isinstance(value, bool) and value:
            return "true"
        elif isinstance(value, bool) and not value:
            return "false"
        elif isinstance(value, GraphQLVariable):
            return "${0}".format(value.name)
        elif isinstance(value, int) or isinstance(value, float):
            return "{0}".format(value)
        elif value is None:
            return "null"
        elif isinstance(value, list):
            vals = ", ".join(
                list([GraphQLArgument._value_str(listel) for listel in value])
            )
            return "[{}]".format(vals)
        else:
            raise GraphQLError([{"message": "Unsupported data type"}], None)

    def __str__(self):
        return "{0}: {1}".format(self.name, self.value)


class GraphQLVariables:
    """
    This class holds a list of GraphQLVariable objects and can serialize
    them as a variable list string or generate an object containing
    variable-value mappings
    """

    def __init__(self, *args):
        self._variables = list(*args)

    def __str__(self):
        if self._variables:
            return "({})".format(", ".join([str(v) for v in self._variables]))
        else:
            return ""

    def variables(self):
        return {v.name: v.value for v in self._variables}


class GraphQLQueryType(enum.Enum):
    QUERY = "query"
    MUTATION = "mutation"
    SUBSCRIPTION = "subscription"


class GraphQLSelection:
    def __init__(self, name, arguments=None, selections=None):
        self.name = name
        self.arguments = arguments
        self.selections = []
        if selections is not None:
            for s in selections:
                if isinstance(s, GraphQLSelection):
                    self.selections.append(s)
                else:
                    self.selections.append(GraphQLSelection(s))

    def __str__(self):
        if self.arguments:
            argument_string = ", ".join([str(a) for a in self.arguments])
            argument_string = "({})".format(argument_string)
        else:
            argument_string = ""
        if len(self.selections) > 0:
            selection_string = " {{ {} }}".format(
                " ".join([str(s) for s in self.selections])
            )
        else:
            selection_string = ""

        return self.name + argument_string + selection_string


class GraphQLQuery:
    def __init__(self, query_type, selection, variables=None):
        if not isinstance(query_type, GraphQLQueryType):
            query_type = GraphQLQueryType(query_type)
        if variables is not None and not isinstance(variables, GraphQLVariables):
            variables = GraphQLVariables(variables)
        self.query_type = query_type
        self.query_body = selection
        self.variables = variables

    def __str__(self):
        if self.variables:
            var_string = " " + str(self.variables) + " "
        else:
            var_string = " "

        return "{}{}{{ {} }}".format(
            self.query_type.value, var_string, str(self.query_body)
        )
