import os
import textwrap
import typing as t

import click
from flask import current_app

import globus_sdk
from gcs_cli.config import internal_auth_client

import globus.manager.api.models as models
from globus.manager.api.app import make_app
from globus.manager.appsync.query import GraphQLError
from globus.manager.constants import ConfigurationObjectTag
from globus.manager.utils.keychain import NodeKey


def endpoint_cleanup(
    ctx: click.Context,
    client_id: str,
    secret: str,
    deployment_node_key: t.Optional[NodeKey],
    deployment_key_path: str,
    development_appsync: t.Optional[t.Tuple[str, str]],
    skip_delete_registration: bool,
    lost_deployment_node_key: bool,
    agree_to_delete_endpoint: bool,
) -> None:
    app = make_app(app_config="globus.manager.api.config.endpoint_cli")
    models.db.init_app(app)
    environment = os.environ.get("GLOBUS_SDK_ENVIRONMENT", "production")

    with app.app_context():
        models.db.create_all()
        models.db.session.commit()
        if deployment_node_key:
            click.echo(f"Loaded deployment node key {deployment_node_key.thumbprint()}")
            app.config["DEPLOYMENT_KEY"] = deployment_node_key
        current_app.config["CLIENT_ID"] = client_id
        current_app.config["CLIENT_SECRET"] = secret
        current_app.config["ENVIRONMENT"] = environment
        if development_appsync:
            current_app.config["GCS_APPSYNC_URL"] = development_appsync[0]
            current_app.config["GCS_APPSYNC_CLIENT_ID"] = development_appsync[1]
            current_app.config["ENVIRONMENT"] = environment

        if not deployment_node_key and not lost_deployment_node_key:
            click.echo(
                "\n".join(
                    textwrap.wrap(
                        "Your deployment key does not provide a valid encryption key. "
                        "This may happen if you lose the original deployment key "
                        "created by endpoint setup. "
                        "If you still have access to the original deployment key, "
                        "please rerun this command with that deployment key. "
                        "Otherwise, not all resources can be deleted."
                    )
                )
            )
            if click.confirm("Do you still have access to your deployment key?"):
                ctx.exit(0)

        click.echo(
            "\n".join(
                textwrap.wrap(
                    "This command will completely remove all configuration "
                    "information associated with this endpoint. This is "
                    "irreversible, and the client_id used to create the "
                    "endpoint may not be reused to recreate this endpoint.",
                    width=78,
                )
            )
        )

        if not agree_to_delete_endpoint and not click.confirm(
            "Do you wish to continue?"
        ):
            ctx.exit(1)

        command = EndpointDelete(
            ctx,
            client_id,
            secret,
            deployment_node_key,
            deployment_key_path,
            environment,
            skip_delete_registration,
        )
        command.run()


class EndpointDelete:
    def __init__(
        self,
        ctx,
        client_id: str,
        secret: str,
        deployment_node_key: t.Optional[NodeKey],
        deployment_key_path: str,
        environment: str,
        skip_delete_registration: bool,
    ):
        self.ctx = ctx
        self.client_id = client_id
        self.secret = secret
        self.deployment_node_key = deployment_node_key
        self.deployment_key_path = deployment_key_path
        self.environment = environment
        self.skip_delete_registration = skip_delete_registration
        self.endpoint_id = client_id
        self.items = None

    def lookup_project_id(self) -> None:
        "Lookup the project ID"

        # Use the endpoint's credentials to find the Auth project containing it.
        try:
            client: globus_sdk.response.GlobusHTTPResponse
            client = current_app.auth_client.get(f"/v2/api/clients/{self.client_id}")
        except Exception as e:
            if isinstance(e, globus_sdk.AuthAPIError) and e.code == "UNAUTHORIZED":
                raise click.ClickException(
                    "\n".join(
                        textwrap.wrap(
                            "Error authenticating to Globus with the "
                            "endpoint's client credentials. This can "
                            "occur if the endpoint has already been "
                            "deleted or if the client_id or secret are "
                            "no longer valid."
                        )
                        + [""]
                        + textwrap.wrap(
                            "Verify the client_id and secret on "
                            "{}v2/web/developers".format(
                                globus_sdk.config.get_service_url(
                                    "auth",
                                    environment=os.environ.get(
                                        "GLOBUS_SDK_ENVIRONMENT", "production"
                                    ),
                                )
                            )
                        )
                    )
                )
            current_app.logger.warn(str(e))
            raise click.ClickException(
                "\n".join(
                    textwrap.wrap(
                        "An error occurred while looking up the endpoint's project "
                        f"ID. {e}"
                    )
                )
            )

        if not client.get("client", {}).get("project", None):
            current_app.logger.warn(
                f"lookup_project_id(): Unexpected response from Auth: {client.data}"
            )
            raise click.ClickException(
                "\n".join(
                    textwrap.wrap(
                        "Unexpected response when looking up the endpoint's "
                        "project ID. Please report this to support@globus.org."
                    )
                )
            )

        self.project_id = client["client"]["project"]

    def do_auth_flow(self) -> None:
        """
        Perform any necessary Auth flows. Outputs self.auth_token.
        """
        if os.getenv("GCS_CLI_CLIENT_ID") and os.getenv("GCS_CLI_CLIENT_SECRET"):
            self._do_client_auth_flow()
        else:
            self._do_link_auth_flow()

    def _do_client_auth_flow(self) -> None:
        current_app.logger.info("_do_client_auth_flow: start")
        self.auth_token = globus_sdk.ClientCredentialsAuthorizer(
            globus_sdk.ConfidentialAppAuthClient(
                t.cast(str, os.getenv("GCS_CLI_CLIENT_ID")),
                t.cast(str, os.getenv("GCS_CLI_CLIENT_SECRET")),
                environment=self.environment,
            ),
            ["urn:globus:auth:scope:auth.globus.org:manage_projects"],
        ).access_token
        current_app.logger.info("_do_client_auth_flow: end")

    def _do_link_auth_flow(self):
        "Perform auth flows"
        assert getattr(self, "project_id", None) is not None

        current_app.logger.info("_do_link_auth_flow: start")

        session_params = {"prompt": "login"}
        session_message = "delete the endpoint credentials "
        scopes = ["urn:globus:auth:scope:auth.globus.org:manage_projects"]

        auth_client = internal_auth_client()
        try:
            auth_client.oauth2_start_flow(
                redirect_uri=auth_client.base_url + "v2/web/auth-code",
                refresh_tokens=False,
                requested_scopes=scopes,
            )
            session_params["session_message"] = session_message.capitalize()
            linkprompt = f"Please authenticate with Globus to {session_message}"
            click.echo(
                "\n{0}:\n{1}\n{2}\n{1}\n".format(
                    linkprompt,
                    "-" * len(linkprompt),
                    auth_client.oauth2_get_authorize_url(query_params=session_params),
                )
            )
            auth_code = click.prompt(
                "Enter the resulting Authorization Code here", default=""
            ).strip()
            if not auth_code:
                raise Exception("No auth code")
            token_response = auth_client.oauth2_exchange_code_for_tokens(auth_code)
        except Exception as e:
            raise click.ClickException(f"Authorization flow failed: {e}")

        token_resp = token_response.by_resource_server.get(
            globus_sdk.scopes.AuthScopes.resource_server
        )
        assert token_resp is not None  # mypy kludge
        self.auth_token = token_resp["access_token"]

        current_app.logger.info("_do_link_auth_flow: end")

    def verify_auth_token_identity(self):
        "Verify identity"
        assert getattr(self, "project_id", None) is not None
        assert getattr(self, "auth_token", None) is not None

        # Verify that the Auth token gives us access to the correct project.
        projects = []
        try:
            client = globus_sdk.AuthClient(
                authorizer=globus_sdk.AccessTokenAuthorizer(self.auth_token),
                app_name="GCS CLI endpoint cleanup",
            )
            projects = client.get("/v2/api/projects")
        except Exception as e:
            raise click.ClickException(
                "\n".join(
                    textwrap.wrap(f"Failed to retrieve the available projects. {e}")
                )
            )

        if self.project_id not in [p["id"] for p in projects["projects"]]:
            raise click.ClickException(
                "\n".join(
                    textwrap.wrap(
                        "The authorization flow was not authenticated with an "
                        f"identity that is an admin on Auth project {self.project_id}. "
                        "Rerun this command and authenticate with the correct identity."
                    )
                )
            )

    def synchronize_state(self):
        "Synchronize endpoint state"
        kc = None

        if self.deployment_node_key:
            kc = current_app.appsync_client.query_all(
                id=ConfigurationObjectTag.KEY_CHAIN.value
            )
            for k in kc:
                current_app.appsync_client.process_item(k)

        try:
            self.items = current_app.appsync_client.query_all()
        except GraphQLError:
            self.items = []

        for it in self.items:
            if it["tag"] == ConfigurationObjectTag.ENDPOINT:
                if self.deployment_node_key:
                    self.endpoint_id = current_app.keychain.decrypt(
                        it["configuration"]
                    )["id"]

        if kc:
            current_app.appsync_client.commit_items_to_session(
                models.db.session, self.items
            )

    def check_nodes(self):
        "Check for existing node"
        domain = models.GlobusDnsDomain.query.one_or_none()

        if domain:
            existing_nodes = []
            for node in domain.nodes:
                existing_nodes.append(node)

            if existing_nodes:
                text = "\n".join(
                    ["\n"]
                    + textwrap.wrap(
                        "There are still node(s) configured to serve "
                        "this endpoint running on the following IP addresses:",
                        width=78,
                    )
                    + [""]
                    + [
                        "    " + ", ".join([str(n) for n in node.ip_addresses])
                        for node in existing_nodes
                    ]
                    + [""]
                    + textwrap.wrap(
                        "If you still have access to these nodes, log into "
                        "them and run the following command as root to remove "
                        "them from the endpoint and disable the Globus services "
                        "on them:",
                        width=78,
                    )
                    + ["\n    globus-connect-server node cleanup"]
                    + [""]
                )
                click.echo(text)
                if not click.confirm("Continue without cleaning up nodes?"):
                    self.ctx.exit(1)

    def check_oidc(self) -> None:
        "Check for OIDC server"
        endpoint: t.Optional[models.Endpoint] = models.Endpoint.query.one_or_none()

        if endpoint and endpoint.has_oidc_server:
            text = "\n".join(
                ["\n"]
                + textwrap.wrap(
                    "This endpoint configuration contains a Globus OIDC server "
                    "for the {} domain. If you delete this endpoint, the OIDC "
                    "server will also be deleted, and the domain ".format(
                        endpoint.oidc_server.domain_name
                    )
                    + click.style("can not be reused", bold=True)
                    + " as a Globus identity domain."
                )
            )
            click.echo(text)
            if not click.confirm("Continue deleting this endpoint?"):
                self.ctx.exit(1)
            endpoint.oidc_server.delete()
        elif endpoint and endpoint.has_oidc_registration:
            text = "\n".join(
                ["\n"]
                + textwrap.wrap(
                    "This endpoint configuration owns an OIDC server "
                    "registration for the {} domain. If you delete this "
                    "endpoint, the OIDC server registration will also be "
                    "deleted, and the domain will no longer be valid "
                    " as a Globus identity domain.".format(
                        endpoint.oidc_registration.domain_name
                    )
                )
            )
            click.echo(text)
            if not click.confirm("Continue deleting this endpoint?"):
                self.ctx.exit(1)
            endpoint.oidc_registration.delete()

    def delete_collections(self):
        "Delete collections"
        for collection in models.Collection.query.filter(
            ~models.Collection.deleted, models.Collection.is_guest_collection
        ):
            try:
                collection.delete()
            except globus_sdk.GlobusAPIError as e:
                if e.http_status != 404:
                    current_app.logger.debug("Error deleting collection %s", str(e))

        for collection in models.Collection.query.filter(
            ~models.Collection.deleted, models.Collection.is_mapped_collection
        ):
            try:
                collection.delete()
            except globus_sdk.GlobusAPIError as e:
                if e.http_status != 404:
                    current_app.logger.warn("Unable to delete collection %s", str(e))

        for ep in current_app.transfer_client.endpoint_search(
            filter_scope="my-endpoints",
        ):
            if not ep.get("non_functional"):
                try:
                    current_app.transfer_client.delete_endpoint(ep["id"])
                except globus_sdk.GlobusAPIError as e:
                    if e.http_status != 404:
                        current_app.logger.warn(
                            "Unable to delete collection %s", str(e)
                        )
            elif not self.endpoint_id:
                self.endpoint_id = ep["id"]

    def delete_user_credentials(self):
        "Delete user credentials"
        for uc in models.UserCredential.query.filter(~models.UserCredential.deleted):
            try:
                uc.delete()
            except Exception as e:
                current_app.logger.warn(
                    "Unable to delete user credential %s: %s", uc, str(e)
                )

    def delete_domain(self):
        "Delete DNS name"
        domain = models.GlobusDnsDomain.query.one_or_none()
        if domain:
            try:
                domain.delete()
            except Exception as e:
                current_app.logger.warn("Unable to delete DNS name: %s", str(e))

    def delete_endpoint(self):
        "Delete endpoint"
        try:
            if self.endpoint_id:
                current_app.transfer_client.delete_endpoint(self.endpoint_id)
        except globus_sdk.TransferAPIError as e:
            if e.http_status != 404:
                current_app.logger.warn("Unable to delete endpoint: %s", str(e))

    def delete_auth_clients(self):
        "Delete auth clients"
        clients = []
        try:
            result = current_app.auth_client.get("/v2/api/clients")
            clients = result.data.get("clients")
        except globus_sdk.GlobusAPIError as e:
            current_app.logger.warning(e)

        for client in clients:
            try:
                client_id = client["id"]
                if client["id"] != self.client_id:
                    current_app.auth_client.delete(
                        "/v2/api/clients/{}".format(client_id)
                    )
            except Exception as e:
                current_app.logger.warning(e)

    def delete_appsync_data(self):
        "Delete appsync data"

        endpoint_object = None
        keychain_object = None

        all_deleted = True

        for item in self.items:
            if item["tag"] == ConfigurationObjectTag.ENDPOINT:
                endpoint_object = item
            elif item["tag"] == ConfigurationObjectTag.KEY_CHAIN:
                keychain_object = item
            else:
                try:
                    current_app.appsync_client.delete_object(id=item["id"])
                except Exception as e:
                    current_app.logger.warn(
                        "Unable to delete %s object %s: %s",
                        item["tag"].value,
                        item["id"],
                        str(e),
                    )
                    all_deleted = False

        if all_deleted and keychain_object:
            try:
                current_app.appsync_client.delete_object(id=keychain_object["id"])
            except Exception as e:
                current_app.logger.warn(
                    "Unable to delete %s object %s: %s",
                    keychain_object["tag"].value,
                    keychain_object["id"],
                    str(e),
                )
                all_deleted = False

        if all_deleted and endpoint_object:
            try:
                current_app.appsync_client.delete_object(id=endpoint_object["id"])
            except Exception as e:
                current_app.logger.warn(
                    "Unable to delete %s object %s: %s",
                    endpoint_object["tag"].value,
                    endpoint_object["id"],
                    str(e),
                )
                all_deleted = False
        if not all_deleted:
            current_app.logger.info("Not all objects deleted cleanly")

    def delete_endpoint_registration(self):
        "Delete endpoint's registration"
        assert getattr(self, "project_id", None) is not None
        assert getattr(self, "auth_token", None) is not None

        try:
            client = globus_sdk.AuthClient(
                authorizer=globus_sdk.AccessTokenAuthorizer(self.auth_token),
                app_name="GCS CLI endpoint cleanup",
            )
            client.delete(
                f"/v2/api/clients/{self.client_id}",
            )
        except Exception as e:
            current_app.logger.warning(e)
            click.echo(
                "\n".join(
                    textwrap.wrap(
                        f"Failed to delete client {self.client_id} from project "
                        f"{self.project_id}. You'll need to delete it manually.",
                    )
                ),
                err=True,
            )

    def delete_deployment_key(self):
        """
        Delete deployment key
        """
        try:
            os.unlink(self.deployment_key_path)
            click.echo(
                "The deployment key file is no longer usable and has been deleted."
            )
        except Exception as e:
            click.echo(
                "\n".join(
                    textwrap.wrap(f"Could not delete the deployment key: {e}.")
                    + [""]
                    + textwrap.wrap(
                        "They deployment key file is no longer usable "
                        "and should be deleted."
                    )
                )
            )

    def run(self):
        tasks = [
            self.synchronize_state,
            self.check_nodes,
            self.check_oidc,
            self.delete_collections,
            self.delete_user_credentials,
            self.delete_domain,
            self.delete_endpoint,
            self.delete_auth_clients,
            self.delete_appsync_data,
        ]

        if not self.skip_delete_registration:
            # Add interactive/initializers to the front
            tasks.insert(0, self.lookup_project_id)
            tasks.insert(1, self.do_auth_flow)
            tasks.insert(2, self.verify_auth_token_identity)
            # Add endpoint deletion to the back
            tasks += [self.delete_endpoint_registration]

        def show(obj):
            if obj and obj.__doc__:
                return obj.__doc__.split("\n")[0]
            return ""

        with click.progressbar(tasks, item_show_func=show) as bar:
            for task in bar:
                task()

        self.delete_deployment_key()
