from __future__ import annotations

import os
import textwrap
import typing as t
import uuid
from pathlib import Path

import click
from flask import current_app
from jwcrypto.jwe import InvalidJWEData

import globus_sdk

from globus.manager.api.app import make_app
from globus.manager.api.models import db
from globus.manager.appsync.query import GraphQLError
from globus.manager.constants import ConfigurationObjectTag
from globus.manager.utils.keychain import NodeKey

from .utils.deployment_key import DeploymentKey, store_deployment_key

if t.TYPE_CHECKING:
    from collections import abc


def endpoint_key_convert(
    client_id: uuid.UUID,
    deployment_node_key: t.Optional[NodeKey],
    deployment_key_path: str,
    secret: str,
    verbose: bool,
) -> None:
    environment = os.environ.get("GLOBUS_SDK_ENVIRONMENT", "production")
    if verbose:
        os.environ["GCS_MANAGER_LOG_LEVEL"] = "DEBUG"

    app = make_app(app_config="globus.manager.api.config.endpoint_cli")
    db.init_app(app)

    app.config["CLIENT_ID"] = str(client_id)
    app.config["CLIENT_SECRET"] = secret
    app.config["ENVIRONMENT"] = environment

    if deployment_node_key:
        app.config["DEPLOYMENT_KEY"] = deployment_node_key
        app.keychain.add_node_key(deployment_node_key)

    with app.app_context():
        db.create_all()

        click.echo("Verifying ID and secret...")
        key_chain_items = query_key_chain()

        if not key_chain_items:
            raise click.ClickException(f"Endpoint {client_id} does not exist")

        if deployment_node_key:
            click.echo("Verifying old deployment key...")
            import_key_chain(key_chain_items)
        else:
            click.echo("Skipping verification of old deployment key...")
            click.echo("Creating a deployment key without a valid node key...")

    click.echo("Exporting the new deployment key...")
    store_deployment_key(
        Path(deployment_key_path),
        DeploymentKey(deployment_node_key, client_id, secret),
    )


def query_key_chain() -> list[dict[str, t.Any]]:
    """
    Initiate access of current_app.appsync_client which forces the client to
    connect and authenticate with Auth using the endpoint credentials in order
    to get a token. This will verify that the endpoint credentials are correct.
    """
    try:
        return current_app.appsync_client.query_all(
            id=ConfigurationObjectTag.KEY_CHAIN.value,
        )
    except globus_sdk.AuthAPIError as e:
        if e.code != "UNAUTHORIZED":
            raise

        raise click.ClickException(
            "\n".join(
                textwrap.wrap(
                    "Error authenticating to Globus with the "
                    "endpoint's client credentials. This can "
                    "occur if the client_id "
                    "or secret has been entered incorrectly or if the "
                    "client_id or secret are no longer valid."
                )
                + [""]
                + textwrap.wrap(
                    "Verify the client_id and secret on "
                    "{}v2/web/developers".format(
                        globus_sdk.config.get_service_url(
                            "auth",
                            environment=os.environ.get(
                                "GLOBUS_SDK_ENVIRONMENT", "production"
                            ),
                        )
                    )
                )
            )
        )


def import_key_chain(items: abc.Sequence[abc.Mapping[str, t.Any]]) -> None:
    """
    Import the key chain which will make use of our node_key to decrypt it
    thereby verifying that we have a usable node_key.
    """
    try:
        for item in items:
            current_app.appsync_client.process_item(item, catch_exceptions=False)
    except GraphQLError as e:
        raise click.ClickException(str(e))
    except InvalidJWEData as e:
        current_app.logger.error("download_keychain_data: %s", str(e))
        raise click.ClickException(
            "Unable to unlock keychain. "
            "This is likely caused by using the incorrect deployment key."
        )
